Require Import Reals.
Require Import Ring_tac.
Require Import NewField.
Import Pol.



Module RField .
Module insideRing := REntries.
Import insideRing.
 
Definition rdiv := Rdiv.
 
Definition rinv := Rinv.
 
Definition rI_neq_rO: ~ req rI rO.
apply R1_neq_R0.
Qed.
 
Theorem rdiv_def: forall p q,  req (rdiv p q) (rmul p (rinv q)).
intros p q; exact (refl_equal (Rdiv p q)).
Qed.
 
Theorem rinv_l: forall p, ~ req p rO ->  req (rmul (rinv p) p) rI.
exact Rinv_l.
Qed.
 
Theorem SRinv_ext: forall p q, req p q ->  req (rinv p) (rinv q).
intros p q; apply f_equal with ( f := rinv ); auto.
Qed.
 

End RField.

Module Rfield := MakeFieldPol RField.
Import Rfield.
Import REntries.


Fixpoint Fcons (e : PExpr) (l : list PExpr) {struct l} : list PExpr :=
 match l with
   nil => cons e (nil PExpr)
  | cons a l1 =>
      match PExpr_eq e a with true => l | false => cons a (Fcons e l1) end
 end.
 
Theorem PFcons_fcons_inv:
 forall l a l1, PCond l (Fcons a l1) ->  ~ req (PEeval l a) rO /\ PCond l l1.
intros l a l1; elim l1; simpl Fcons; auto.
simpl; auto.
intros a0 l0.
generalize (PExpr_eq_semi_correct l a a0); case (PExpr_eq a a0).
intros H H0 H1; split; auto.
rewrite H; auto.
generalize (PCond_cons_inv_l _ _ _ H1); simpl; auto.
intros H H0 H1;
 assert (Hp: ~ req (PEeval l a0) rO /\ (~ req (PEeval l a) rO /\ PCond l l0)).
split.
generalize (PCond_cons_inv_l _ _ _ H1); simpl; auto.
apply H0.
generalize (PCond_cons_inv_r _ _ _ H1); simpl; auto.
generalize Hp; case l0; simpl; intuition.
Qed.
 
Fixpoint Fcons1 (e : PExpr) (l : list PExpr) {struct e} : list PExpr :=
 match e with
   PEmul e1 e2 => Fcons1 e1 (Fcons1 e2 l)
  | PEc c => if ceqb c Z0 then Fcons e l else l
  | _ => Fcons e l
 end.
 
 
Theorem gen_phiPOS1_IZR:
 forall p,
  gen_phiPOS Rdefinitions.R1 Rdefinitions.Rplus Rdefinitions.Rmult p =
  IZR (Zpos p).
intros p; elim p; simpl gen_phiPOS.
intros p0; case p0.
intros p1 H; rewrite H.
unfold IZR; rewrite (nat_of_P_xI (xI p1)); (try rewrite S_INR);
 (try rewrite plus_INR); (try rewrite mult_INR).
pose (x:= INR (nat_of_P (xI p1))); fold x; simpl; ring.
intros p1 H; rewrite H.
unfold IZR; rewrite (nat_of_P_xI (xO p1)); (try rewrite S_INR);
 (try rewrite plus_INR); (try rewrite mult_INR).
pose (x:= INR (nat_of_P (xO p1))); fold x; simpl; ring.
simpl; intros; ring.
intros p0; case p0.
intros p1 H; rewrite H.
unfold IZR; rewrite (nat_of_P_xO (xI p1)); (try rewrite S_INR);
 (try rewrite plus_INR); (try rewrite mult_INR).
pose (x:= INR (nat_of_P (xO p1))); fold x; simpl; ring.
intros p1 H; rewrite H.
unfold IZR; rewrite (nat_of_P_xO (xO p1)); (try rewrite S_INR);
 (try rewrite plus_INR); (try rewrite mult_INR).
pose (x:= INR (nat_of_P (xO p1))); fold x; simpl; ring.
simpl; intros; ring.
simpl; intros; ring.
Qed.
 
Theorem gen_phiZ1_IZR:
 forall p,
  gen_phiZ
   Rdefinitions.R0 Rdefinitions.R1 Rdefinitions.Rplus Rdefinitions.Rmult
   Rdefinitions.Ropp p = IZR p.
intros p; case p; simpl; auto.
intros p0; rewrite gen_phiPOS1_IZR; auto.
intros p0; rewrite gen_phiPOS1_IZR; auto.
Qed.
 
Theorem PFcons1_fcons_inv:
 forall l a l1, PCond l (Fcons1 a l1) ->  ~ req (PEeval l a) rO /\ PCond l l1.
intros l a; elim a; (try (intros; apply PFcons_fcons_inv; auto; fail)).
intros c; simpl; unfold ceqb, Zeq_bool.
generalize (Zcompare_Eq_iff_eq c Z0).
case (Zcompare c Z0); simpl; auto.
intros H l1 H0; apply (PFcons_fcons_inv _ _ _ H0).
intros H l1 H0; split; auto.
unfold RField.insideRing.phi.
rewrite gen_phiZ1_IZR; intros H1.
absurd (c = Z0).
intros H2; case H; auto.
intros H3 H4; generalize (H4 H2); intros; discriminate.
apply eq_IZR; simpl; auto.
intros H l1 H0; split; auto.
unfold RField.insideRing.phi.
rewrite gen_phiZ1_IZR; intros H1.
absurd (c = Z0).
intros H2; case H; auto.
intros H3 H4; generalize (H4 H2); intros; discriminate.
apply eq_IZR; simpl; auto.
intros p H p0 H0 l1 H1.
simpl in H1 |-.
case (H _ H1); intros H2 H3.
case (H0 _ H3); intros H4 H5; split; auto.
intros H6; simpl in H6 |-; case (Rmult_integral _ _ H6).
intros H7; case H2; auto.
intros H7; case H4; auto.
Qed.

Definition Fcons2 := fun e l => Fcons1 (PExpr_simp e) l.
 
Theorem PFcons2_fcons_inv:
 forall l a l1, PCond l (Fcons2 a l1) ->  ~ req (PEeval l a) rO /\ PCond l l1.
unfold Fcons2; intros l a l1 H; split;
 case (PFcons1_fcons_inv l (PExpr_simp a) l1); auto.
intros H1 H2 H3; case H1.
apply eq_trans with ( 2 := H3 ).
apply PExpr_simp_correct.
Qed.


Ltac rfield :=
  let T := Rfield.Make_field_tac in
  T Rplus Rmult Rminus Ropp Rinv Rdiv (@eq R) Fcons2  PFcons2_fcons_inv RCst.

Ltac unfold_field := try (unfold RField.insideRing.req,
                              RField.insideRing.rmul, RAE.rmul,
                              RField.insideRing.radd, RAE.radd,
                              RField.insideRing.rsub, RAE.rsub,
                              RField.insideRing.ropp, RAE.ropp,
                              RField.insideRing.rmul, RAE.rmul,
                              RField.rdiv,
                              RField.insideRing.rI, RAE.rI,
                              RField.insideRing.rO, RAE.rO).

Ltac newfield :=  
 match goal with
 | |- (?r1 = ?r2) =>
   match (type of r1) with
   | _ =>  (apply Rminus_diag_uniq;
           rfield; unfold_field)
   end
  |  _ => fail  "not equality"
 end.


Ltac newfield_rewrite r :=
  let T := Rfield.Make_field_rewrite in
  T Rplus Rmult Rminus Ropp Rinv Rdiv Fcons2 PFcons2_fcons_inv RCst r;
  unfold_field.

	      

