Require Import Pol.
Require Export BinPos.


Ltac FV1 Cst add mul sub opp inv div t fv :=
 let rec TFV t fv :=
  match t with
  | (add ?t1 ?t2) => 
    let fv1 := TFV t1 fv in let fv2 := TFV t2 fv1 in constr:fv2
  | (mul ?t1 ?t2) => 
    let fv1 := TFV t1 fv in let fv2 := TFV t2 fv1 in constr:fv2
  | (sub ?t1 ?t2) => 
    let fv1 := TFV t1 fv in let fv2 := TFV t2 fv1 in constr:fv2
  | (opp ?t1) => 
    let fv1 := TFV t1 fv in constr:fv1
  | (inv ?t1) => 
    let fv1 := TFV t1 fv in constr:fv1
  | (div ?t1 ?t2) => 
    let fv1 := TFV t1 fv in let fv2 := TFV t2 fv1 in constr:fv2
  | _ =>  
    match Cst t with
    | RBfalse => let fv1 := AddFv t fv in constr:fv1
    | _ => constr:fv
    end
  end 
 in TFV t fv.


Module Type FIELD_POL_ENTRIES.
(* Ring elements *)
Declare Module insideRing :RING_POL_ENTRIES.
Import insideRing.
Parameters (rdiv : R -> R ->  R) (rinv : R ->  R).
Parameter rI_neq_rO : ~ req rI rO.
Parameter rdiv_def : forall p q,  req (rdiv p q) (rmul p (rinv q)).
Parameter rinv_l : forall p, ~ req p rO ->  req (rmul (rinv p) p) rI.
Parameter SRinv_ext : forall p q, req p q ->  req (rinv p) (rinv q).
 
End FIELD_POL_ENTRIES.

Module MakeFieldPol (FE:FIELD_POL_ENTRIES).
Import FE.
Import FE.insideRing.
Module Rring := MakeRingPol insideRing.
Export Rring.
 
Let eq_trans := Setoid.Seq_trans _ _ Rsth.
 
Let eq_sym := Setoid.Seq_sym _ _ Rsth.
 
Let eq_refl := Setoid.Seq_refl _ _ Rsth.
Hint Resolve eq_refl rdiv_def rinv_l rI_neq_rO CRmorph.(morph1) .
Hint Resolve (Rmul_ext Reqe) (Rmul_ext Reqe) (Radd_ext Reqe) (ARsub_ext Rsth Reqe ARth) (Ropp_ext Reqe) SRinv_ext.
Hint Resolve (ARadd_0_l  ARth) (ARadd_sym  ARth) (ARadd_assoc ARth) (ARmul_1_l  ARth) (ARmul_0_l  ARth) 
            (ARmul_sym  ARth) (ARmul_assoc ARth) (ARdistr_l  ARth) (ARopp_mul_l ARth) (ARopp_add  ARth) 
            (ARsub_def  ARth) .

(***************************************************************************
                                                                          
                       Properties of division                             
                                                                          
  ***************************************************************************)
 
Theorem rdiv_simpl: forall p q, ~ req q rO ->  req (q * rdiv p q) p.
intros p q H; apply eq_trans with (rmul q (rmul p (rinv q))); auto.
apply eq_trans with ((p * rinv q) * q); auto.
apply eq_trans with (p * (rinv q * q)); auto.
apply eq_sym; auto.
apply eq_trans with (p * 1); auto.
apply eq_trans with (1 * p); auto.
Qed.
Hint Resolve rdiv_simpl .
 
Theorem SRdiv_ext:
 forall p1 p2 q1 q2, req p1 p2 -> req q1 q2 ->  req (rdiv p1 q1) (rdiv p2 q2).
intros p1 p2 q1 q2 H H0.
apply eq_trans with (p1 * rinv q1); auto.
apply eq_trans with (p2 * rinv q2); auto.
apply eq_sym; auto.
Qed.
Hint Resolve SRdiv_ext .

(***************************************************************************
                                                                          
                       Some equality test                             
                                                                          
  ***************************************************************************)

Fixpoint positive_eq (p1 p2 : positive) {struct p1} : bool :=
 match p1, p2 with
   xH, xH => true
  | xO p3, xO p4 => positive_eq p3 p4
  | xI p3, xI p4 => positive_eq p3 p4
  | _, _ => false
 end.
 
Theorem positive_eq_correct:
 forall p1 p2,  if positive_eq p1 p2 then p1 = p2 else p1 <> p2.
intros p1; elim p1;
 (try (intros p2; case p2; simpl; auto; intros; discriminate)).
intros p3 rec p2; case p2; simpl; auto; (try (intros; discriminate)); intros p4.
generalize (rec p4); case (positive_eq p3 p4); auto.
intros H1; apply f_equal with ( f := xI ); auto.
intros H1 H2; case H1; injection H2; auto.
intros p3 rec p2; case p2; simpl; auto; (try (intros; discriminate)); intros p4.
generalize (rec p4); case (positive_eq p3 p4); auto.
intros H1; apply f_equal with ( f := xO ); auto.
intros H1 H2; case H1; injection H2; auto.
Qed.
 
Fixpoint PExpr_eq (e1 e2 : PExpr) {struct e1} : bool :=
 match e1, e2 with
   PEc c1, PEc c2 => ceqb c1 c2
  | PEX p1, PEX p2 => positive_eq p1 p2
  | PEadd e3 e5, PEadd e4 e6 => if PExpr_eq e3 e4 then PExpr_eq e5 e6 else false
  | PEsub e3 e5, PEsub e4 e6 => if PExpr_eq e3 e4 then PExpr_eq e5 e6 else false
  | PEmul e3 e5, PEmul e4 e6 => if PExpr_eq e3 e4 then PExpr_eq e5 e6 else false
  | PEopp e3, PEopp e4 => PExpr_eq e3 e4
  | _, _ => false
 end.
 
Theorem PExpr_eq_semi_correct:
 forall l e1 e2, PExpr_eq e1 e2 = true ->  req (PEeval l e1) (PEeval l e2).
intros l e1; elim e1.
intros c1; intros e2; elim e2; simpl; (try (intros; discriminate)).
intros c2; apply (morph_eq CRmorph).
intros p1; intros e2; elim e2; simpl; (try (intros; discriminate)).
intros p2; generalize (positive_eq_correct p1 p2); case (positive_eq p1 p2);
 (try (intros; discriminate)); intros H; rewrite H; auto.
intros e3 rec1 e5 rec2 e2; case e2; simpl; (try (intros; discriminate)).
intros e4 e6; generalize (rec1 e4); case (PExpr_eq e3 e4);
 (try (intros; discriminate)); generalize (rec2 e6); case (PExpr_eq e5 e6);
 (try (intros; discriminate)); auto.
intros e3 rec1 e5 rec2 e2; case e2; simpl; (try (intros; discriminate)).
intros e4 e6; generalize (rec1 e4); case (PExpr_eq e3 e4);
 (try (intros; discriminate)); generalize (rec2 e6); case (PExpr_eq e5 e6);
 (try (intros; discriminate)); auto.
intros e3 rec1 e5 rec2 e2; case e2; simpl; (try (intros; discriminate)).
intros e4 e6; generalize (rec1 e4); case (PExpr_eq e3 e4);
 (try (intros; discriminate)); generalize (rec2 e6); case (PExpr_eq e5 e6);
 (try (intros; discriminate)); auto.
intros e3 rec e2; (case e2; simpl; (try (intros; discriminate))).
intros e4; generalize (rec e4); case (PExpr_eq e3 e4);
 (try (intros; discriminate)); auto.
Qed.

Definition NPEadd :=
   fun e1 e2 =>
      match e1, e2 with
        PEc c1, PEc c2 => PEc (cadd c1 c2)
       | _, _ => PEadd e1 e2
      end.
 
Theorem NPEadd_correct:
 forall l e1 e2,  req (PEeval l (NPEadd e1 e2)) (PEeval l (PEadd e1 e2)).
intros l e1 e2; case e1; simpl; auto.
intros c; case e2; simpl; auto.
intros; apply (morph_add CRmorph).
Qed.
 
Definition NPEmul :=
   fun e1 e2 =>
      match e1, e2 with
        PEc c1, PEc c2 => PEc (cmul c1 c2)
       | _, _ => PEmul e1 e2
      end.
 
Theorem NPEmul_correct:
 forall l e1 e2,  req (PEeval l (NPEmul e1 e2)) (PEeval l (PEmul e1 e2)).
intros l e1 e2; case e1; simpl; auto.
intros c; case e2; simpl; auto.
intros; apply (morph_mul CRmorph).
Qed.
 
Definition NPEsub :=
   fun e1 e2 =>
      match e1, e2 with
        PEc c1, PEc c2 => PEc (csub c1 c2)
       | _, _ => PEsub e1 e2
      end.
 
Theorem NPEsub_correct:
 forall l e1 e2,  req (PEeval l (NPEsub e1 e2)) (PEeval l (PEsub e1 e2)).
intros l e1 e2; case e1; simpl; auto.
intros c; case e2; simpl; auto.
intros; apply (morph_sub CRmorph).
Qed.
 
Definition NPEopp :=
   fun e1 =>
      match e1 with PEc c1 => PEc (copp c1) | _ => PEopp e1 end.
 
Theorem NPEopp_correct:
 forall l e1,  req (PEeval l (NPEopp e1)) (PEeval l (PEopp e1)).
intros l e1; case e1; simpl; auto.
intros; apply (morph_opp CRmorph).
Qed.
 
Fixpoint PExpr_simp (e : PExpr) : PExpr :=
 match e with
   PEadd e1 e2 => NPEadd (PExpr_simp e1) (PExpr_simp e2)
  | PEmul e1 e2 => NPEmul (PExpr_simp e1) (PExpr_simp e2)
  | PEsub e1 e2 => NPEsub (PExpr_simp e1) (PExpr_simp e2)
  | PEopp e1 => NPEopp (PExpr_simp e1)
  | _ => e
 end.
 
Theorem PExpr_simp_correct:
 forall l e,  req (PEeval l (PExpr_simp e)) (PEeval l e).
intros l e; elim e; simpl; auto.
intros e1 He1 e2 He2.
apply eq_trans with (PEeval l (PEadd (PExpr_simp e1) (PExpr_simp e2))); auto.
apply NPEadd_correct.
simpl; auto.
intros e1 He1 e2 He2.
apply eq_trans with (PEeval l (PEsub (PExpr_simp e1) (PExpr_simp e2))); auto.
apply NPEsub_correct.
simpl; auto.
intros e1 He1 e2 He2.
apply eq_trans with (PEeval l (PEmul (PExpr_simp e1) (PExpr_simp e2))); auto.
apply NPEmul_correct.
simpl; auto.
intros e1 He1.
apply eq_trans with (PEeval l (PEopp (PExpr_simp e1))); auto.
apply NPEopp_correct.
simpl; auto.
Qed.


(***************************************************************************
                                                                          
                       Properties of division                             
                                                                          
  ***************************************************************************)
 
Theorem rdiv1: forall r,  req r (rdiv r rI).
intros r; apply eq_trans with (rmul rI (rdiv r rI)); auto.
apply eq_sym; apply rdiv_simpl; auto.
Qed.
 
Theorem rdiv2:
 forall r1 r2 r3 r4,
 ~ req (rmul r2 r4) rO ->
  req
   (radd (rdiv r1 r2) (rdiv r3 r4))
   (rdiv (radd (rmul r1 r4) (rmul r3 r2)) (rmul r2 r4)).
intros r1 r2 r3 r4 H1.
apply eq_trans with (radd (rmul r1 (rinv r2)) (rmul r3 (rinv r4))); auto.
apply eq_trans with (rmul 1 (r1 * rinv r2 + r3 * rinv r4)); auto.
apply eq_sym; auto.
apply eq_trans
     with (rmul (rmul (rinv (r2 * r4)) (r2 * r4)) (r1 * rinv r2 + r3 * rinv r4));
 auto.
apply eq_sym; auto.
apply eq_trans
     with (rinv (r2 * r4) * ((r2 * r4) * (r1 * rinv r2 + r3 * rinv r4))); auto.
apply eq_sym; auto.
apply eq_trans
     with (((r2 * r4) * (r1 * rinv r2 + r3 * rinv r4)) * rinv (r2 * r4)); auto.
apply eq_trans with ((r1 * r4 + r3 * r2) * rinv (r2 * r4)); auto.
2:apply eq_sym; auto.
apply (Rmul_ext Reqe); auto.
apply eq_trans with ((r2 * r4) * (r1 * rinv r2) + (r2 * r4) * (r3 * rinv r4));
 auto.
apply eq_trans with ((r1 * rinv r2 + r3 * rinv r4) * (r2 * r4)); auto.
apply eq_trans with ((r1 * rinv r2) * (r2 * r4) + (r3 * rinv r4) * (r2 * r4));
 auto.
assert (Hk:
 forall r1 r2 r4, ~ req r2 rO ->  req ((r2 * r4) * (r1 * rinv r2)) (r1 * r4));
 auto.
clear r1 r2 r3 r4 H1; intros r1 r2 r4 H2.
apply eq_trans with ((r1 * rinv r2) * (r2 * r4)); auto.
apply eq_trans with (r1 * (rinv r2 * (r2 * r4))); auto.
apply eq_sym; auto.
apply (Rmul_ext Reqe); auto.
apply eq_trans with ((rinv r2 * r2) * r4); auto.
apply eq_trans with (rI * r4); auto.
assert (K1: ~ req r2 rO).
intros H2; case H1.
apply eq_trans with (0 * r4); auto.
assert (K2: ~ req r4 rO).
intros H2; case H1.
apply eq_trans with (r4 * r2); auto.
apply eq_trans with (0 * r2); auto.
apply (Radd_ext Reqe); auto.
apply eq_trans with ((r4 * r2) * (r3 * rinv r4)); auto.
Qed.
 
Theorem rdiv5: forall r1 r2,  req (- rdiv r1 r2) (rdiv (- r1) r2).
intros r1 r2.
apply eq_trans with (- (r1 * rinv r2)); auto.
apply eq_trans with (- r1 * rinv r2); auto.
apply eq_sym; auto.
Qed.
Hint Resolve rdiv5 .
 
Theorem rdiv3:
 forall r1 r2 r3 r4,
 ~ req (rmul r2 r4) rO ->
  req
   (rsub (rdiv r1 r2) (rdiv r3 r4))
   (rdiv (rsub (rmul r1 r4) (rmul r3 r2)) (rmul r2 r4)).
intros r1 r2 r3 r4 H.
apply eq_trans with (rdiv r1 r2 + - rdiv r3 r4); auto.
apply eq_trans with (rdiv r1 r2 + rdiv (- r3) r4); auto.
apply eq_trans with (rdiv (r1 * r4 + - r3 * r2) (r2 * r4)); auto.
apply rdiv2; auto.
apply SRdiv_ext; auto.
apply eq_trans with (r1 * r4 + - (r3 * r2)); apply eq_sym; auto.
Qed.
 
Theorem rdiv4:
 forall r1 r2 r3 r4,
 ~ req (rmul r2 r4) rO ->
  req (rmul (rdiv r1 r2) (rdiv r3 r4)) (rdiv (rmul r1 r3) (rmul r2 r4)).
intros r1 r2 r3 r4 H.
apply eq_trans with (1 * (rdiv r1 r2 * rdiv r3 r4)); auto.
apply eq_sym; auto.
apply eq_trans with ((rinv (r2 * r4) * (r2 * r4)) * (rdiv r1 r2 * rdiv r3 r4));
 auto.
apply eq_sym; auto.
apply eq_trans with (rinv (r2 * r4) * ((r2 * r4) * (rdiv r1 r2 * rdiv r3 r4)));
 auto.
apply eq_sym; auto.
apply eq_trans with ((r1 * r3) * rinv (r2 * r4)); auto.
2:apply eq_sym; auto.
apply eq_trans with (rinv (r2 * r4) * (r1 * r3)); auto.
apply (Rmul_ext Reqe); auto.
apply eq_trans with (r2 * (r4 * (rdiv r1 r2 * rdiv r3 r4))); auto.
apply eq_sym; auto.
apply eq_trans with (r2 * ((r4 * rdiv r1 r2) * rdiv r3 r4)); auto.
apply eq_trans with (r2 * ((rdiv r1 r2 * r4) * rdiv r3 r4)); auto.
apply eq_trans with (r2 * (rdiv r1 r2 * (r4 * rdiv r3 r4))); auto.
apply eq_sym; auto.
apply eq_trans with ((r2 * rdiv r1 r2) * (r4 * rdiv r3 r4)); auto.
assert (~ req r2 rO).
intros H1; case H.
apply eq_trans with (0 * r4); auto.
assert (~ req r4 rO); auto.
intros H1; case H.
apply eq_trans with (r2 * 0); auto.
apply eq_trans with (0 * r2); auto.
Qed.
 
Theorem rdiv6:
 forall r1 r2,
 ~ req r1 rO -> ~ req r2 rO ->  req (rinv (rdiv r1 r2)) (rdiv r2 r1).
intros r1 r2 H H0.
assert (Hk: ~ req (rdiv r1 r2) rO).
intros H1; case H.
apply eq_trans with (r2 * rdiv r1 r2); auto.
apply eq_sym; auto.
apply eq_trans with (r2 * rO); auto.
apply eq_trans with (rO * r2); auto.
apply eq_trans with (1 * rdiv r2 r1); auto.
apply eq_trans with ((rinv (rdiv r1 r2) * rdiv r1 r2) * rdiv r2 r1); auto.
apply eq_trans with (rinv (rdiv r1 r2) * (rdiv r1 r2 * rdiv r2 r1)); auto.
apply eq_trans with (rinv (rdiv r1 r2) * 1); auto.
apply eq_trans with (1 * rinv (rdiv r1 r2)); auto.
apply eq_sym; auto.
apply (Rmul_ext Reqe); auto.
apply eq_trans with ((r1 * rinv r2) * (r2 * rinv r1)); auto.
2:apply eq_sym; auto.
apply eq_trans with (((r1 * rinv r2) * r2) * rinv r1); auto.
2:apply eq_sym; auto.
apply eq_trans with ((r1 * (rinv r2 * r2)) * rinv r1); auto.
apply eq_trans with (rinv r1 * (r1 * (rinv r2 * r2))); auto.
apply eq_trans with ((rinv r1 * r1) * (rinv r2 * r2)); auto.
2:apply eq_sym; auto.
apply eq_trans with (rI * (rinv r2 * r2)); apply eq_sym; auto.
apply eq_trans with (rinv r2 * r2); auto.
Qed.
Hint Resolve rdiv6 .
 
Theorem rdiv7:
 forall r1 r2 r3 r4,
 ~ req (r2 * r3) rO ->
 ~ req r4 rO ->  req (rdiv (rdiv r1 r2) (rdiv r3 r4)) (rdiv (r1 * r4) (r2 * r3)).
intros r1 r2 r3 r4 H H0.
assert (Hk: ~ req r2 rO).
intros H1; case H.
apply eq_trans with (rO * r3); auto.
assert (Hk1: ~ req r3 rO).
intros H1; case H.
apply eq_trans with (r2 * rO); auto.
apply eq_trans with (rO * r2); auto.
apply eq_trans with (rdiv r1 r2 * rinv (rdiv r3 r4)); auto.
apply eq_trans with (rdiv r1 r2 * rdiv r4 r3); auto.
apply rdiv4; auto.
Qed.
 
Theorem rdiv8: forall r1 r2, ~ req r2 rO -> req r1 rO ->  req (rdiv r1 r2) rO.
intros r1 r2 H H0.
apply eq_trans with (r1 * rinv r2); auto.
apply eq_trans with (rO * rinv r2); auto.
Qed.

Add Morphism rdiv : rdiv_ext1.
exact SRdiv_ext.
Qed.
Add Morphism rinv : rinv_ext1.
exact SRinv_ext.
Qed.

(****************************************************************************
                                                                          
                               Datastructure                              
                                                                          
  ***************************************************************************)
 
Inductive FExpr : Type :=
  FEc: C ->  FExpr
 | FEX: positive ->  FExpr
 | FEadd: FExpr -> FExpr ->  FExpr
 | FEsub: FExpr -> FExpr ->  FExpr
 | FEmul: FExpr -> FExpr ->  FExpr
 | FEopp: FExpr ->  FExpr
 | FEinv: FExpr ->  FExpr
 | FEdiv: FExpr -> FExpr ->  FExpr .

   (* syntaxification *)
 Ltac mkFieldexpr Cst add mul sub opp inv div t fv := 
 let rec mkP t :=
  match t with 
  | (add ?t1 ?t2) => 
    let e1 := mkP t1 in
    let e2 := mkP t2 in constr:(FEadd e1 e2)
  | (mul ?t1 ?t2) => 
    let e1 := mkP t1 in
    let e2 := mkP t2 in constr:(FEmul e1 e2)
  | (sub ?t1 ?t2) => 
    let e1 := mkP t1 in
    let e2 := mkP t2 in constr:(FEsub e1 e2)
  | (opp ?t1) =>
    let e1 := mkP t1 in constr:(FEopp e1)
  | (inv ?t1) =>
    let e1 := mkP t1 in constr:(FEinv e1)
  | (div ?t1 ?t2) => 
    let e1 := mkP t1 in
    let e2 := mkP t2 in constr:(FEdiv e1 e2)
  | _ =>
    match Cst t with
    | RBfalse => let p := Find_at t fv in constr:(FEX p)
    | ?c => constr:(FEc c)
    end
  end
 in mkP t.

(* The result of the normalisation *)
 
Record linear : Type := mk_linear {
   num : PExpr;
   denum : PExpr;
   condition : list PExpr }.

(***************************************************************************
                                                                          
                               Normalisation                              
                                                                          
  ***************************************************************************)
 
Definition FPEmul :=
  fun x y => match x, y with
                (PEc c), _ => if (ceqb c cI) then y else (PEmul x y) 
             |  _, (PEc c) => if (ceqb c cI) then x else (PEmul x y) 
             |  _, _ => (PEmul x y) 
             end.
             
Theorem FPEmul_correct : forall l e1 e2,  req (PEeval l (FPEmul e1 e2) ) (PEeval l (PEmul e1 e2)).
intros l e1 e2; case e1; simpl; auto; case e2; auto; 
try (
  ( (intros p1 p2 c0; generalize (morph_eq  CRmorph c0 cI);  case (ceqb c0 cI); simpl; auto)  ||
    (intros c0 p1 p2; generalize (morph_eq  CRmorph c0 cI);  case (ceqb c0 cI); simpl; auto) ||
    (intros c c0; generalize (morph_eq  CRmorph c0 cI);  case (ceqb c0 cI); simpl; auto) ||
    (intros c c0; generalize (morph_eq  CRmorph c cI);  case (ceqb c cI); simpl; auto));
    intros H; rewrite (H (refl_equal true)); rewrite (morph1 CRmorph); apply eq_sym; auto).
(* Ugly!! *)	
apply eq_trans with (1*nth 0 c0 l);auto.
apply eq_trans with (1*((PEeval l p1 + PEeval l p2)));auto.
apply eq_trans with (1*((PEeval l p1 - PEeval l p2)));auto.
apply eq_trans with (1*((PEeval l p1 * PEeval l p2)));auto.
apply eq_trans with (1*(-(PEeval l c0)));auto.
Qed.

Section Fcons.


Variable Fcons : PExpr -> list PExpr -> list PExpr.


Fixpoint Fapp (l m:list PExpr) {struct l} : list PExpr :=
  match l with
  | nil => m
  | (cons a l1) => Fcons a (Fapp l1 m)
  end.


Definition Fnorm: FExpr ->  linear.
fix 1; intros e; case e.
intros c; exact (mk_linear (PEc c) (PEc cI) (@nil PExpr)).
intros x; exact (mk_linear (PEX x) (PEc cI) (@nil PExpr)).
intros e1 e2; pose (x:= Fnorm e1); pose (y:= Fnorm e2).
exact
 (mk_linear
   (PEadd (FPEmul (num x) (denum y)) (FPEmul (num y) (denum x)))
   (FPEmul (denum x) (denum y))
   (Fcons (FPEmul (denum x) (denum y)) (Fapp (condition x) (condition y)))).
intros e1 e2; pose (x:= Fnorm e1); pose (y:= Fnorm e2).
exact
 (mk_linear
   (PEsub (FPEmul (num x) (denum y)) (FPEmul (num y) (denum x)))
   (FPEmul (denum x) (denum y))
   (Fcons (FPEmul (denum x) (denum y)) (Fapp (condition x) (condition y)))).
intros e1 e2; pose (x:= Fnorm e1); pose (y:= Fnorm e2).
exact
 (mk_linear
   (FPEmul (num x) (num y)) (FPEmul (denum x) (denum y))
   (Fcons (FPEmul (denum x) (denum y)) (Fapp (condition x) (condition y)))).
intros e1; pose (x:= Fnorm e1).
exact (mk_linear (PEopp (num x)) (denum x) (condition x)).
intros e1; pose (x:= Fnorm e1).
exact (mk_linear (denum x) (num x) (Fcons (num x) (condition x))).
intros e1 e2; pose (x:= Fnorm e1); pose (y:= Fnorm e2).
exact
 (mk_linear
   (FPEmul (num x) (denum y)) (FPEmul (denum x) (num y))
   (Fcons (FPEmul (denum x) (num y)) (Fapp (condition x) (condition y)))).
Defined.


(* Example *)

Eval compute
   in (Fnorm
        (FEdiv
          (FEc cI)
          (FEadd (FEinv (FEX xH%positive)) (FEinv (FEX (xO xH)%positive))))).


(***************************************************************************
                                                                          
                               Evaluation                                 
                                                                          
  ***************************************************************************)
 
Definition FEeval: forall (l : list R) (pe : FExpr),  R.
fix 2; intros l pe.
case pe.
intros c; exact (phi c).
intros x; exact (nth 0 x l).
intros x y; exact (radd (FEeval l x) (FEeval l y)).
intros x y; exact (rsub (FEeval l x) (FEeval l y)).
intros x y; exact (rmul (FEeval l x) (FEeval l y)).
intros x; exact (ropp (FEeval l x)).
intros x; exact (rinv (FEeval l x)).
intros x y; exact (rdiv (FEeval l x) (FEeval l y)).
Defined.


(***************************************************************************
                                                                          
                       Properties of side condition                       
                                                                          
  ***************************************************************************)
 
Definition PCond: list R -> list PExpr ->  Prop.
fix 2; intros l le; case le.
exact True.
intros e1 l1; pose (x:= PCond l l1); case l1.
exact (~ req (PEeval l e1) rO).
intros e2 l2; exact (~ req (PEeval l e1) rO /\ x).
Defined.

 
Theorem PCond_def:
 forall l le,  (PCond l le <-> (forall e, In e le ->  ~ req (PEeval l e) rO)).
intros l le; elim le; simpl; auto.
split; intros H1; auto.
intros a l0; case l0.
intros [H1 H2].
split; auto.
intros H e [H3|H4]; (try subst e); auto.
intros p l1 [H1 H2].
split; auto.
intros [H H'] e [H3|H4]; (try subst e); auto.
Qed.

 
Theorem PCond_correct1:
 forall l le e, PCond l le -> In e le ->  ~ req (PEeval l e) rO.
intros l le e H1 H2; case (PCond_def l le); auto.
Qed.

Theorem PCond_correct2:
 forall l le, (forall e, In e le ->  ~ req (PEeval l e) rO) ->  PCond l le.
intros l le H; case (PCond_def l le); auto.
Qed.

Hypothesis
   PCond_fcons_inv :
   forall l a l1, PCond l (Fcons a l1) ->  ~ req (PEeval l a) rO /\ PCond l l1.
 

Theorem
   PCond_fcons_inv_l :
   forall l a l1, PCond l (Fcons a l1) ->  ~ req (PEeval l a) rO.
intros l a l1 H; case (PCond_fcons_inv l a l1); auto.
Qed.
 
Theorem PCond_fcons_inv_r : forall l a l1, PCond l (Fcons a l1) ->  PCond l l1.
intros l a l1 H; case (PCond_fcons_inv l a l1); auto.
Qed.
 
Theorem PCond_cons_inv_l:
 forall l a l1, PCond l (cons a l1) ->  ~ req (PEeval l a) rO.
intros l a l1 H; apply PCond_correct1 with (cons a l1); simpl; auto.
Qed.
 
Theorem PCond_cons_inv_r: forall l a l1, PCond l (cons a l1) ->  PCond l l1.
intros l a l1 H; apply PCond_correct2; intros e H1;
 apply PCond_correct1 with (cons a l1); red; auto.
Qed.

Theorem PCond_fapp_inv_l: forall l l1 l2, PCond l (Fapp l1 l2) ->  PCond l l1.
intros l l1 l2; elim l1; simpl Fapp.
simpl; auto.
intros a l0 H H0; apply PCond_correct2.
intros e H1; simpl in H1 |-; case H1.
intros H2; subst e; simpl in H0 |-; apply PCond_fcons_inv_l with ( 1 := H0 ).
intros H2; apply (PCond_correct1 l l0); auto; apply H;
 apply PCond_fcons_inv_r with ( 1 := H0 ).
Qed.
 
Theorem PCond_fapp_inv_r: forall l l1 l2, PCond l (Fapp l1 l2) ->  PCond l l2.
intros l l1 l2; elim l1; simpl Fapp; auto.
intros a l0 H H0; apply H; apply PCond_fcons_inv_r with ( 1 := H0 ).
Qed.
 
Theorem Pcond_Fnorm:
 forall l e,
 PCond l (condition (Fnorm e)) ->  ~ req (PEeval l (denum (Fnorm e))) rO.
intros l e; elim e; auto;
 (simpl; intros c H1 H2; case rI_neq_rO; rewrite <- CRmorph.(morph1); auto; fail)
   ||
    (intros e1 He1 e2 He2 H; apply (PCond_fcons_inv_l _ _ _ H); simpl; auto;
      fail)
      || simpl; intros e1 He1 H.
apply (PCond_fcons_inv_l _ _ _ H); simpl; auto.
Qed.
Hint Resolve Pcond_Fnorm .

(* Weird hack *)
Opaque PCond.


(***************************************************************************
                                                                          
                       Main theorem                                       
                                                                          
  ***************************************************************************)

Theorem Fnorm_FEeval_PEeval:
 forall l fe,
 PCond l (condition (Fnorm fe)) ->
  req
   (FEeval l fe)
   (rdiv (PEeval l (num (Fnorm fe))) (PEeval l (denum (Fnorm fe)))).
intros l fe; elim fe; simpl.
intros c H; rewrite CRmorph.(morph1); apply rdiv1.
intros p H; rewrite CRmorph.(morph1); apply rdiv1.
intros e1 He1 e2 He2 HH.
assert (HH1: PCond l (condition (Fnorm e1))).
apply PCond_fapp_inv_l with (condition (Fnorm e2)).
apply PCond_fcons_inv_r with ( 1 := HH ).
assert (HH2: PCond l (condition (Fnorm e2))).
apply PCond_fapp_inv_r with (condition (Fnorm e1)).
apply PCond_fcons_inv_r with ( 1 := HH ).
rewrite (He1 HH1); rewrite (He2 HH2).
rewrite (FPEmul_correct l  (num (Fnorm e1))  (denum (Fnorm e2))).
rewrite (FPEmul_correct l  (num (Fnorm e2))  (denum (Fnorm e1))).
rewrite (FPEmul_correct l  (denum (Fnorm e1))  (denum (Fnorm e2))).
simpl; apply rdiv2; auto.
intros Hk; absurd (req (PEeval l (FPEmul (denum (Fnorm e1)) (denum (Fnorm e2)))) rO).
apply PCond_fcons_inv_l with ( 1 := HH ).
rewrite (FPEmul_correct l  (denum (Fnorm e1))  (denum (Fnorm e2))); auto.

intros e1 He1 e2 He2 HH.
assert (HH1: PCond l (condition (Fnorm e1))).
apply PCond_fapp_inv_l with (condition (Fnorm e2)).
apply PCond_fcons_inv_r with ( 1 := HH ).
assert (HH2: PCond l (condition (Fnorm e2))).
apply PCond_fapp_inv_r with (condition (Fnorm e1)).
apply PCond_fcons_inv_r with ( 1 := HH ).
rewrite (He1 HH1); rewrite (He2 HH2).
rewrite (FPEmul_correct l  (num (Fnorm e1))  (denum (Fnorm e2))).
rewrite (FPEmul_correct l  (num (Fnorm e2))  (denum (Fnorm e1))).
rewrite (FPEmul_correct l  (denum (Fnorm e1))  (denum (Fnorm e2))).
simpl; apply rdiv3; auto.
intros Hk; absurd (req (PEeval l (FPEmul (denum (Fnorm e1)) (denum (Fnorm e2)))) rO).
apply PCond_fcons_inv_l with ( 1 := HH ).
rewrite (FPEmul_correct l  (denum (Fnorm e1))  (denum (Fnorm e2))); auto.

intros e1 He1 e2 He2 HH.
assert (HH1: PCond l (condition (Fnorm e1))).
apply PCond_fapp_inv_l with (condition (Fnorm e2)).
apply PCond_fcons_inv_r with ( 1 := HH ).
assert (HH2: PCond l (condition (Fnorm e2))).
apply PCond_fapp_inv_r with (condition (Fnorm e1)).
apply PCond_fcons_inv_r with ( 1 := HH ).
rewrite (He1 HH1); rewrite (He2 HH2).
rewrite (FPEmul_correct l  (num (Fnorm e1))  (num (Fnorm e2))).
rewrite (FPEmul_correct l  (denum (Fnorm e1))  (denum (Fnorm e2))).
simpl; apply rdiv4; auto.
intros Hk; absurd (req (PEeval l (FPEmul (denum (Fnorm e1)) (denum (Fnorm e2)))) rO).
apply PCond_fcons_inv_l with ( 1 := HH ).
rewrite (FPEmul_correct l  (denum (Fnorm e1))  (denum (Fnorm e2))); auto.

intros e1 He1 HH.
rewrite (He1 HH); apply rdiv5; auto.

intros e1 He1 HH.
assert (HH1: PCond l (condition (Fnorm e1))).
apply PCond_fcons_inv_r with ( 1 := HH ).
rewrite (He1 HH1); apply rdiv6; auto.
apply PCond_fcons_inv_l with ( 1 := HH ); auto.

intros e1 He1 e2 He2 HH.
assert (HH1: PCond l (condition (Fnorm e1))).
apply PCond_fapp_inv_l with (condition (Fnorm e2)).
apply PCond_fcons_inv_r with ( 1 := HH ).
assert (HH2: PCond l (condition (Fnorm e2))).
apply PCond_fapp_inv_r with (condition (Fnorm e1)).
apply PCond_fcons_inv_r with ( 1 := HH ).
rewrite (He1 HH1); rewrite (He2 HH2).
rewrite (FPEmul_correct l  (num (Fnorm e1))  (denum (Fnorm e2))).
rewrite (FPEmul_correct l  (denum (Fnorm e1))  (num (Fnorm e2))).
simpl; apply rdiv7; auto.
intros Hk; absurd (req (PEeval l (FPEmul (denum (Fnorm e1)) (num (Fnorm e2)))) rO).
apply PCond_fcons_inv_l with ( 1 := HH ).
rewrite (FPEmul_correct l  (denum (Fnorm e1))  (num (Fnorm e2))); auto.
Qed.
 
Theorem Fnorm_ok:
 forall l fe,
 PCond l (condition (Fnorm fe)) ->
  req
   (FEeval l fe)
   (rdiv (Pphi l (norm (num (Fnorm fe)))) (Pphi l (norm (denum (Fnorm fe))))).
intros l fe H;
 apply eq_trans with ( 1 := Fnorm_FEeval_PEeval l fe H ).
apply SRdiv_ext; apply norm_ok; auto.
Qed.

Theorem Fnorm_correct:
 forall l fe,
 Peq (norm (num (Fnorm fe))) (Pc cO) = true ->
 PCond l (condition (Fnorm fe)) ->  req (FEeval l fe) rO.
intros l fe H H1;
 apply eq_trans  with ( 1 := Fnorm_FEeval_PEeval l fe H1 ).
apply rdiv8; auto.
apply eq_trans with (PEeval l (PEc cO)); auto.
apply ring_correct; auto.
simpl; apply (morph0 CRmorph); auto.
Qed.



Theorem Pphi_dev_div_ok:
 forall l fe,
 PCond l (condition (Fnorm fe)) ->
  req
   (FEeval l fe)
   (rdiv (Pphi_dev l (norm (num (Fnorm fe))))
         (Pphi_dev l (norm (denum (Fnorm fe))))).
Proof.
  intros l fe H.
  apply eq_trans with (1 := Fnorm_FEeval_PEeval _ _ H).
  apply SRdiv_ext; apply Pphi_dev_ok.
Qed.
			   

End Fcons.

(* Weird hack *)
Transparent PCond.

 Ltac Make_field_rewrite1 add mul div cons Pcons fv r fe :=
  let Heq := fresh "Heq" in
  let x := fresh "x" in 
  let y := fresh "y" in 
  (assert (Heq := Pphi_dev_div_ok cons Pcons fv fe);
   set (x := (norm (num (Fnorm cons fe)))) in Heq;
   compute in x;
   let npx := eval red in x in
   let nrx := TPphi_dev add mul fv npx in
   (set (y := (norm (denum (Fnorm cons fe)))) in Heq;
   compute in y;
   let npy := eval red in y in
   let nry := TPphi_dev add mul fv npy in
   (change     
      (PCond fv (condition (Fnorm cons fe)) ->
       (req r (rdiv nrx nry))) in Heq; 
    (simpl PCond in Heq); try rewrite Heq; clear Heq x y))).

 Ltac Make_field_rewrite add mul sub opp inv div cons Pcons Cst_tac r :=
  let rfv := FV1 Cst_tac add mul sub opp inv div r (nil R) in
  let fv := Trev rfv in
  let fe := mkFieldexpr Cst_tac add mul sub opp inv div r fv in
  Make_field_rewrite1 add mul div cons Pcons fv r fe.


Ltac Make_field_tac add mul sub opp inv div req cons Pcons Cst_tac :=
  match goal with
  | |- ?r1 = ?r2 =>
    let rfv := FV1 Cst_tac add mul sub opp inv div r1 (nil R) in
    let fv := Trev rfv in
    let pe1 := mkFieldexpr Cst_tac add mul sub opp inv div r1 fv in
    let H := fresh "h" in 
    let pe := fresh "pe" in 
    let re := fresh "pe" in 
       ( 
         generalize (Fnorm_correct cons Pcons fv pe1);
         set (pe := Fnorm cons pe1);
         compute in pe;  
         set (re := (norm (num pe)));
         compute in re;
	 intros H;
         apply H ; [(apply (refl_equal true) || 
                      fail "Can't prove the equality") |
                    unfold pe; lazy beta zeta iota delta
                        [PCond  condition 
                         Rring.PEeval insideRing.rO
                         BinList.nth BinList.hd rO
                         BinList.jump BinList.tl]; 
                    try (apply I); try (clear H pe re)]
        )
  | _ => 
     fail  "not equality"
  end.



End MakeFieldPol.

