Set Implicit Arguments.
Require Import Pol.

 
(*** Intanciation for Z*)
Require Import ZArith.
Open Scope Z_scope.
		    
Module ZCE.
  Definition R := Z.
  Definition rO := 0.
  Definition rI := 1.
  Definition radd := Zplus.
  Definition rmul := Zmult.
  Definition rsub := Zminus.
  Definition ropp := Zopp.
  Definition Rth := Zth.
  Definition reqb := Zeq_bool.
  Definition req_morph := Zeqb_ok.
End ZCE.

Module ZEntries := CE_Entries ZCE.

Module Zring := MakeRingPol ZEntries.
		  
Ltac is_PCst p :=
 match p with
 | xH => constr:RBtrue
 | (xO ?p') => is_PCst p'
 | (xI ?p') => is_PCst p'
 | _ => constr:RBfalse
 end.

Ltac ZCst t :=
 match t with
 | Z0 => constr:t 
 | (Zpos ?p) =>
    match (is_PCst p) with
    | RBfalse => constr:RBfalse
    | _ => constr:t
    end
 | (Zneg ?p) =>
    match (is_PCst p) with
    | RBfalse => constr:RBfalse
    | _ => constr:t
    end
 | _ => constr:RBfalse
 end.
		  
Ltac zring := 
  let T := Zring.Make_ring_tac in
  T Zplus Zmult Zminus Zopp (@eq Z) ZCst.

Ltac zrewrite :=
  let T := Zring.Make_ring_rewrite in
  T Zplus Zmult Zminus Zopp ZCst.
	      
(* Instanciation for Bool *)	      
Require Import Bool.
	      
Module BCE.
 Definition R := bool.
 Definition rO := false.
 Definition rI := true.
 Definition radd := xorb.
 Definition rmul := andb.
 Definition rsub := xorb.
 Definition ropp b:bool := b.
 Lemma Rth : ring_theory rO rI radd rmul rsub ropp (@eq bool).
 Proof.
  constructor.
   exact false_xorb.
   exact xorb_comm.
   intros; symmetry  in |- *; apply xorb_assoc.
   exact andb_true_b.
   exact andb_comm.
   exact andb_assoc.
   destruct x; destruct y; destruct z; reflexivity.
   intros; reflexivity.
   exact xorb_nilpotent.
 Qed.

 Definition reqb := eqb.
 Definition req_morph := eqb_prop.
End BCE.

Module BEntries := CE_Entries BCE.

Module Bring := MakeRingPol BEntries.

Ltac BCst t :=
  match t with
   | true => constr:true
   | false => constr:false
   | _ => constr:RBfalse
   end.

Ltac bring := 
  let T := Bring.Make_ring_tac in
  T xorb andb xorb (fun b:bool => b) (@eq bool) BCst.

Ltac brewrite :=
  let T := Zring.Make_ring_rewrite in
  T Zplus Zmult Zminus Zopp ZCst.

(* Instanciation for R *)	      
Require Import Rbase.
Open Scope R_scope.	      

Module RAE.
 Definition R := R.
 Definition rO := 0.
 Definition rI := 1.
 Definition radd := Rplus.
 Definition rmul := Rmult.
 Definition rsub := Rminus.
 Definition ropp := Ropp.
 Lemma Rth : ring_theory 0 1 Rplus Rmult Rminus Ropp (@eq R).
 Proof.
  constructor. exact Rplus_0_l. exact Rplus_comm. 
  intros;symmetry;apply Rplus_assoc.
  exact Rmult_1_l. exact Rmult_comm. 
  intros;symmetry;apply Rmult_assoc.
  exact Rmult_plus_distr_r. trivial. exact Rplus_opp_r.
 Qed.
End RAE.

Module REntries := AE_Entries RAE.

Module Rring := MakeRingPol REntries.

Ltac genCstZ rO rI t :=
  match t with
   | rO => constr:(Z0)
   | rI => constr:(Zpos xH)
   | _ => constr:RBfalse
   end.

Ltac RCst := genCstZ 0 1.
		       
Ltac rring :=
  let T := Rring.Make_ring_tac in
  T Rplus Rmult Rminus Ropp (@eq R) RCst.

Ltac rrewrite :=
  let T := Rring.Make_ring_rewrite in
  T Rplus Rmult Rminus Ropp RCst.

Ltac rtest := 
  let T := Rring.Make_test in
  T Rplus Rmult Rminus Ropp RCst.	     
(* Instanciation for N *)	     
Require Import NArith.
Open Scope N_scope.
		
Module NCSE.
 Definition R := N.
 Definition rO := 0.
 Definition rI := 1.
 Definition radd := Nplus.
 Definition rmul := Nmult.
 Definition SRth := Nth.
 Definition reqb := Neq_bool.
 Definition req_morph := Neq_bool_ok.
End NCSE.

Module NEntries := CSE_Entries NCSE.

Module Nring := MakeRingPol NEntries.

Ltac NCst t :=
 match t with
 | N0 => constr:t 
 | (Npos ?p) =>
    match (is_PCst p) with
    | RBfalse => constr:RBfalse
    | _ => constr:t
    end
 | _ => constr:RBfalse
 end.

Ltac nring :=  
  let T := Nring.Make_ring_tac in
  T Nplus Nmult (@SRsub N Nplus) (@SRopp N) (@eq N) NCst.		

Ltac nrewrite :=
  let T := Nring.Make_ring_rewrite in
  T Nplus Nmult  (@SRsub N Nplus) (@SRopp N) NCst.
	       
(* Instanciation for nat *)	       
Open Scope nat_scope.

Module NatASE.
 Definition R := nat.
 Definition rO := 0.
 Definition rI := 1.
 Definition radd := plus.
 Definition rmul := mult.
 Lemma SRth : semi_ring_theory O (S O) plus mult (@eq nat).
 Proof.
  constructor. exact plus_0_l. exact plus_comm. exact plus_assoc.
  exact mult_1_l. exact mult_0_l. exact mult_comm. exact mult_assoc.
  exact mult_plus_distr_r. 
 Qed.
End NatASE.

Module NatEntries := ASE_Entries NatASE.

Module Natring := MakeRingPol NatEntries.

Ltac natCst t := 
 match t with
 | O => constr:(N0)
 | (S ?n) =>
    match (natCst n) with
    | RBfalse => constr:(RBfalse)
    | ?p => constr:(Nsucc p)
    end
 | _ => constr:(RBfalse)
 end.
		  
Ltac natring :=
 let T := Natring.Make_ring_tac in
 T plus mult (@SRsub nat plus) (@SRopp nat) (@eq nat) natCst.

Ltac natrewrite :=
  let T := Natring.Make_ring_rewrite in
  T plus mult (@SRsub nat plus) (@SRopp nat) natCst.
	  
(* Generic tactic, checks the type of the terms and applies the
suitable instanciation*)
		  
Ltac newring :=  
 match goal with
 | |- (?r1 = ?r2) =>
   match (type of r1) with
   | Z => zring
   | R => rring
   | bool => bring
   | N => nring
   | nat => natring
   end
 end.




