Require Export realbasis.
Require Export proof_utils.
Require Export real_facts.
Require Export add_correct.
Require Export sub_correct.

Ltac interval_from_represents H :=
 (generalize (R_interval _ _ H) || 
  generalize (L_interval _ _ H) ||
  generalize (C_interval _ _ H) ||
  generalize (represents_interval _ _ H)).

Ltac intervals_from_represents H H0 :=
  interval_from_represents H; interval_from_represents H0; intros; dec_tac.

Ltac interval x :=
  match goal with 
    id : represents (_::_::x) _ |- _ => inversion id
    | _ => idtac
  end;
  match goal with 
    id : represents (_::x) _ |- _ => interval_from_represents id
    | _ => idtac
  end.

Ltac find_represents x :=
  match goal with 
    id : represents (_::_::x) _ |- _ => inversion id
    | _ => idtac
  end;
  match goal with 
    id : represents (_::x) _ |- _ => inversion id
    | _ => idtac
  end.

Lemma mult_correct_aux_lemma :
  forall x y, 
    (-1 <= x <= 1)%R ->
    (-1 / 4 <= y <= 1 / 4)%R -> 
    (-1 / 4 <= x * y <= 1 / 4)%R.
Proof.
intros x y [Hxge Hxle] [Hyge Hyle].
elim (Rlt_le_dec 0 x); elim (Rlt_le_dec 0 y).

intros Hgt1 Hgt2.
split.
apply Rle_trans with 0%R.
fourier.
apply Rlt_le.
apply Rmult_lt_0_compat; fourier.
replace (1 / 4)%R  with (1*(1/4))%R; [idtac| newfield].
apply Rmult_le_compat; fourier.

intros Hle Hgt.
split.
apply Ropp_le_cancel.
replace (- (x * y))%R with (x * (-y))%R; [idtac| newfield].
replace (- (-1 / 4))%R with (1*(1/4))%R; [idtac| newfield].
apply Rmult_le_compat; fourier.
apply Rle_trans with 0%R.
replace 0%R with (x * 0)%R; [idtac| newfield].
apply Rmult_le_compat_l; fourier.
fourier.

intros Hgt Hle.
split.
apply Ropp_le_cancel.
replace (- (x * y))%R with ((-x) * y)%R; [idtac| newfield].
replace (- (-1 / 4))%R with (1*(1/4))%R; [idtac| newfield].
apply Rmult_le_compat; fourier.
apply Rle_trans with 0%R.
replace 0%R with (0*y)%R; [idtac| newfield].
apply Rmult_le_compat_r; fourier.
fourier.

intros Hle1 Hle2.
replace (x * y)%R with ((-x) * (-y))%R; [idtac| newfield].
split.
apply Rle_trans with 0%R.
fourier.
replace 0%R with (0*0)%R; [idtac| newfield].
apply Rmult_le_compat; fourier.
replace (1 / 4)%R  with (1*(1/4))%R; [idtac| newfield].
apply Rmult_le_compat; fourier.
Qed.

Theorem mult_correct_body :
  forall mult_a' : stream idigit -> stream idigit*stream idigit -> stream idigit,
    (forall v1 v2 v3 x u u', 
      (-1/4 <= v3 <= 1/4)%R ->
      (-1 <= v1 +v2*v3 <= 1)%R ->
      represents x v1 -> 
      represents u v2 -> 
      represents u' v3 ->
      represents (mult_a' x (u, u')) (v1+v2*v3)%R) ->
    forall v1 v2 v3 x u u', 
      (-1/4 <= v3 <= 1/4)%R ->
      (-1 <= v1 +v2*v3 <= 1)%R ->
      represents x v1 -> 
      represents u (2*v2) -> 
      represents u' v3 ->
      represents (series_body _ mult_a' x (u, u')) (v1+v2*v3)%R.
Proof.
intros mult_a' mult_correct_body v1 v2 v3 x u u' Hv3lt [Hsum1 Hsum2] Hv1 Hv2 Hv3.
unfold series_body.
assert ((-1/4<=2*v2*v3 <= 1/4)%R).
clear mult_correct_body.
generalize (represents_interval _ _ Hv2).
intros.
apply mult_correct_aux_lemma; assumption.

destruct x as [ [ | | ] x1]; inversion Hv1.
(* case x = L.. *)

assert (-1<= r + 2 * v2 * v3)%R.
clear mult_correct_body.
subst v1.
assert (H0le2: (0<=2)%R).
fourier.
generalize (Rmult_le_compat_l 2 _ _ H0le2 Hsum1).
replace (2 * (r / 2 - 1 / 2 + v2 * v3))%R with 
  (r + 2*v2*v3 -1)%R; [intros; fourier|newfield].

destruct x1 as [ [ | | ] x2].
(* case x = L::L.. *)
match goal with id : represents (L::x2) _ |- _ => 
  intervals_from_represents Hv1 id
end.
assert ((-1 <= r + 2 * v2 * v3 <= 1)%R).
clear mult_correct_body.
split; [assumption |fourier].
constructor' mult_correct_body; 
(replace (2*(r/2-1/2+v2*v3)+1)%R with (r + 2*v2*v3)%R; [idtac|newfield]).
apply mult_correct_body; clear mult_correct_body; auto.
assumption.

(*case : x = L::R.. *)
destruct x2 as [ [ | | ] x3].

(*case : x = L::R::L.. *)
assert ((-1 <= r + 2 * v2 * v3 <= 1)%R).
clear mult_correct_body.
split; [assumption |idtac].
interval x3.
intros; dec_tac; fourier.
constructor' mult_correct_body;
(replace (2*(r/2-1/2+v2*v3)+1)%R with (r + 2*v2*v3)%R; [idtac|newfield]).
apply mult_correct_body; clear mult_correct_body; auto.
assumption.

(*case : x = L::R::R.. *)
assert (-1 <= r - 1 + 2 * v2 * v3 <= 1)%R.
clear mult_correct_body.
interval x3.
intros; dec_tac; split; fourier.
constructor' mult_correct_body;
(replace (2*(r/2-1/2+v2*v3))%R with (r -1 + 2*v2*v3)%R; [idtac|newfield]).
apply mult_correct_body; clear mult_correct_body; auto.
find_represents x3.
replace ((r1 / 2 + 1 / 2) / 2 + 1 / 2 - 1)%R with
  ((r1 / 2 + 1 / 2) / 2 - 1 / 2)%R; [idtac|newfield].
constructor.
constructor; assumption.
subst; assumption.
assumption.

(*case : x = L::R::C.. *)
assert ((-1 <= r + 2 * v2 * v3 <= 1)%R).
clear mult_correct_body.
split; [assumption |idtac].
interval x3.
intros; dec_tac; fourier.
constructor' mult_correct_body;
(replace (2*(r/2-1/2+v2*v3)+1)%R with (r + 2*v2*v3)%R; [idtac|newfield]).
apply mult_correct_body; clear mult_correct_body; auto.
assumption.

(*case : x = L::C.. *)
assert ((-1 <= r + 2 * v2 * v3 <= 1)%R).
clear mult_correct_body.
split; [assumption |idtac].
interval x2.
intros; dec_tac; fourier.
constructor' mult_correct_body;
(replace (2*(r/2-1/2+v2*v3)+1)%R with (r + 2*v2*v3)%R; [idtac|newfield]).
apply mult_correct_body; clear mult_correct_body; auto.
assumption.

(* case x = R.. *)

assert (r + 2 * v2 * v3<=1)%R.
clear mult_correct_body.
subst v1.
assert (H0le2: (0<=2)%R).
fourier.
generalize (Rmult_le_compat_l 2 _ _ H0le2 Hsum2).
replace (2 * (r / 2 + 1 / 2 + v2 * v3))%R with 
  (r + 2*v2*v3 +1)%R; [intros; fourier|newfield].

destruct x1 as [ [ | | ] x2].

(*case : x = R::L.. *)
destruct x2 as [ [ | | ] x3].

(*case : x = R::L::L.. *)
assert (-1 <= r + 1 + 2 * v2 * v3 <= 1)%R.
clear mult_correct_body.
interval x3.
intros; dec_tac; split; fourier.
constructor' mult_correct_body;
(replace (2*(r/2+1/2+v2*v3))%R with (r +1 + 2*v2*v3)%R; [idtac|newfield]).
apply mult_correct_body; clear mult_correct_body; auto.
find_represents x3.
replace ((r1 / 2 - 1 / 2) / 2 - 1 / 2 + 1)%R with
  ((r1 / 2 - 1 / 2) / 2 + 1 / 2)%R; [idtac|newfield].
constructor.
constructor; assumption.
subst; assumption.
assumption.

(*case : x = R::L::R.. *)
assert ((-1 <= r + 2 * v2 * v3 <= 1)%R).
clear mult_correct_body.
split; [idtac| assumption].
interval x3.
intros; dec_tac; fourier.
constructor' mult_correct_body;
(replace (2*(r/2+1/2+v2*v3)-1)%R with (r + 2*v2*v3)%R; [idtac|newfield]).
apply mult_correct_body; clear mult_correct_body; auto.
assumption.

(*case : x = R::L::C.. *)
assert ((-1 <= r + 2 * v2 * v3 <= 1)%R).
clear mult_correct_body.
split; [idtac|assumption].
interval x3.
intros; dec_tac; fourier.
constructor' mult_correct_body;
(replace (2*(r/2+1/2+v2*v3)-1)%R with (r + 2*v2*v3)%R; [idtac|newfield]).
apply mult_correct_body; clear mult_correct_body; auto.
assumption.

(* case x = R::R.. *)
inversion Hv1.
match goal with id : represents (R::x2) _ |- _ => 
  intervals_from_represents Hv1 id
end.
assert ((-1 <= r + 2 * v2 * v3 <= 1)%R).
clear mult_correct_body.
split; [fourier| assumption].
constructor' mult_correct_body; 
(replace (2*(r/2+1/2+v2*v3)-1)%R with (r + 2*v2*v3)%R; [idtac|newfield]).
apply mult_correct_body; clear mult_correct_body; auto.
assumption.

(*case : x = R::C.. *)
assert ((-1 <= r + 2 * v2 * v3 <= 1)%R).
clear mult_correct_body.
split; [idtac|assumption].
interval x2.
intros; dec_tac; fourier.
constructor' mult_correct_body;
(replace (2*(r/2+1/2+v2*v3)-1)%R with (r + 2*v2*v3)%R; [idtac|newfield]).
apply mult_correct_body; clear mult_correct_body; auto.
assumption.

(*case : x = C.. *)
destruct x1 as [ [ | | ] x2].

(* case x = C::L.. *)
destruct x2 as [ [ | | ] x3].

(* case x = C::L::L... *)
assert (-1 <= r + 1 + 2 * v2 * v3 <= 1)%R.
clear mult_correct_body.
interval x3.
intros; dec_tac; split; fourier.
constructor' mult_correct_body;
(replace (2*(r/2+v2*v3)+1)%R with ((r+1) + 2*v2*v3)%R; [idtac|newfield]).
apply mult_correct_body; clear mult_correct_body; auto.
find_represents x3.
replace ((r1 / 2 - 1 / 2) / 2 - 1 / 2 + 1)%R with
  ((r1 / 2 - 1 / 2) / 2 + 1 / 2)%R; [idtac|newfield].
constructor.
constructor; assumption.
subst; assumption.
assumption.

(* case x = C::L::R... *)
assert  (-1 <= r + 2 * v2 * v3 <= 1)%R.
clear mult_correct_body.
interval x3.
intros; dec_tac; split; fourier.
constructor' mult_correct_body;
(replace (2*(r/2+v2*v3))%R with (r + 2*v2*v3)%R; [idtac|newfield]).
apply mult_correct_body; clear mult_correct_body; auto.
assumption.

(* case x = C::L::C... *)
assert  (-1 <= r + 2 * v2 * v3 <= 1)%R.
clear mult_correct_body.
interval x3.
intros; dec_tac; split; fourier.
constructor' mult_correct_body;
(replace (2*(r/2+v2*v3))%R with (r + 2*v2*v3)%R; [idtac|newfield]).
apply mult_correct_body; clear mult_correct_body; auto.
assumption.

(* case x = C::R.. *)
destruct x2 as [ [ | | ] x3].

(* case x = C::R::L... *)
assert  (-1 <= r + 2 * v2 * v3 <= 1)%R.
clear mult_correct_body.
interval x3.
intros; dec_tac; split; fourier.
constructor' mult_correct_body;
(replace (2*(r/2+v2*v3))%R with (r + 2*v2*v3)%R; [idtac|newfield]).
apply mult_correct_body; clear mult_correct_body; auto.
assumption.

(* case x = C::R::R... *)
assert (-1 <= r - 1 + 2 * v2 * v3 <= 1)%R.
clear mult_correct_body.
interval x3.
intros; dec_tac; split; fourier.
constructor' mult_correct_body;
(replace (2*(r/2+v2*v3)-1)%R with ((r-1) + 2*v2*v3)%R; [idtac|newfield]).
apply mult_correct_body; clear mult_correct_body; auto.
find_represents x3.
replace ((r1 / 2 + 1 / 2) / 2 + 1 / 2 - 1)%R with
  ((r1 / 2 + 1 / 2) / 2 - 1 / 2)%R; [idtac|newfield].
constructor.
constructor; assumption.
subst; assumption.
assumption.

(* case x = C::R::C... *)
assert  (-1 <= r + 2 * v2 * v3 <= 1)%R.
clear mult_correct_body.
interval x3.
intros; dec_tac; split; fourier.
constructor' mult_correct_body;
(replace (2*(r/2+v2*v3))%R with (r + 2*v2*v3)%R; [idtac|newfield]).
apply mult_correct_body; clear mult_correct_body; auto.
assumption.

(* case x = C::C... *)
assert  (-1 <= r + 2 * v2 * v3 <= 1)%R.
clear mult_correct_body.
match goal with 
  id : represents (C :: x2) _ |- _ => generalize (C_interval _ _ id)
end.
intros; dec_tac; split; fourier.
constructor' mult_correct_body;
(replace (2*(r/2+v2*v3))%R with (r + 2*v2*v3)%R; [idtac|newfield]).
apply mult_correct_body; clear mult_correct_body; auto.
assumption.
Defined.

Ltac force_represents :=
 match goal with
  id: represents ?x ?y |- represents ?x ?z =>
  replace z with y;[assumption | field; auto with real1]
 end.

Theorem mult_a_correct :
  forall v1 v2 v3 x u u', 
    (-1/4 <= v3 <= 1/4)%R ->
    (-1 <= v1 +v2*v3 <= 1)%R ->
    represents x v1 ->
    represents u v2 -> 
    represents u' v3 ->
    represents (mult_a x (u, u')) (v1+v2*v3)%R.
cofix.
intros v1 v2 v3 x u u' Hv3lt Hsum Hv1 Hv2 Hv3.
rewrite (str_dec_thm (mult_a x (u, u'))).
lazy beta iota zeta delta [str_decompose mult_a]; fold mult_a.

destruct u as [ [ | | ] u1].

(*case L*)
destruct u1 as [ [ | | ] u2].

(*case L::L*)
fold (str_decompose (series_body (stream idigit * stream idigit) mult_a (x - (C:: u')) (L::u2, u'))).
rewrite <- str_dec_thm.
inversion Hv2.
replace (v1 + (r / 2 - 1 / 2) * v3)%R with ((v1-v3/2)+r/2*v3)%R; 
    [idtac | clear mult_a_correct; field; auto with real1].
apply mult_correct_body; (exact mult_a_correct || clear mult_a_correct); auto.
replace ((v1-v3/2)+r/2*v3)%R with (v1 + (r / 2 - 1 / 2) * v3)%R; 
    [rewrite H2; assumption | newfield].
apply sub_correct.
assumption.
constructor.
assumption.
dec_tac; split; fourier.
generalize (represents_interval _ _ Hv1).
generalize (L_interval _ _ H0).
intros ; dec_tac.
assert (r = 2*v2+1)%R;[subst v2; newfield|subst r].
elim (Rlt_le_dec 0 v3); intros; split.

apply Rle_trans with (1:=H1).
unfold Rminus.
apply Rplus_le_compat_l.
replace (- (v3 / 2))%R with ((-1/2)*v3)%R; [idtac|newfield].
apply Rmult_le_compat_r.
fourier.
fourier.
fourier.

fourier.
apply Rle_trans with (2:=H9).
unfold Rminus.
apply Rplus_le_compat_l.
replace (- (v3 / 2))%R with ((1/2)*(-v3))%R; [idtac|newfield].
replace (v2*v3)%R with (-v2*(-v3))%R; [idtac|newfield].
apply Rmult_le_compat_r.
fourier.
fourier.
force_represents.

(*case L::R*)
fold (str_decompose (series_body (stream idigit * stream idigit) mult_a x (L::u2, u'))).
rewrite <- str_dec_thm.
find_represents u2.
replace ((r0 / 2 + 1 / 2) / 2 - 1 / 2)%R with 
  ((r0/2 - 1/2)/2)%R; [idtac|newfield].
apply mult_correct_body; (exact mult_a_correct || clear mult_a_correct); auto.
replace ((r0 / 2 - 1 / 2) / 2)%R with v2; [assumption|subst; newfield].
replace (2 * ((r0 / 2 - 1 / 2) / 2))%R with 
  (r0 / 2 - 1 / 2)%R ; [idtac|newfield].
constructor; assumption.

(*case L::C*)
fold (str_decompose (series_body (stream idigit * stream idigit) mult_a (x - (C::C::u')) (L::u2, u'))).
rewrite <- str_dec_thm.
find_represents u2.
replace (v1 + (r0 / 2 / 2 - 1 / 2) * v3)%R with 
  ((v1-v3/2/2)+ ((r0/2-1/2)/2)*v3)%R; [idtac|newfield].
apply mult_correct_body; (exact mult_a_correct || clear mult_a_correct); auto.
replace (v1 - v3 / 2 / 2 + (r0 / 2 - 1 / 2) / 2 * v3)%R with 
  (v1+v2*v3)%R;[assumption|subst; newfield].
generalize (represents_interval _ _ Hv3); intros.
apply sub_correct.
assumption.
constructor.
constructor; assumption.
dec_tac; split; fourier.

generalize (represents_interval _ _ Hv1).
generalize (C_interval _ _ H0).
intros ; dec_tac.
assert (r = 2*v2+1)%R;[subst v2; newfield|subst r].
replace (v3/2/2)%R with (v3/4)%R; [idtac |newfield].
elim (Rlt_le_dec 0 v3); intros; split.

apply Rle_trans with (1:=H1).
unfold Rminus.
apply Rplus_le_compat_l.
replace (- (v3 / 4))%R with ((-1/4)*v3)%R; [idtac|newfield].
apply Rmult_le_compat_r.
fourier.
fourier.
fourier.

fourier.
apply Rle_trans with (2:=H16).
unfold Rminus.
apply Rplus_le_compat_l.
replace (- (v3 / 4))%R with ((1/4)*(-v3))%R; [idtac|newfield].
replace (v2*v3)%R with (-v2*(-v3))%R; [idtac|newfield].
apply Rmult_le_compat_r.
fourier.
fourier.
replace (2 * ((r0 / 2 - 1 / 2) / 2))%R with 
  (r0 / 2 - 1 / 2)%R ; [idtac|newfield].
constructor; assumption.

(*case R*)
destruct u1 as [ [ | | ] u2].

(*case R::L*)
fold (str_decompose (series_body (stream idigit * stream idigit) mult_a x (R::u2, u'))).
rewrite <- str_dec_thm.
find_represents u2.
replace ((r0 / 2 - 1 / 2) / 2 + 1 / 2)%R with 
  ((r0/2 + 1/2)/2)%R; [idtac|newfield].
apply mult_correct_body; (exact mult_a_correct || clear mult_a_correct); auto.
replace ((r0 / 2 + 1 / 2) / 2)%R with v2; [assumption|subst; newfield].
replace (2 * ((r0 / 2 + 1 / 2) / 2))%R with 
  (r0 / 2 + 1 / 2)%R ; [idtac|newfield].
constructor; assumption.

(*case R::R*)
fold (str_decompose (series_body (stream idigit * stream idigit) mult_a (x + (C:: u')) (R::u2, u'))).
rewrite <- str_dec_thm.
inversion Hv2.
replace (v1 + (r / 2 + 1 / 2) * v3)%R with ((v1+v3/2)+r/2*v3)%R; 
    [idtac | clear mult_a_correct; field; auto with real1].
apply mult_correct_body; (exact mult_a_correct || clear mult_a_correct); auto.
replace ((v1+v3/2)+r/2*v3)%R with (v1 + (r / 2 + 1 / 2) * v3)%R; 
    [rewrite H2; assumption | newfield].
apply add_correct.
assumption.
constructor.
assumption.
dec_tac; split; fourier.
generalize (represents_interval _ _ Hv1).
generalize (R_interval _ _ H0).
intros ; dec_tac.
assert (r = 2*v2-1)%R;[subst v2; newfield|subst r].
elim (Rlt_le_dec 0 v3); intros; split.

fourier.
apply Rle_trans with (2:=H9).
apply Rplus_le_compat_l.
replace (v3 / 2)%R with ((1/2)*v3)%R; [idtac|newfield].
apply Rmult_le_compat_r.
fourier.
fourier.

apply Rle_trans with (1:=H1).
apply Rplus_le_compat_l.
replace (v3 / 2)%R with ((-1/2)*(-v3))%R; [idtac|newfield].
replace (v2*v3)%R with (-v2*(-v3))%R; [idtac|newfield].
apply Rmult_le_compat_r.
fourier.
fourier.
fourier.
force_represents.

(*case R::C*)
fold (str_decompose (series_body (stream idigit * stream idigit) mult_a (x + (C::C::u')) (R::u2, u'))).
rewrite <- str_dec_thm.
find_represents u2.
replace (v1 + (r0 / 2 / 2 + 1 / 2) * v3)%R with 
  ((v1+v3/2/2)+ ((r0/2+1/2)/2)*v3)%R; [idtac|newfield].
apply mult_correct_body; (exact mult_a_correct || clear mult_a_correct); auto.
replace (v1 + v3 / 2 / 2 + (r0 / 2 + 1 / 2) / 2 * v3)%R with 
  (v1+v2*v3)%R;[assumption|subst; newfield].
generalize (represents_interval _ _ Hv3); intros.
apply add_correct.
assumption.
constructor.
constructor; assumption.
dec_tac; split; fourier.

generalize (represents_interval _ _ Hv1).
generalize (C_interval _ _ H0).
intros ; dec_tac.
assert (r = 2*v2-1)%R;[subst v2; newfield|subst r].
replace (v3/2/2)%R with (v3/4)%R; [idtac |newfield].
elim (Rlt_le_dec 0 v3); intros; split.

fourier.
apply Rle_trans with (2:=H16).
apply Rplus_le_compat_l.
replace (v3 / 4)%R with ((1/4)*v3)%R; [idtac|newfield].
apply Rmult_le_compat_r.
fourier.
fourier.

apply Rle_trans with (1:=H1).
apply Rplus_le_compat_l.
replace (v3 / 4)%R with ((-1/4)*(-v3))%R; [idtac|newfield].
replace (v2*v3)%R with (-v2*(-v3))%R; [idtac|newfield].
apply Rmult_le_compat_r.
fourier.
fourier.
fourier.
replace (2 * ((r0 / 2 + 1 / 2) / 2))%R with 
  (r0 / 2 + 1 / 2)%R ; [idtac|newfield].
constructor; assumption.

(*case C*)

fold (str_decompose (series_body (stream idigit * stream idigit) mult_a x (u1, u'))).
rewrite <- str_dec_thm.
inversion Hv2.
apply mult_correct_body; (exact mult_a_correct || clear mult_a_correct); auto.
subst; assumption.
force_represents.
Qed.

Theorem mult_correct :
  forall x y vx vy, 
    represents x vx -> 
    represents y vy ->
    represents (mult x y) (vx * vy).
intros x y vx vy Hx Hy.
replace (vx * vy)%R with (2*(2*(vx * (vy/2 /2))))%R; [idtac|newfield].
unfold mult.
interval_from_represents Hx; interval_from_represents Hy; intros.
assert (-1 / 4 <= vy / 2 / 2 <= 1 / 4)%R.
replace (vy/2/2)%R with (vy/4)%R;[idtac|newfield].
dec_tac; split; fourier.
assert (-1/4<= vx * (vy/2/2) <= 1/4)%R.
apply mult_correct_aux_lemma.
assumption.
assumption.
apply mult2_correct.
dec_tac; split; fourier.
apply mult2_correct.
dec_tac; split; fourier.
replace (vx * (vy / 2 / 2))%R with
  (0 + (vx * (vy / 2 / 2)))%R; [idtac|ring].
apply mult_a_correct.
assumption.
dec_tac; split; fourier.
apply represents_zero.
assumption.
constructor.
constructor; assumption.
dec_tac; split; fourier.
Qed.

