Require Export realbasis.
Require Export proof_utils.
Require Export Field_tac.

Definition str_decompose (A:Set)(x:stream A) :=
  match x with Cons a tl => Cons a tl end.

Implicit Arguments str_decompose.

Theorem str_dec_thm : forall A:Set, forall x : stream A,
  x = str_decompose x.
intros; case x; auto.
Qed.

Implicit Arguments str_dec_thm.

Ltac str_unfold term :=
  rewrite (str_dec_thm term).

Theorem R_interval :
  forall x y, represents (Cons R x) y -> (0 <= y <= 1)%R.
intros; inversion H; dec_tac; split; fourier.
Qed.

Theorem L_interval :
  forall x y, represents (Cons L x) y -> (-1 <= y <= 0)%R.
intros; inversion H ; dec_tac; split; fourier .
Qed.

Theorem C_interval :
  forall x y, represents (Cons C x) y -> (-1/2 <= y <= 1/2)%R.
intros; inversion H; dec_tac; split; fourier.

Qed.

Theorem represents_interval : 
  forall x r, represents x r -> (-1 <= r <= 1)%R.
intros x r H; destruct x as [ [ | | ] x']; 
 [generalize (L_interval _ _ H) | generalize (R_interval _ _ H)
   | generalize (C_interval _ _ H)];
 intros; dec_tac; split; fourier.
Qed.

Theorem represents_zero :
  represents zero 0%R.
cofix.
str_unfold zero.
simpl.
replace 0%R with (0/2)%R.
constructor.
assumption.
clear represents_zero.
split; fourier.
clear represents_zero.
newfield.
Qed.

Theorem represents_one : 
  represents one 1%R.
cofix.
str_unfold one.
simpl.
replace 1%R with (1/2+1/2)%R.
constructor.
assumption.
clear represents_one.
split; fourier.
clear represents_one.
newfield.
Qed.

Theorem represents_sub_one : 
  represents sub_one (-1)%R.
cofix.
str_unfold sub_one.
simpl.
replace (-1)%R with (-1/2-1/2)%R.
constructor.
assumption.
clear represents_sub_one.
split; fourier.
clear represents_sub_one.
newfield.
Qed.

Ltac rep_tac := 
  match goal with
    | id : represents ?x ?r |- represents ?x ?r => assumption
    | id : represents ?x ?r1 |- represents ?x ?r2 =>
	replace r1 with r2; [idtac|newfield]
    |id : represents ?x ?r1 |- represents (?d::?x) ?r2 => 
	replace r2 with 
	  match d with 
	    | L => (r1/2 -1/2)%R
	    | C => (r1/2)%R
	    | R => (r1/2 +1/2)%R
	  end; [constructor| newfield]
  end.

Ltac constructor' hyp :=
  match goal with
    |- represents (C::_) ?v =>
    replace v with ((2*v)/2)%R;
         [constructor; [idtac|clear hyp] | clear hyp; newfield]
  | |- represents (L::_) ?v =>
    replace v with ((2*v+1)/2-1/2)%R;
         [constructor; [idtac|clear hyp] | clear hyp; newfield]
  | |- represents (R::_) ?v  =>
    replace v with ((2*v-1)/2+1/2)%R;
         [constructor; [idtac|clear hyp] | clear hyp; newfield]
  end.

Ltac rat_to_str_tac hyp :=
  match goal with 
    |- represents (rat_to_stream ?q ?r) ?w =>
      replace w with (IZR q / IZR r)%R; [apply hyp; clear hyp| clear hyp]
end.

Ltac inversion_to_variable x :=
  match goal with
  | id : represents x _ |- _ => idtac
  | id : represents (Cons ?d x) _ |- context [Cons ?d x] => idtac
  | id : represents (Cons ?d1 x) _ |- context [Cons ?d2 x] => inversion id
  | id : represents (Cons ?d1 (Cons ?d2 x)) _ |- context [ Cons ?d3 x] =>
    inversion id; inversion_to_variable x
  end.

Lemma rat_to_stream_correct_aux1 : 
  forall a b,
    0<>b ->
    (-1 <= IZR a/ IZR b <= 1)%R -> 
    0<=a*b ->
    (-1 <= 2 * (IZR a / IZR b) - 1 <= 1)%R.
Proof.
intros a b Hbneq0 [Hdivge Hdivle] Hmultpos.
split; [idtac| fourier].
assert ( 0<= (IZR a / IZR b))%R; [idtac|fourier].
unfold Rdiv.
elim (Ztrichotomy_inf 0 b).
intros H1; elim H1; clear H1.
intros Hgt.
apply Rle_mult_inv_pos.
apply (IZR_le 0).
apply Zmult_le_0_reg_r with b; omega.
apply (IZR_lt 0 b Hgt).

intuition.

intros Hlt.
rewrite <- Rmult_opp_opp.
replace (-/IZR b)%R with (/(-IZR b))%R; [idtac| newfield].
apply Rle_mult_inv_pos.
rewrite <- Zmult_opp_opp in Hmultpos.
rewrite <- Ropp_Ropp_IZR.
apply (IZR_le 0).
apply Zmult_le_0_reg_r with (-b); omega.
rewrite <- Ropp_Ropp_IZR.
apply (IZR_lt 0); omega.
split.
apply (IZR_neq b 0); omega.
rewrite <- Ropp_Ropp_IZR.
apply (IZR_neq (-b) 0); omega.
Qed.

Lemma rat_to_stream_correct_aux2 : 
  forall a b,
    0<>b ->
    (-1 <= IZR a/ IZR b <= 1)%R -> 
    0 > a * b ->
    (-1 <= 2 * (IZR a / IZR b) + 1 <= 1)%R.
Proof.
intros a b Hbneq0 [Hdivge Hdivle] Hmultneg.
split; [fourier|idtac].
assert ( 0< -(IZR a / IZR b))%R; [idtac|fourier].
unfold Rdiv.
elim (Ztrichotomy_inf 0 b).
replace (- (IZR a * / IZR b))%R with 
  (IZR (- a) * / IZR b)%R.
intros H1; elim H1; clear H1.
intros Hgt.
apply Rlt_mult_inv_pos.
apply (IZR_lt 0).
apply Zmult_lt_0_reg_r with b. 
assumption.
replace (- a * b) with (-(a*b)); [idtac|ring].
omega.
apply (IZR_lt 0 b Hgt).

intuition.

field.
apply Ropp_Ropp_IZR.
apply (IZR_neq b 0); omega.

intros Hlt.
rewrite <- Rmult_opp_opp.
replace (-(- IZR a * -/IZR b))%R with (IZR a * /(-IZR b))%R; [idtac| newfield].
apply Rlt_mult_inv_pos.
apply (IZR_lt 0).
apply Zmult_lt_0_reg_r with (-b).
omega.
replace (a*-b) with (-(a*b)); [omega|ring].
rewrite <- Ropp_Ropp_IZR.
apply (IZR_lt 0).
omega.
split.
apply (IZR_neq b 0); omega.
rewrite <- Ropp_Ropp_IZR.
apply (IZR_neq (-b) 0); omega.
Qed.

Theorem rat_to_stream_correct :
  forall a b, 
    0 <> b ->
    (-1 <= IZR a/ IZR b <= 1)%R -> 
    represents (rat_to_stream a b) (IZR a/IZR b).
Proof.
cofix.
intros a b Hbneq0 Hbound.
assert (HIZRb: IZR b<>0%R).
apply (IZR_neq b 0); omega.
str_unfold (rat_to_stream a b).
unfold str_decompose.
unfold rat_to_stream.
fold rat_to_stream.
elim (Z_le_gt_dec 0 (a * b)).

intros Hle.
generalize (rat_to_stream_correct_aux1 a b Hbneq0 Hbound Hle). 
intros H1.
assert (IZR (2 * a - b) / IZR b = 2 * (IZR a / IZR b) - 1)%R.
rewrite <- (Z_R_minus).
replace ((IZR (2 * a) - IZR b) / IZR b)%R with 
  (IZR (2 * a)/ IZR b - IZR b / IZR b)%R;[idtac|newfield; assumption].
rewrite mult_IZR.
unfold Rdiv.
change (IZR 2) with 2%R.
newfield; assumption.
constructor' rat_to_stream_correct; auto with real.
rewrite <- H.
apply rat_to_stream_correct.
assumption.
clear rat_to_stream_correct.
rewrite H; assumption.

intros Hgt.
generalize (rat_to_stream_correct_aux2 a b Hbneq0 Hbound Hgt). 
intros H1.
assert (H2: (IZR (2 * a + b) / IZR b = 2 * (IZR a / IZR b) + 1)%R).
rewrite plus_IZR.
replace ((IZR (2 * a) + IZR b) / IZR b)%R with 
  (IZR (2 * a)/ IZR b + IZR b / IZR b)%R;[idtac|newfield; assumption].
rewrite mult_IZR.
unfold Rdiv.
change (IZR 2) with 2%R.
newfield; assumption.
constructor' rat_to_stream_correct ; try auto with real.
rewrite <- H2.
apply rat_to_stream_correct.
assumption.
rewrite H2.
assumption.
Qed.

Lemma mult2_aux : forall r0,
  (-1 / 2 <= (r0 / 2 - 1 / 2) / 2 - 1 / 2 -> 1<= r0)%R.
Proof.
intros r0 H.
generalize (Rplus_le_compat_l (1/2) _ _ H )%R; clear H.
replace (1 / 2 + -1 / 2)%R with 0%R; [idtac| newfield].
replace ( 1 / 2 + ((r0 / 2 - 1 / 2) / 2 - 1 / 2))%R with 
  ((r0 / 2 - 1 / 2) / 2)%R; [idtac| newfield].
intros H.
assert (H0: (0<= 2)%R); [fourier|idtac].
generalize (Rmult_le_compat_l 2 _ _ H0 H )%R; clear H H0.
replace (2*0)%R with 0%R; [idtac| newfield].
replace (2 * ((r0 / 2 - 1 / 2) / 2))%R with
  (r0 / 2 - 1 / 2)%R; [idtac| newfield].
intros; fourier.
Qed.

Lemma mult2_aux2 : forall r0,
  ((r0 / 2 + 1 / 2) / 2 + 1 / 2 <= 1/2 -> r0 <= -1)%R.
Proof.
intros r0 H.
generalize (Rplus_le_compat_l (-1/2) _ _ H )%R; clear H.
replace (-1 / 2 + 1 / 2)%R with 0%R; [idtac| newfield].
replace ( -1 / 2 + ((r0 / 2 + 1 / 2) / 2 + 1 / 2))%R with 
  ((r0 / 2 + 1 / 2) / 2)%R; [idtac| newfield].
intros H.
assert (H0: (0<= 2)%R); [fourier|idtac].
generalize (Rmult_le_compat_l 2 _ _ H0 H )%R; clear H H0.
replace (2*0)%R with 0%R; [idtac| newfield].
replace (2 * ((r0 / 2 + 1 / 2) / 2))%R with
  (r0 / 2 + 1 / 2)%R; [idtac| newfield].
intros; fourier.
Qed.

Theorem mult2_correct :
  forall x u, (-1/2 <= u <= 1/2)%R -> represents x u ->
   represents (mult2 x) (2*u)%R.
cofix.
intros x u [H1 H2] Hrep.
str_unfold (mult2 x).
(* MODIF POUR LE PAPIER *)
destruct x as [ [ | | ] x'];[ destruct x' as [ [ | | ] x''] |
                              destruct x' as [ [ | | ] x''] |
                change (represents (str_decompose x') (2*u));
                rewrite <- str_dec_thm ]; simpl.
inversion Hrep.
inversion H0.
clear mult2_correct.
subst.
dec_tac.
assert (r0 =1)%R.
apply Rle_antisym.
assumption.
apply (mult2_aux _ H1). 
subst r0.
replace (2*((1 / 2 - 1 / 2) / 2 - 1 / 2))%R with 
  ((-1)/2 - 1/2)%R; [idtac| newfield].
constructor.
apply represents_sub_one.
split; fourier. 

inversion Hrep.
inversion H0.
rep_tac; assumption.

inversion_to_variable x''.
replace (2 * (r0 / 2 / 2 - 1 / 2))%R with 
  (2*((r0-1)/2)/2-1/2)%R;[idtac|newfield].
constructor.
apply mult2_correct.
dec_tac; split.


 fourier.


*)
destruct x as [[ | |] x']; inversion Hrep.

destruct x' as [[ | |] x'']; simpl.
(*L::L*)
clear mult2_correct.
inversion H0.
subst.
dec_tac.
assert (r0 =1)%R.
apply Rle_antisym.
assumption.
apply (mult2_aux _ H1). 
subst r0.
replace (2*((1 / 2 - 1 / 2) / 2 - 1 / 2))%R with 
  ((-1)/2 - 1/2)%R; [idtac| newfield].
constructor.
apply represents_sub_one.
split; fourier. 

(*L::R*)
inversion_to_variable x''.
rep_tac; assumption.

(*L::C*)
inversion_to_variable x''.
constructor' mult2_correct.
replace (2 * (2 * (r0 / 2 / 2 - 1 / 2)) + 1)%R with 
  (2*(r0 / 2-1/2))%R; [idtac|newfield].
apply mult2_correct; clear mult2_correct.
dec_tac.
split; fourier.
rep_tac; assumption.
subst.
split.
generalize H1 H2. 
generalize (r0 / 2 / 2 - 1 / 2)%R.
intros. 
replace (2 * (2 * r))%R with (4*r)%R; [idtac|newfield].
fourier.
clear H1 H2.
dec_tac.
replace (2 * (2 * (r0 / 2 / 2 - 1 / 2)) + 1)%R with 
  (r0 -1)%R; [idtac|newfield].
fourier.

destruct x' as [[ | |] x'']; simpl.

(*R::L*)
inversion_to_variable x''.
rep_tac; assumption.

(*R::R*)
clear mult2_correct.
inversion H0.
subst.
assert (r0 =-1)%R.
dec_tac.
apply Rle_antisym.
apply (mult2_aux2 _ H2). 
assumption.
subst r0.
replace (2*((-1 / 2 + 1 / 2) / 2 + 1 / 2))%R with 
  (1/2 + 1/2)%R; [idtac| newfield].
constructor.
apply represents_one.
split; fourier. 

(*R::C*)
inversion_to_variable x''.
constructor' mult2_correct.
replace (2 * (2 * (r0 / 2 / 2 + 1 / 2)) - 1)%R with 
  (2*(r0 / 2+1/2))%R; [idtac|newfield].
apply mult2_correct; clear mult2_correct.
dec_tac.
split; fourier.
rep_tac; assumption.
split.
replace (2 * (2 * (r0 / 2 / 2 + 1 / 2)) - 1)%R with 
  (r0 +1)%R; [idtac|newfield].
dec_tac; fourier.
subst.
generalize H1 H2. 
generalize (r0 / 2 / 2 + 1 / 2)%R.
intros.
replace (2 * (2 * r))%R with (4*r)%R; [idtac|newfield].
fourier.

(*C*)
clear mult2_correct.
simpl.
replace (2*(r/2))%R with r; [idtac|newfield].
generalize (str_dec_thm x').
unfold str_decompose.
intros H5; rewrite <- H5; assumption.
Qed.

Theorem Sopp_correct : 
  forall s r, represents s r -> represents (Sopp s) (-r).
Proof.
cofix.
intros s r Hrep.
str_unfold (Sopp s).
destruct s as [[ | | ] x']; simpl; inversion Hrep.

replace (- (r0 / 2 - 1 / 2))%R with ((-r0)/2+1/2)%R;[idtac|newfield].
constructor.
apply Sopp_correct.
assumption.
clear Sopp_correct.
dec_tac.
split; fourier.

replace (- (r0 / 2 + 1 / 2))%R with ((-r0)/2-1/2)%R;[idtac|newfield].
constructor.
apply Sopp_correct.
assumption.
clear Sopp_correct.
dec_tac.
split; fourier.

replace (- (r0 / 2))%R with ((-r0)/2)%R;[idtac|newfield].
constructor.
apply Sopp_correct.
assumption.
clear Sopp_correct.
dec_tac.
split; fourier.
Qed.



Ltac prepare_represents_constructor :=
  match goal with
    |- represents (Cons R ?a) (?b/2) =>
      replace (b/2)%R with ((b - 1)/2+1/2)%R;[idtac | field ; auto with real1]
    | |- represents (Cons L ?a) (?b/2) => 
      replace (b/2)%R with ((b + 1)/2-1/2)%R;[idtac | field ; auto with real1]
    | |- represents (Cons C ?a) (?b/2) => idtac
    | |- _ => idtac
  end.
 
Ltac interval_from_represents H :=
 (generalize (R_interval _ _ H) || 
  generalize (L_interval _ _ H) ||
  generalize (C_interval _ _ H) ||
  generalize (represents_interval _ _ H)).

Ltac intervals_from_represents H H0 :=
  interval_from_represents H; interval_from_represents H0; intros; dec_tac.



