Require Export realbasis.
Require Export proof_utils.
Require Export real_facts.
Require Export add_correct.

Open Scope Z_scope.

Theorem fact_pos :
  forall p n, p = Z_of_nat (fact n) -> 0 < p.
intros; subst; apply (inj_lt 0); apply lt_O_fact.
Qed.

Theorem fact_step : forall n, (fact n * (n + 1) = fact (n+1))%nat.
intros; rewrite plus_comm; simpl;
rewrite mult_comm; simpl; auto.
Qed.

Theorem fact_gt_1 : forall p, 2 <= p -> 1 < (Z_of_nat (fact (Zabs_nat p))).
intros p; elim p using (well_founded_ind (Zwf_well_founded 2)).
clear p.
intros p Hrec Hle; elim (Z_le_gt_dec p 2); intros Hcase.
replace p with 2; compute; auto.
omega.
assert (Zabs_nat p - 1 = Zabs_nat (p - 1))%nat.
rewrite Zabs_nat_minus.
auto.
omega.
replace (Zabs_nat p) with (S (Zabs_nat (p - 1))).
rewrite fact_simpl.
replace (S (Zabs_nat (p - 1))) with (1+Zabs_nat (p - 1))%nat.
rewrite inj_mult.
assert (1 < Z_of_nat (fact (Zabs_nat (p - 1)))).
apply Hrec; unfold Zwf; omega.
apply Zlt_le_trans with (Z_of_nat (1 + (Zabs_nat (p - 1)))*1).
rewrite Zmult_1_r.
rewrite inj_plus.
rewrite Zabs_nat_minus.
rewrite inj_minus1.
repeat rewrite Z_of_nat_Zabs_nat; simpl Z_of_nat; omega.

rewrite Zabs_nat_1.
assert (0 < Zabs_nat p)%nat.
apply (Zabs_nat_lt 0);omega.
omega.

omega.
apply Zmult_le_compat_l.
omega.
rewrite inj_plus.
apply Zplus_le_0_compat.
omega.
omega.
ring_nat.
rewrite <- H.
match goal with |- S ?x = _ => change (S x) with (1 + x)%nat end.
rewrite le_plus_minus_r.
auto.
assert (0 < Zabs_nat p)%nat.
apply (Zabs_nat_lt 0);omega.
omega.
Qed.

Theorem fact_step_minus : forall n, (1 <= n -> fact (n - 1) * n = fact n)%nat.
intros n H; induction H; simpl; auto.
rewrite <- minus_n_O; ring_nat.
Qed.

Theorem lt_0_inv_fact :
 forall k, (0 < 1 /INR(fact k))%R.
intros k; unfold Rdiv.
apply Rlt_mult_inv_pos.
fourier.
rewrite INR_IZR_INZ.
apply (IZR_lt 0).
apply (inj_lt 0).
apply lt_O_fact.
Qed.

Theorem fact_under_estimate :
 forall k i, 0<= k ->
          Z_of_nat (fact(Zabs_nat k)) * k^Z_of_nat i
                 <= Z_of_nat (fact(i+(Zabs_nat k))).
intros k i Hkpos; induction i.
simpl.
rewrite Zmult_1_r; auto with zarith.
pattern (S i) at 1;replace (S i) with (i + 1)%nat.
rewrite inj_plus.
rewrite Zpower_exp.
rewrite Zmult_assoc.
replace (k^Z_of_nat 1) with k.
simpl (S i + Zabs_nat k)%nat.
rewrite fact_simpl.
rewrite inj_mult.
rewrite Zmult_comm.
apply Zmult_le_compat.
replace (S (i + Zabs_nat k)) with (1 + (i + Zabs_nat k))%nat.
repeat rewrite inj_plus.
simpl Z_of_nat.
rewrite Z_of_nat_Zabs_nat.
omega.
assumption.
ring_nat.
exact IHi.
assumption.
apply Zmult_le_0_compat.
apply Zle_0_nat.
apply Zpower_le_0_compat; assumption.
simpl; unfold Zpower_pos; simpl; ring.
omega.
omega.
ring_nat.
Qed.

Theorem fact_pm1_fact_p :
  forall p fact_pm1, 2 <= p ->
    fact_pm1 = Z_of_nat (fact (Zabs_nat (p - 1))) ->
     fact_pm1 * p = Z_of_nat (fact (Zabs_nat p)).
intros p fact_pm1 Hpge2 Hfact.
rewrite Hfact.
pattern p at 2; rewrite <- Z_of_nat_Zabs_nat.
rewrite <- inj_mult.
rewrite Zabs_nat_minus.
rewrite Zabs_nat_1.
rewrite fact_step_minus.
auto.
apply (Zabs_nat_le 1).
omega.
omega.
omega.
Qed.

Theorem infinit_sum_inv_fact_bound :
  forall p r, 
    2 <= p ->
    infinit_sum (fun i => 1/INR(fact (i+Zabs_nat p)))%R r ->
          (0 < r <= 1/(INR (fact(Zabs_nat p-1))*(IZR p - 1)))%R.
intros p r Hpeg2 Hinf.
generalize (infinit_sum_cv _ _ Hinf).
intros Hcv.
split.
apply Rlt_le_trans with (1/(INR(fact(Zabs_nat p))))%R.
apply lt_0_inv_fact.
assert (Hpos : (forall i,  0 <= 1/(INR(fact(i + Zabs_nat p))))%R).
intros n.
apply Rlt_le.
apply lt_0_inv_fact.
exact (sum_incr _ 0 r Hcv Hpos).
simpl sum_f_R0.

assert ((forall n, 
          sum_f_R0 (fun i => (1/INR(fact (i+Zabs_nat p)))%R) n <=
          1/(INR (fact(Zabs_nat p - 1))*(IZR p -1)))%R).
intros n.
replace (1/(INR(fact (Zabs_nat p - 1))*(IZR p - 1)))%R with
        (1/(INR(fact (Zabs_nat p))) *(IZR p/(IZR p - 1)))%R.
assert (H1ltp : 1 < p).
omega.
assert (Hfactpos : (0 < INR(fact(Zabs_nat p)))%R).
apply lt_INR_0.
apply lt_O_fact.

apply Rle_trans with
  (2:= tech_exp (INR (fact (Zabs_nat p))) p n H1ltp Hfactpos).
apply sum_Rle.
intros n' _; unfold Rdiv; repeat rewrite Rmult_1_l.
apply Rle_Rinv.
apply Rmult_lt_0_compat.
apply (lt_INR 0); apply lt_O_fact.
rewrite power_IZR; apply pow_lt; apply (IZR_lt 0); omega.
apply (lt_INR 0); apply lt_O_fact.
repeat rewrite INR_IZR_INZ.
rewrite <- mult_IZR.
apply IZR_le.
apply fact_under_estimate.
omega.
replace (INR(fact (Zabs_nat p))) with
  (INR(fact (Zabs_nat p - 1)*(Zabs_nat p))).
repeat rewrite INR_IZR_INZ.
rewrite inj_mult.
rewrite Z_of_nat_Zabs_nat.
rewrite mult_IZR.
field.
assert ((IZR p <> 0)%R).
apply not_O_IZR; omega.
assert ((IZR p + -1 <> 0)%R).
replace (-1)%R with (IZR (-1)).
rewrite <- plus_IZR.
apply not_O_IZR; omega.
apply (Ropp_Ropp_IZR 1).
assert ((IZR (Z_of_nat (fact (Zabs_nat p - 1))) <> 0)%R).
rewrite <- INR_IZR_INZ.
apply not_O_INR.
apply fact_neq_0.
repeat apply prod_neq_R0;auto.
omega.
rewrite fact_step_minus.
auto.
apply (Zabs_nat_le 1).
omega.
apply Rle_cv_lim  with
         (Vn:= (fun i:nat => 1/(INR(fact (Zabs_nat p - 1))*(IZR p -1)))%R) 
 (1:= H) (2:=Hcv).
intros eps; exists 0%nat; intros n _; rewrite R_dist_eq.
auto with real.
Qed.

Theorem tech_bound :
  forall n r u, (0 < IZR u)%R -> (0 < r <= 1/IZR u)%R -> 8*n <= u -> 
    (2 * (IZR n * r) <= 1/4)%R.
intros; dec_tac.
apply Rmult_le_reg_l with 4%R.
fourier.
replace (4*(1/4))%R with 1%R;[idtac | field; auto with real1].
replace (4 * (2 * (IZR n * r)))%R with (IZR (8 *n) *r)%R.
apply Rle_trans with (IZR u * r)%R.
apply Rmult_le_compat_r.
fourier.
apply IZR_le; assumption.
replace 1%R with (IZR u * (1/IZR u))%R.
apply Rmult_le_compat_l.
fourier.
assumption.
auto with real.
field.
apply Rlt_neq_0; assumption.
rewrite mult_IZR;simpl IZR;ring.
Qed.

Ltac apply_e_correct1' e_series' :=
 match goal with 
   e_correct1 : _, id : represents ?v ?vr, id1 : infinit_sum _ ?r
   |- represents (e_series' ?v (?n, ?p, ?f)) ?w =>
      replace w with (vr + IZR n * r)%R;
        [ apply e_correct1;clear e_correct1;auto with real1 | clear e_correct1]
 end.

Theorem e_series_body_correct :
  forall (e_series : stream idigit -> Z*Z*Z -> stream idigit),
  (forall v vr r n p fact_pm1,
    fact_pm1 = (Z_of_nat (fact (Zabs_nat (p-1)))) ->
    4 * n <= fact_pm1 * (p - 1) ->
    2 <= p -> 
    1 <= n -> 
    represents v vr ->
    infinit_sum (fun i => (1/INR(fact (i+Zabs_nat p)))%R) r ->
    (vr + (IZR n) *r <= 1)%R ->
    represents (e_series v (n,p,fact_pm1))(vr+(IZR n)*r)%R) ->
  forall v vr r n p fact_pm1,
    fact_pm1 = (Z_of_nat (fact (Zabs_nat (p-1)))) ->
    8 * n <= fact_pm1 * (p - 1) ->
    2 <= p -> 1 <= n -> 
    represents v vr ->
    infinit_sum (fun i => (1/INR(fact (i+Zabs_nat p)))%R) r ->
    (vr + (IZR n) *r <= 1)%R ->
    represents (series_body_pos (Z*Z*Z) e_series v (2*n, p, fact_pm1))
       (vr+(IZR n)*r)%R.
intros e_series' e_correct1.
intros v vr r n p fact_pm1 Hfact Hle Hpge2 Hnge1 Hv Hsum Hin_res.
unfold series_body_pos.

assert (Hfactpos : 0 < fact_pm1).
clear e_correct1.
rewrite Hfact.
apply (inj_lt 0).
apply lt_O_fact.
assert (Hprodpos : (0 < IZR(fact_pm1 *(p-1)))%R).
clear e_correct1.
apply (IZR_lt 0); apply Zmult_lt_O_compat.
assumption.
omega.
assert (Hboundr : (2*(IZR n* r) <= 1/4)%R).
clear e_correct1.
apply tech_bound with (fact_pm1 * (p - 1)).
assumption.
generalize (infinit_sum_inv_fact_bound _ _ Hpge2  Hsum).
rewrite INR_IZR_INZ; rewrite mult_IZR; rewrite Hfact.
rewrite <- Z_R_minus; simpl IZR.
rewrite Zabs_nat_minus.
intros; assumption.
omega.
assumption.
assert (Hnrpos : (0 < IZR n * r)%R).
apply Rmult_lt_0_compat.
apply (IZR_lt 0); omega.
generalize (infinit_sum_inv_fact_bound _ _ Hpge2  Hsum).
intros; dec_tac; assumption.

destruct v as [ [ | | ] v'']; inversion Hv.

destruct v'' as [ [ | | ] v3]; inversion_to_variable v3.
(* case : L::L *)
add_fact v3.
constructor' e_correct1.
apply_e_correct1' e_series'.
omega.
omega.
rewrite mult_IZR; simpl IZR; rewrite Rmult_assoc.
dec_tac; fourier.
rewrite mult_IZR; simpl IZR; field; auto with real1.
replace ( 2 * (r0 / 2 - 1 / 2 + IZR n * r) + 1)%R with
  (r0 + 2 * (IZR n * r))%R; [idtac| newfield].
dec_tac; split; fourier.

(* case : L::R *)
add_fact v3.

constructor' e_correct1.
apply_e_correct1' e_series'.
omega.
omega.
rewrite mult_IZR; simpl IZR; rewrite Rmult_assoc.
dec_tac; fourier.
rewrite mult_IZR; simpl IZR; field; auto with real1.
replace (2 * ((r1 / 2 + 1 / 2) / 2 - 1 / 2 + IZR n * r))%R with 
  (r1/2 -1/2 + 2 * (IZR n * r))%R; [idtac| newfield].
dec_tac; split; fourier.

(* case : L::C *)
add_fact v3.
constructor' e_correct1.
apply_e_correct1' e_series'.
omega.
omega.
rewrite mult_IZR; simpl IZR; rewrite Rmult_assoc.
dec_tac; fourier.
rewrite mult_IZR; simpl IZR; field; auto with real1.
replace (2 * (r0 / 2 - 1 / 2 + IZR n * r) + 1)%R with 
  (r0 + 2 * (IZR n * r))%R; [idtac| newfield].
dec_tac; split;fourier.

(* case : R *)
constructor' e_correct1.
apply_e_correct1' e_series'.
omega.
omega.
rewrite mult_IZR; replace r0 with (2*vr -1)%R;
  [simpl IZR; rewrite Rmult_assoc| subst vr;ring].
dec_tac; fourier.
field; auto with real1.
rewrite mult_IZR; simpl IZR; field; auto with real1.
replace (2 * (r0 / 2 + 1 / 2 + IZR n * r) - 1)%R with 
  (r0 + 2 * (IZR n * r))%R; [idtac| newfield].
dec_tac;split; fourier.


destruct v'' as [ [ | | ] v3]; inversion_to_variable v3.
(* case : C::L *)
add_fact v3.
constructor' e_correct1.
apply_e_correct1' e_series'.
omega.
omega.
rewrite mult_IZR; simpl IZR; rewrite Rmult_assoc.
dec_tac; fourier.
rewrite mult_IZR; simpl IZR; field; auto with real1.
replace (2 * (r0 / 2 + IZR n * r))%R with 
  (r0 + 2 * (IZR n * r))%R; [idtac| newfield].
dec_tac; split;fourier.

(* case : C::R *)
add_fact v3.
constructor' e_correct1.
apply_e_correct1' e_series'.
omega.
omega.
rewrite mult_IZR; replace r1 with (2*r0 -1)%R;
  [simpl IZR; rewrite Rmult_assoc| subst r0;field;auto with real1;fail].
autorewrite with real1 real2.
dec_tac; fourier.
rewrite mult_IZR; simpl IZR; field; auto with real1.
replace (2 * ((r1 / 2 + 1 / 2) / 2 + IZR n * r))%R with 
  (r1/2 +1/2 + 2 * (IZR n * r))%R; [idtac| newfield].
dec_tac; split; fourier.

(* case C::C *)
add_fact v3.
constructor' e_correct1.
apply_e_correct1' e_series'.
replace 4%R with (IZR 4);[idtac | simpl IZR; ring].
omega.
omega.
rewrite mult_IZR; simpl IZR; rewrite Rmult_assoc.
dec_tac; fourier.
rewrite mult_IZR; simpl IZR; field; auto with real1.
rewrite double_plus_distr.
autorewrite with real1.
dec_tac; split;fourier.
Defined.

Theorem e_correct1_tech_lemma : 
  forall n fact_pm1 p,
    fact_pm1 = Z_of_nat (fact (Zabs_nat (p - 1))) ->
    2 <= p ->
    1 <= n ->
    4 * n <= fact_pm1 * (p - 1) ->
    (-1 <= IZR n / IZR (fact_pm1 * p) <= 1)%R.
Proof. 
intros n fact_pm1 p Hfact Hpge2 Hnge1 Hinv.
unfold Rdiv.
assert (0 < IZR (fact_pm1 * p))%R.
apply (IZR_lt 0).
apply Zlt_le_trans with (fact_pm1 * (p - 1)).
omega.
apply Zmult_le_compat_l; omega.

split.
apply Rle_trans with 0%R.
fourier.
apply Rle_mult_inv_pos.
apply (IZR_le 0); omega.
assumption.

apply Rmult_le_reg_l with (IZR (fact_pm1 * p))%R.
assumption.
rewrite <- Rmult_assoc.
rewrite Rinv_r_simpl_m.
rewrite Rmult_1_r.
apply IZR_le.
apply Zle_trans with (4*n).
omega.
apply Zle_trans with (1:=Hinv).
apply Zmult_le_compat_l; omega.
apply Rgt_not_eq.
auto with real.
Qed.

Theorem e_correct1 :
 forall v vr r n p fact_pm1, fact_pm1 = (Z_of_nat (fact (Zabs_nat (p-1)))) ->
   4 * n <= fact_pm1 * (p-1) ->
   2 <= p ->
   1 <= n ->
   represents v vr ->
   infinit_sum (fun i => (1/INR(fact (i+Zabs_nat p)))%R) r ->
   (vr + (IZR n)*r <= 1)%R ->
   represents (e_series v (n, p, fact_pm1)) (vr+(IZR n)*r)%R.
cofix.
intros v vr r n p fact_pm1 Hfact Hinv Hpge2 Hnge1 Hrepr Hinf Hin.
rewrite (str_dec_thm (e_series v (n, p, fact_pm1))); unfold str_decompose.
lazy zeta beta iota delta [e_series]; fold e_series.
elim (Z_le_gt_dec (8*n) (fact_pm1 *(p - 1))).
intros Hle.
fold (str_decompose (series_body_pos (Z*Z*Z) e_series v (2*n,p,fact_pm1))).
rewrite <- str_dec_thm.
exact (e_series_body_correct e_series e_correct1 v vr r n p fact_pm1
   Hfact Hle Hpge2 Hnge1 Hrepr Hinf Hin).

intros Hgt.
replace (vr + IZR n * r)%R with
   ((vr + (IZR n)/(IZR (fact_pm1 * p))) + (IZR n) *
     (r - 1/(IZR (fact_pm1 * p))))%R;
  [idtac | clear e_correct1; newfield].
fold (str_decompose (series_body_pos (Z*Z*Z) e_series
        (v+rat_to_stream n (fact_pm1 * p))%S
        (2 * n, p + 1, fact_pm1 * p))); rewrite <- str_dec_thm.

apply (e_series_body_correct e_series e_correct1 
        (v + rat_to_stream n (fact_pm1 * p))%S
        (vr + (IZR n)/(IZR (fact_pm1 * p)))%R
        (r - 1/(IZR (fact_pm1 * p)))%R
        n (p + 1) (fact_pm1 * p)); clear e_correct1.
ring (p + 1 - 1).
pattern p at 1; rewrite <- (Z_of_nat_Zabs_nat p).
rewrite Hfact.
rewrite <- inj_mult.
rewrite Zabs_nat_minus.
rewrite Zabs_nat_1.
rewrite (fact_step_minus (Zabs_nat p)).
auto.
rewrite <- Zabs_nat_1.
apply lt_le_weak.
apply Zabs_nat_lt.
omega.
omega.
omega.
apply Zle_trans with (2* (fact_pm1 * (p - 1))).
omega.
rewrite Zmult_permute.
rewrite <- Zmult_assoc.
apply Zmult_le_compat_l.
apply Zmult_le_compat; omega.
rewrite Hfact.
apply Zle_0_nat.
omega.
omega.
apply add_correct.
assumption.
apply rat_to_stream_correct.

apply Zlt_not_eq.
apply Zlt_le_trans with (fact_pm1 * (p - 1)).
omega.
apply Zmult_le_compat_l; omega.

apply e_correct1_tech_lemma; assumption.

split.
rewrite mult_IZR.
generalize (represents_interval v vr Hrepr).
intros [Hvr _].
assert (0 <= IZR n / (IZR fact_pm1 * IZR p))%R.
apply Rdiv_le_0_compat.
apply (IZR_le 0); omega.
rewrite <- mult_IZR.
apply (IZR_lt 0). 
apply Zmult_lt_O_compat.
apply (fact_pos _ _ Hfact).
omega.
fourier.

apply Rle_trans with (2:= Hin).
apply Rplus_le_compat_l.
replace (IZR n/IZR(fact_pm1 * p))%R with (IZR n * (1/IZR(fact_pm1 * p)))%R.
2: field;auto with real1.
apply Rmult_le_compat_l.
apply (IZR_le 0).
omega.

generalize (infinit_sum_step _ _ Hinf); intros Hinf_step.
simpl (0 + Zabs_nat p)%nat in Hinf_step.
assert (Hext:forall n,
           ((fun i => 1/INR(fact (i+1+Zabs_nat p))) n=
            (fun i => 1/INR(fact (i+(Zabs_nat (p + 1))))) n)%R).
intros i; simpl; replace (i+1+Zabs_nat p)%nat 
            with (i + (Zabs_nat (p + 1)))%nat.
auto.
rewrite Zabs_nat_plus.
rewrite Zabs_nat_1.
ring.
omega.
omega.
generalize 
  (infinit_sum_ext 
     (fun i => 1/INR(fact(i + 1 + Zabs_nat p)))%R
     (fun i => 1/INR(fact(i + (Zabs_nat (p + 1)))))%R  _ Hext Hinf_step).
intros Hinf_step'.
assert (Hpp1ge2 : 2 <= p + 1).
omega.
generalize (infinit_sum_inv_fact_bound (p+1) _ Hpp1ge2 Hinf_step').
intros Hint'.
rewrite INR_IZR_INZ in Hint'.
dec_tac.
rewrite <- fact_step_minus in H.
rewrite <- Zabs_nat_1 in H.
rewrite <- Zabs_nat_minus in H.
rewrite inj_mult in H.
rewrite <- Hfact in H.
rewrite Z_of_nat_Zabs_nat in H.
fourier.
omega.
omega.
rewrite <- Zabs_nat_1.
apply Zabs_nat_le.
omega.

rewrite mult_IZR.
apply prod_neq_R0.
apply not_O_IZR.
assert (0 < fact_pm1).
rewrite Hfact.
apply (inj_lt 0).
apply lt_O_fact.
omega.
apply not_O_IZR.
omega.

apply infinit_sum_ext with (fun i => 1/INR(fact ((i+1)+Zabs_nat p)))%R.
replace (IZR (fact_pm1 * p)) with (INR (fact (0 + (Zabs_nat p)))).
intros i; rewrite Zabs_nat_plus.
rewrite Zabs_nat_1.
rewrite (plus_comm (Zabs_nat p)).
rewrite plus_assoc.
auto.
omega.
omega.
simpl.
rewrite INR_IZR_INZ.
rewrite Hfact.
pattern p at 3; rewrite <- Z_of_nat_Zabs_nat.
rewrite <- inj_mult.
rewrite Zabs_nat_minus.
rewrite Zabs_nat_1.
rewrite fact_step_minus.
auto.
rewrite <- Zabs_nat_1; apply Zabs_nat_le.
omega.
omega.
omega.
rewrite fact_pm1_fact_p.
rewrite <- INR_IZR_INZ.
apply infinit_sum_step with (An := (fun i => 1/INR (fact (i + Zabs_nat p)))%R).
assumption.
assumption.
assumption.

match goal with |- (?x <= 1)%R => replace x with (vr + IZR n * r)%R end.
assumption.
unfold Rdiv; ring.

rewrite mult_IZR.
apply prod_neq_R0.
apply not_O_IZR.
assert (0 < fact_pm1).
rewrite Hfact.
apply (inj_lt 0).
apply lt_O_fact.
omega.
apply not_O_IZR.
omega.
Qed.

Theorem e_correct :
  forall r,
  infinit_sum (fun i => (1/INR(fact(i+Zabs_nat 2)))%R) r ->
  represents number_e_minus2 r.
intros r H.
unfold number_e_minus2.
replace r with (1/2+(IZR 1)*(r-(1/2)))%R;[idtac | ring].
apply e_correct1.
auto.
omega.
omega.
omega.
replace 2%R with (IZR 2);[idtac | simpl IZR; ring].
replace 1%R with (IZR 1);[idtac | simpl IZR; ring].
apply rat_to_stream_correct.
omega.
change (IZR 1) with 1%R.
change (IZR 2) with 2%R.
split; fourier.

apply infinit_sum_ext with 
  (fun i => (fun j =>(1/INR(fact(j + Zabs_nat 2)))%R) (i+1)%nat).
intros n; unfold Zabs_nat, nat_of_P; simpl.
replace (n +1 + 2)%nat with (n+3)%nat;[auto|ring].
replace (1/2)%R with ((fun j =>(1/INR(fact(j + Zabs_nat 2)))) 0%nat)%R.
apply infinit_sum_step with (An := (fun j => (1/INR(fact(j + Zabs_nat 2)))%R)).
assumption.
compute; trivial.
assert (Hint: (0 < r <= 1/(INR(fact(Zabs_nat 2 -1)) *(IZR 2 - 1)))%R).
apply infinit_sum_inv_fact_bound; auto with zarith.
destruct Hint as [H1 H2].
match goal with id : (r <= 1/?x)%R |- _ => replace x with 1%R in id end.
simpl IZR; match goal with |- (?x <=1)%R => replace x with r end.
fourier.
compute; ring.
simpl IZR; simpl fact; ring.
Qed.

