Require Export realbasis.
Require Export proof_utils.
Require Export real_facts.

Lemma sub_aux_stable : 
  forall u v, (-1<=u <= 1 -> -1 <= v <=1-> -1 <= (u-v)/2 <= 1)%R. 
Proof.
intros u v H1 H2; dec_tac.
autorewrite with real1.
split; fourier.
Qed.

(*en fonction du bit d remplace b par ce qu'il faut en fonction de u
  pour que le constructeur de represents trouve ce qu'il faut
*)
Ltac represents_replace :=
  match goal with id : represents ?s ?u
      |- represents (?d::?s) ?b =>
	replace b with 
	  match d with 
	    | L => (u/2-1/2)%R
	    | C => (u/2 )%R
	    | R => (u/2+1/2)%R
	  end; 
	  [constructor; [assumption| split; try auto with real1] | field; discrR]
  end.
(*on veut montrer qu'un stream represente un reel et on connait ses 2 premiers digits
  on a besoin d'une hypothese ou on connnait la representation du stream 
  sans le premier digit
*)
Ltac rep_prepare_hyp := 
  match goal with 
    | id1 : represents ?s ?u, id2 : (-1 <= ?u <=1)%R
      |- represents (?d1::?d2::?s) ?b =>
	match d2 with 
	  | L => generalize (reprL s u id1 id2)
	  | C => generalize (reprC s u id1 id2)
	  | R => generalize (reprR s u id1 id2)
 	end; intros
  end.

Theorem impl_CL_LR :
  forall s u , represents (C::L::s) u -> represents (L::R::s) u.
Proof.
  intros s u H;  inversion H.
  inversion H1.
  inversion H6.
  rep_prepare_hyp; represents_replace.
Qed.

Theorem impl_CR_RL : 
  forall s u , represents (C::R::s) u -> represents (R::L::s) u.
Proof.
  intros s u H;  inversion H.
  inversion H1.
  inversion H6.
  rep_prepare_hyp; represents_replace.
Qed.

Theorem impl_LR_CL : 
  forall s u , represents (L::R::s) u -> represents (C::L::s) u.
Proof.
  intros s u H; inversion H.
  inversion H1.
  inversion H6.
  rep_prepare_hyp; represents_replace.
Qed.

Theorem impl_RL_CR : 
  forall s u , represents (R::L::s) u -> represents (C::R::s) u.
Proof.
  intros s u H; inversion H.
  inversion H1.
  inversion H6.
  rep_prepare_hyp; represents_replace.
Qed.

Theorem decalage_C_L : 
  forall d s u, represents (d::C::s) u -> represents (d::L::s) (u-1/4).
Proof.
  intros d; case d; intros s u H; inversion H;
  inversion H1;  inversion H6;
  rep_prepare_hyp;  represents_replace.
Qed.

Theorem decalage_C_R : 
  forall d s u, represents (d::C::s) u -> represents (d::R::s) (u+1/4).
Proof.
  intros d; case d; intros s u H;
  inversion H; inversion H1;  inversion H6;
  rep_prepare_hyp; represents_replace.
Qed.

Theorem decalage_R_C : 
  forall d s u, represents (d::R::s) u -> represents (d::C::s) (u-1/4).
Proof.
  intros d; case d; intros s u H; inversion H;
  inversion H1;  inversion H6;
  rep_prepare_hyp; represents_replace.
Qed.

Theorem decalage_L_C : 
  forall d s u, represents (d::L::s) u -> represents (d::C::s) (u+1/4).
Proof.
  intros d; case d; intros s u H; inversion H;
  inversion H1; inversion H6;
  rep_prepare_hyp; represents_replace.
Qed.

(*on veut fabriquer une hypothese du type represents (d::x) u
  a partir des autres hypotheses et des equivalences entre prefixe
*)
Ltac find_represents_eq x :=
  match x with 
    |L::?x' => 
      match goal with 
	|id : represents (L::R::x') _ |- _ => 
	  generalize  (impl_LR_CL _ _ id); clear id; inversion 1
	|id : represents (C::R::x') _ |- _ => 
	  generalize  (impl_CR_RL _ _ id); clear id; inversion 1
	| id : represents (_::C::x') _ |- _ => 
	  generalize  (decalage_C_L _ _ _ id); clear id; inversion 1
	| _ => idtac "find_represents_eq L"
      end
    |C::?x' => 
      match goal with 
	|id : represents (_::L::x') _ |- _ => 
	  generalize  (decalage_L_C _ _ _ id); clear id; inversion 1
	|id : represents (_::R::x') _ |- _ => 
	  generalize  (decalage_R_C _ _ _ id); clear id; inversion 1
	| _ =>idtac "find_represents_eq C"
      end     
    |R::?x' => 
      match goal with 
	|id : represents (R::L::x') _ |- _ => 
	  generalize  (impl_RL_CR _ _ id); clear id; inversion 1
	|id : represents (C::L::x') _ |- _ => 
	  generalize  (impl_CL_LR _ _ id); clear id; inversion 1
	| id : represents (_::C::x') _ |- _ => 
	  generalize  (decalage_C_R _ _ _ id); clear id; inversion 1
	| _ => idtac "find_represents_eq R"
      end
  end.

Ltac find_represents x :=
 match goal with 
   |id : represents x _ |- _ => idtac
   |id : represents (_::x) _ |- _ => inversion id; clear id
   |_ => find_represents_eq x 
 end.


(*essaye de fabriquer les hypotheses qui donnent les valeurs representees par 
  les termes sur lesquels on veut appliquer sub_aux 
  si on connait le premier digit d'un argument, on va essayer si necessaire 
  d'utiliser l'equivalence de certains prefixes
*)
Ltac prepare_hyp := 
  match goal with 
    | |- represents (_::half_sub (?d1::?x) (?d2::?y)) _ =>
      find_represents_eq d1 x;
      find_represents_eq d2 y
    | |- represents (_::half_sub (?d1::?x) ?y) _ =>
      find_represents_eq d1 x;
      find_represents y
    | |- represents (_::half_sub ?x (?d2::?y)) _ =>
      find_represents x;
      find_represents_eq d2 y
    | |- represents (_::half_sub ?x ?y) _ => 
      find_represents x;
      find_represents y
    | _ => idtac "prepare_hyp"
  end.
    
(*on va essayer de transformer b dans le but de type represents (d::sub_aux x y) b
  en connaissant les valeurs representes par x et y, de telle sorte que l'on puisse 
  utiliser le bon constructeur de represents pour simplifier et qu'on puisse ensuite
  appliquer l'hypothese de coinduction H fournie.
  on fait la preuve du replace et de l'inegalite engendree par constructor
 *)
Ltac simpl_represent_sub_aux H :=
match goal with 
  id1 : represents ?x ?u, id2 : represents ?y ?v
  |- represents (?d::half_sub ?x ?y) ?b => 
    replace b with 
      match d with 
	| L => (((u-v)/2)/2-1/2)%R 
	| C => ((u-v)/2/2)%R
	| R => (((u-v)/2)/2+1/2)%R
      end;
      [constructor; [apply H; assumption| apply sub_aux_stable;assumption ]| field; discrR]
end.

(*on a une hypothese qui nous donne un terme equivalent a "a +/- 1/8",
  on va faire le replace qu'il faut pour pouvoir faire un subst a ensuite
*)
Ltac subst_aux a :=
  match goal with 
    | id : ?b = (a + 1/4)%R |- _ => 
      assert (b-1/4=a)%R; [rewrite id; field; discrR | subst a]
    | id : ?b = (a - 1/4)%R |- _ => 
      assert (b+1/4=a)%R; [rewrite id; field; discrR | subst a]
  end.	

Theorem subs_aux_correct :
  forall x y u v, represents x u -> represents y v -> 
    represents (half_sub x y) ((u - v)/2).
Proof.
cofix.
intros x y u v H1 H2.
str_unfold (half_sub x y).
(*on enleve les cas simples ou un bit de chaque cote suffit*)
destruct x as [ [ | | ] x'] ; destruct y as [ [ | | ] y'];
try (simpl; prepare_hyp; simpl_represent_sub_aux subs_aux_correct);
(*on enleve ensuite ceux ou le 2e bit de y est necesaire*)
destruct y' as [ [ | | ] y''];
intros;
try(simpl; prepare_hyp; simpl_represent_sub_aux subs_aux_correct);
(*on enleve enfin ceux ou le 2e bit de x est necesaire*)
destruct x' as [ [ | | ] x''];
try(simpl; prepare_hyp; simpl_represent_sub_aux subs_aux_correct);
(*pour les 8 cas pathetiques ou on a fait des decalages
  il reste a demontrer la validite d'un replace que field n'a pas su demontrer
*)
subst_aux v;
subst_aux u;
field; discrR.
Qed.

Theorem sub_correct : 
  forall x y u v, represents x u -> represents y v -> (-1 <= u - v<=1)%R -> 
    represents (x - y)%S (u - v)%R.
Proof.
intros x y u v H1 H2 H3.
unfold sub.
replace (u-v)%R with  (2*((u-v)/2))%R;[idtac| field; discrR].
apply mult2_correct.
dec_tac; split; auto with real1.
apply subs_aux_correct; assumption.
Qed.
