Require Import realbasis.
Require Import real_value.

Definition cst_si := fun A : Set=> fun x : A => stream idigit.
Implicit Arguments cst_si.

Definition lower_bound_to_digit (rat : Z * Z) : idigit :=
  let (num, den) := rat in
  match 4*num ?= den with
    Lt => L | _ => match 2*num ?= den with Lt => C | _ => R end
  end.

Definition compute_lower_bound (d:idigit) : Z*Z :=
  match d with L => (0,1) | C => (1,4) | R => (1,2) end.

Ltac solve_it f opp tl :=
 match goal with
  |- cst_si ?d => 
  apply Cons with (A:=idigit);
   [exact (lower_bound_to_digit (f (compute_lower_bound d))) | exact (opp tl)]
 end.

Definition rat_add (r1 r2 : Z*Z) : Z*Z :=
  let (n1,d1) := r1 in let (n2, d2) := r2 in (n1*d2+n2*d1,d1*d2).

Definition rat_mult (r1 r2 : Z*Z) : Z*Z :=
  let (n1,d1) := r1 in let (n2, d2) := r2 in (n1*n2,d1*d2).

Definition opp_map_bound (r: Z*Z) : (Z*Z) :=
   rat_add (1,1) (rat_mult (-1,1) (rat_add r (1,2))).

Definition opp : stream idigit -> stream idigit.
cofix.
intros s; case s; intros d tl; change (cst_si d); case d;
solve_it opp_map_bound opp tl.
Defined.

Definition lift' (d:idigit)(x:Rdefinitions.R):Rdefinitions.R :=
  match d with L => (x/2)%R | R => ((x+1)/2)%R | C => ((2*x+1)/4)%R end.

Definition anti_lift (d:idigit)(x:Rdefinitions.R):Rdefinitions.R :=
  match d with L => (2*x)%R | R => (2*x - 1)%R | C => (2*x-(1/2))%R end.

Theorem lift_inverse : forall d x, lift' d (anti_lift d x) = x.
intros [ | | ] x; unfold lift', anti_lift; field; try apply nonz_2.
apply Rlt_neq_0; fourier.
Qed.

Definition new_digit (d:idigit) : idigit := 
  lower_bound_to_digit (opp_map_bound (compute_lower_bound d)).

Theorem opp_correct : forall x r, represents x r -> represents (opp x) (1-r)%R.
cofix.
intros [[ | | ] tl] r Hrec; inversion Hrec;
match goal with 
  id : (0 <= ?r'<= 1)%R |- represents (opp (?d::?tl)) (1 - ?r)%R =>
   let d' := eval compute in (new_digit d) in
   (rewrite (str_dec_thm (opp (d::tl))); simpl;
   rewrite <- (lift_inverse d' (1-r)); unfold lift'; 
   replace (anti_lift d' (1-r)) with (1 - r')%R);
     [constructor; [apply opp_correct; assumption | 
                    clear opp_correct; destruct id; split; fourier]
      | clear opp_correct; unfold anti_lift; field; apply Rlt_neq_0; fourier]
end.
Qed.

