Require Export realbasis.
Require Export Field_tac.
Require Export proof_utils.
Require Export real_facts.
Require Export add_correct.

Definition terme_arctan_1_n_R (n i: Z) :=
  (IZR (n^2*(4*i+3)-4*i-1)%Z / IZR (n^(4*i+3)*(4*i+1)*(4*i+3))%Z)%R.

Lemma Zle_neq_lt :
  forall m n, m <> n -> m <= n -> m < n.
Proof.
intros; omega.
Qed.

Lemma Zpow_pos :
  forall i n, 0 < n -> 0 <=i -> 0 < n ^i. 
Proof.
induction i; intros.
simpl; omega.
apply  Zle_neq_lt.
apply sym_not_eq.
apply Zpower_not_0.
auto with zarith.
omega.
apply Zpower_le_0_compat.
omega.
generalize (Zlt_neg_0 p).
intros H1.
omega.
Qed.

Theorem Zpower_aux : 
  forall n, 1 <= 2 ^ n -> 0<= n.
Proof.
unfold Zpower.
induction n.
intros; omega.
rewrite Zpower_pos_nat.
rewrite (Zpos_eq_Z_of_nat_o_nat_of_P p).
generalize (nat_of_P p).
intros.
auto with zarith.
intuition.
Qed.

Theorem Zpower_incr : 
  forall a b n, 0<= n -> 0<=a<=b -> a^n <= b^n.
Proof.
intros a b n.
unfold Zpower.
case n; clear n.
intros; omega.
intros p H1 H2. 
repeat rewrite (Zpower_pos_nat).
generalize (nat_of_P p).
induction n.
unfold Zpower_nat.
simpl; omega.
replace (S n) with (n + 1)%nat;[idtac|omega].
repeat rewrite Zpower_nat_is_exp.
apply Zmult_le_compat.
assumption.
unfold Zpower_nat.
simpl; omega.
clear IHn.
induction n.
unfold Zpower_nat.
simpl; omega.
replace (S n) with (n + 1)%nat;[idtac|omega].
rewrite Zpower_nat_is_exp.
apply Zmult_le_0_compat.
assumption.
unfold Zpower_nat.
simpl; omega.
unfold Zpower_nat.
simpl; omega.
intros; omega.
Qed.

Theorem terme_atan_1_n_bound :
  forall n i, 2<=n -> 0<=i -> ( (IZR (2^(4*i+1)) * terme_arctan_1_n_R n i) < 1)%R.
Proof.
intros n i H2len H0lei.
assert (H0: 0 <= 2 <= n); [omega|idtac].
assert (H1: 0<= 4*i+1); [omega|idtac].
generalize (Zpower_incr _ _ _ H1 H0).
intros H2.
assert (H3: 1<= 4*i+1); [omega|idtac].
assert (H9: 0< 2 ^ (4 * i + 1)).
apply Zpow_pos; omega.
assert (H4:  0<= 2 ^ (4 * i + 1)); [omega|idtac].
assert (H5: 0<=1); [omega|idtac].
generalize (Zmult_le_compat _ _ _ _ H2 H3 H4 H5).
replace (2 ^ (4 * i + 1) * 1)%Z with (2 ^ (4 * i + 1) ); [idtac|ring] .
intros H6.
unfold terme_arctan_1_n_R.
assert (H7: 0 < n ^ (4 * i + 3) * (4 * i + 1) * (4 * i + 3)).
apply Zmult_lt_O_compat.
apply Zmult_lt_O_compat.
apply  Zpow_pos; omega.
omega.
omega.
unfold Rdiv.
repeat rewrite <- Rmult_assoc.
rewrite <- Rmult_comm.
apply Rmult_lt_reg_l with (IZR (n ^ (4 * i + 3) * (4 * i + 1) * (4 * i + 3))).
apply (IZR_lt _ _ H7).
rewrite <- Rmult_assoc.
replace (IZR (n ^ (4 * i + 3) * (4 * i + 1) * (4 * i + 3)) *
    / IZR (n ^ (4 * i + 3) * (4 * i + 1) * (4 * i + 3)))%R with 1%R.
change 1%R with (IZR 1).
repeat rewrite <- mult_IZR.
apply IZR_lt. 
apply Zlt_le_trans with (2 ^ (4 * i + 1) * (n ^ 2 * (4 * i + 3)))%Z.
rewrite Zmult_assoc.
replace (1 * (2 ^ (4 * i + 1)))%Z with (2 ^ (4 * i + 1)); [idtac|ring] .
replace (n ^2*(4*i+3) - 4*i-1)%Z with  (n^2*(4*i+3) - (4 * i+1))%Z; [idtac|ring].
rewrite Zmult_minus_distr_l.
apply Zlt_minus_simpl_swap.
apply Zmult_lt_O_compat; omega.
rewrite Zmult_assoc.
replace  (n ^ (4 * i + 3) * (4 * i + 1) * (4 * i + 3) * 1)%Z with 
  (n ^ (4 * i + 3) * (4 * i + 1) * (4 * i + 3))%Z; [idtac|ring].
apply Zmult_le_compat_r; [idtac|omega].
apply Zle_trans with  (n ^ (4 * i + 3)).
replace (4 * i + 3)%Z with ((4*i+1) + 2)%Z; [idtac|ring].
rewrite (Zpower_exp n (4*i+1) 2); try omega.
apply Zmult_le_compat_r.
apply Zpower_incr; omega.
assert (0<n^2).
apply Zpow_pos; omega.
omega.
assert (H10: 0< n ^ (4 * i + 3)).
apply Zpow_pos; omega.
assert (H8: 0<= n ^ (4 * i + 3)).
omega.
generalize (Zmult_le_compat_l _ _ _ H3 H8).
intros; omega.
field.
discrR.
assert (H10: 0< n ^ (4 * i + 3)).
apply Zpow_pos; omega.
apply Zmult_non0.
apply Zmult_non0.
omega.
omega.
omega.
Qed.

Theorem terme_atan_1_n_pos:
  forall n i, 0<n -> 0<=i -> (0 < terme_arctan_1_n_R n i)%R.
Proof.
intros n i H0ltn H0lei.
unfold terme_arctan_1_n_R.
unfold Rdiv.
assert (H1:0< 4*i+1).
omega.
assert (H2:0< 4*i+3).
omega.
assert (H3:0<n ^ (4 * i + 3)).
assert (H3:0<= 4*i+3).
omega.
generalize ( Zpow_pos (4 * i + 3) n H0ltn H3).
intros; omega.
cut ((0 < IZR (n ^ 2 * (4 * i + 3) - 4 * i - 1)%Z)%R).
intros H4.
cut ((0 < / IZR (n ^ (4 * i + 3) * (4 * i + 1) * (4 * i + 3)))%R).
intros H5.
generalize (Rmult_lt_compat_r _ _ _ H5 H4).
replace (0 * / IZR (n ^ (4 * i + 3) * (4 * i + 1) * (4 * i + 3)))%R with 0%R.
tauto.
field.
discrR.
apply Zmult_non0.
apply Zmult_non0.
omega.
omega.
omega.
apply Rinv_0_lt_compat.
change 0%R with (IZR 0). 
apply IZR_lt.
apply Zmult_lt_O_compat.
apply Zmult_lt_O_compat; assumption.
assumption.
change 0%R with (IZR 0). 
apply IZR_lt.
assert (H4: 1*(4*i+3) <= n^2*(4*i+3)).
apply Zmult_le_compat_r.
assert (H5: 0<n^2).
apply Zpow_pos.
assumption.
omega.
omega.
omega.
omega.
Qed.

Theorem finite_sum_x_pow_i: 
  forall x n, x <> 1%R -> sum_f_R0 (fun i => x^i)%R n = ((1-x^(n+1)%nat) / (1-x))%R.
intros x n H.
apply Rmult_eq_reg_l with (x-1)%R.
rewrite Rmult_comm.
rewrite GP_finite.
field.
auto with real.
auto with real.
Qed.

Theorem INR_Zabs_nat_eq :
  forall z, (0 <= z)%Z -> (IZR z = INR(Zabs_nat z))%R.
unfold Zabs_nat.
intros z; case z.
simpl.
intros; auto.
unfold IZR.
intros; auto.
compute.
intros x H'; elim H'; auto.
Qed.

Theorem pow_exp : (forall x, 0 < x -> forall n, x^n=exp (ln x * (INR n)))%R.
intros x Hpos.
induction n.
rewrite Rmult_0_r.
rewrite exp_0; trivial.
rewrite S_INR; simpl.
rewrite Rmult_plus_distr_l.
rewrite Rmult_1_r.
rewrite exp_plus.
rewrite exp_ln.
rewrite IHn; ring.
assumption.
Qed.

Theorem pow_lt_1 :
  (forall eps, 0 < eps -> forall x, 0 <= x < 1 -> exists n:nat, x^n < eps)%R.
Proof.
intros eps Hpos x Hint.
elim (Rlt_dec 0 x).
intros H0ltx.
assert (Hlnxneg: (ln x < 0)%R).
rewrite <- ln_1.
apply ln_increasing; auto.
intuition.
elim (Rlt_dec 1 eps).
intros Hepslt1.
exists 1%nat; simpl; rewrite Rmult_1_r.
apply Rlt_trans with 1%R; intuition auto with real.
intros Hn1lteps.
elim (Rlt_dec eps 1).
intros Hepslt1.
elim (archimed (ln eps / ln x)).
exists (Zabs_nat (up (ln eps / ln x))).
rewrite pow_exp.
rewrite <- INR_Zabs_nat_eq.
apply Rlt_le_trans with (exp (ln x * (ln eps / ln x))).
apply exp_increasing.
apply Ropp_lt_cancel.
repeat rewrite <- Ropp_mult_distr_l_reverse.
apply Rmult_lt_compat_l.
replace 0%R with (-0)%R.
apply Ropp_lt_contravar.
rewrite <- ln_1.
apply ln_increasing.
assumption.
intuition.
ring.
auto with real.
unfold Rdiv; rewrite (Rmult_comm (ln eps));rewrite <- Rmult_assoc.
rewrite Rinv_r; auto with real.
rewrite Rmult_1_l; rewrite exp_ln; auto with real.
auto with real.
apply le_IZR.
apply Rlt_le.
apply Rle_lt_trans with (ln eps / ln x)%R; auto with real.
simpl (IZR 0).
unfold Rdiv; rewrite <- Rmult_opp_opp.
simpl (IZR 0).
apply Rmult_le_pos.
(replace 0%R with (-ln 1)%R; 
   [apply Ropp_le_contravar| rewrite ln_1; ring]).
apply Rlt_le; apply ln_increasing; intuition auto with real.
rewrite Ropp_inv_permute.
apply Rlt_le.
apply Rinv_0_lt_compat.
auto with real;fail.
auto with real.
assumption.
intros Hnepslt1.
exists 1%nat; simpl; rewrite Rmult_1_r.
assert (Heq: eps = 1%R).
apply Rle_antisym.
apply Rnot_lt_le; auto.
apply Rnot_lt_le; auto.
subst eps; intuition.

intros Hx0.
assert (Heq : x = 0%R).
apply Rle_antisym.
apply Rnot_lt_le; auto.
intuition.
subst x.
exists 1%nat; simpl; rewrite Rmult_1_r; assumption.
Qed.

Theorem pow_lt_1_decr : 
  forall x n m, (0 < x <= 1 -> (n<=m)%nat -> x ^m <= x^n)%R.
Proof.
intros x n m [H1 H2] H3.
replace m with ((m-n)+n)%nat.
replace (x  ^ n)%R with (1* x ^ n)%R.
rewrite pow_add. 
apply Rmult_le_compat_r.
apply pow_le.
auto with real.
rewrite <- (pow1 (m-n)).
apply pow_incr.
auto with real.
auto with real.
omega.
Qed.

Axiom le_inv_16_1 : (1/16 < 1)%R.

Theorem arctan_1_2_serie_bound :
  infinit_sum (fun i => (1/16)^i)%R (16/15).
Proof.
apply sum_cv_infinit_sum.
unfold Un_cv.
intros eps Heps_pos.
assert (Heps_pos2: (0<eps)%R).
fourier.
assert (Hinv16bound: (0<=1/16<1)%R).
split; [fourier| apply le_inv_16_1].
generalize (pow_lt_1 eps Heps_pos2 (1/16) Hinv16bound).
intros [N HN].
exists N.
intros n HNge0.
rewrite finite_sum_x_pow_i.
unfold R_dist.
replace ((1 - (1 / 16) ^ (n + 1)) / (1 - 1 / 16) - 16 / 15)%R with
  (-((1/16)^n/15))%R.
rewrite Rabs_Ropp.
rewrite Rabs_right.
apply Rle_lt_trans with ((1 / 16) ^ N/15)%R.
unfold Rdiv.
apply Rmult_le_compat_r.
fourier.
apply pow_lt_1_decr.
split.
fourier.
apply Rlt_le.
apply le_inv_16_1.
omega.
fourier.
unfold Rdiv.
replace 0%R with (0*/15)%R.
apply Rle_ge.
apply Rmult_le_compat_r.
fourier.
apply pow_le.
fourier.
auto with real.
rewrite pow_add.
rewrite pow_1.
newfield.
apply Rlt_not_eq.
apply le_inv_16_1.
Qed.

Theorem artan_serie_bound_1 : 
  forall n p r1 r2,
    2 <= n ->
    1 <= p -> 
    infinit_sum (fun k : nat => 
      terme_arctan_1_n_R n (Z_of_nat (k+ Zabs_nat p))) r1 -> 
    infinit_sum (fun k : nat => 
      (1/(2*16^(k+ Zabs_nat p))))%R r2 -> 
    (r1 <= r2)%R.
Proof.
intros n p r1 r2 H1 H2 H3 H4.
apply (infinit_sum_bound (fun k : nat => 
  terme_arctan_1_n_R n (Z_of_nat (k + Zabs_nat p)))
(fun k : nat => (1 / (2*16 ^ (k + Zabs_nat p)))))%R; try assumption.
intros n0.
apply Rmult_le_reg_l with (2*16 ^ (n0 + Zabs_nat p))%R.
apply Rmult_lt_0_compat.
fourier.
apply pow_lt.
fourier.
unfold Rdiv.
rewrite (Rmult_comm 1). 
rewrite Rmult_1_r.
rewrite Rinv_r.
generalize (n0 + Zabs_nat p)%nat.
intros n1.
replace (2*16^n1)%R with (IZR (2^(4*Z_of_nat n1+1))%Z)%R.
apply Rlt_le.
apply terme_atan_1_n_bound.
assumption.
omega.
rewrite Zpower_exp.
rewrite mult_IZR.
replace ( IZR (2 ^ 1)) with 2%R.
rewrite Rmult_comm.
apply Rmult_eq_compat_l.
replace (4 * Z_of_nat n1)%Z with (Z_of_nat (4 *n1)%nat)%Z.
rewrite power_IZR .
replace (IZR 2) with 2%R.
rewrite  pow_mult.
replace 16%R with (2^4)%R.
reflexivity.
unfold pow.
auto with real.
auto.
rewrite inj_mult.
auto.
auto.
omega.
omega.
apply prod_neq_R0.
auto with real1.
apply Rgt_not_eq.
assert (0 < 16 ^ (n0 + Zabs_nat p))%R.
apply pow_lt.
fourier.
fourier.
Qed.

Theorem scal_infinit_sum : 
  forall (An : nat -> Rdefinitions.R) (x r: Rdefinitions.R),
    infinit_sum An r ->
    (x <> 0)%R ->
    infinit_sum  (fun i : nat => (An i * x)%R) (x*r).
Proof.
intros An x r.
unfold infinit_sum.
intros Hinf.
intros HRabs eps Heps_pos.
assert (Heps'_pos: (eps / Rabs x>0)%R).
unfold Rdiv.
assert (0<eps* / Rabs x)%R.
apply Rlt_mult_inv_pos.
auto with real.
apply Rabs_pos_lt.
assumption.
auto with real.
generalize (Hinf (eps/ (Rabs x))%R Heps'_pos).
clear Hinf.
intros [N0 Hinf].
exists N0.
intros n HngeN0.
generalize (Hinf n HngeN0).
clear Hinf.
unfold R_dist.
unfold Rdiv.
intros Hinf.
rewrite <- scal_sum.
rewrite <- Rmult_minus_distr_l.
rewrite Rabs_mult.
rewrite Rmult_comm.
apply Rmult_lt_reg_l with (/Rabs x).
rewrite <- Rabs_Rinv.
apply Rabs_pos_lt.
auto with real.
assumption.
rewrite Rmult_comm.
rewrite Rinv_r_simpl_l.
rewrite Rmult_comm.
assumption.
apply Rabs_no_R0.
assumption.
Qed.

(*Theorem infinit_sum_pos :
  forall i An l, (forall n, 0 <= (An n))%R -> 
    infinit_sum (fun k => (An (k+i)%nat)) l -> 
    (0 <= l)%R.
Proof.
induction i; intros An l Hpos Hinf.


infinit_sum_step:
  forall (An : nat -> Rdefinitions.R) (l : Rdefinitions.R),
  infinit_sum An l ->
  infinit_sum (fun i : nat => An (i + 1)%nat) (l - An 0%nat)
*)


Theorem arctan_bound_limit : 
  forall p, 
    (1 <= p)%nat ->
    infinit_sum (fun k : nat => (1/(2*16^(k+ p))))%R (1/(30*16^(p-1)%nat))%R. 
Proof.
intros p H1lep.
apply infinit_sum_ext with 
  (fun k : nat => ((fun i : nat => ((1 /  16) ^ i)) k) * (1/(2*16^p)))%R.
intros n.
rewrite pow_add.
unfold Rdiv.
repeat rewrite Rmult_1_l.
rewrite <- Rinv_pow.
newfield.
split; apply pow_nonzero; discrR.
discrR.
replace (1/(30*16^(p-1)%nat))%R with ((1/(2*16^p))*(16/15))%R.
apply (scal_infinit_sum (fun k : nat =>  ((1 /  16) ^ k)) (1 / (2 * 16 ^ p)) (16/15))%R.
apply arctan_1_2_serie_bound.
unfold Rdiv.
apply Rmult_integral_contrapositive.
split.
discrR.
rewrite Rinv_mult_distr.
apply Rmult_integral_contrapositive.
split.
apply Rgt_not_eq.
fourier.
rewrite  Rinv_pow.
apply pow_nonzero.
apply Rgt_not_eq.
fourier.
discrR.
discrR.
apply pow_nonzero.
discrR.
replace ((2 * 16 ^ p))%R with (2 * 16 ^ ((p-1)+1))%R.
rewrite pow_add.
rewrite pow_1.
newfield.
apply pow_nonzero.
discrR.
replace (p-1+1)%nat with p.
tauto.
destruct p.
unfold minus.
simpl.
apply le_antisym.
omega.
assumption.
simpl.
ring_nat.
auto with arith.
Qed.

Theorem arctan_serie_shift_pos:
  forall n p r,
    2 <= n ->
    1 <= p -> 
    infinit_sum (fun k : nat => 
      terme_arctan_1_n_R n (Z_of_nat (k+ Zabs_nat p))) r -> 
    (0<r)%R.
Proof.
intros n p r H2len H1lep Hinf.
generalize (infinit_sum_cv _ _ Hinf).
intros Hcv.
apply Rlt_le_trans with (terme_arctan_1_n_R n p)%R.
apply terme_atan_1_n_pos.
omega.
omega.
generalize (sum_incr _ 0 r Hcv).
simpl.
replace (Z_of_nat (Zabs_nat p)) with p.
intros H; apply H; clear H.
intros n0.
apply Rlt_le.
apply terme_atan_1_n_pos.
omega.
omega.
rewrite Z_of_nat_Zabs_nat.
reflexivity.
omega.
Qed.

Theorem arctan_serie_shift_bound:
  forall n p r,
    2 <= n ->
    1 <= p -> 
    infinit_sum (fun k : nat => 
      terme_arctan_1_n_R n (Z_of_nat (k+ Zabs_nat p))) r -> 
    (r <= 1/30 *1/16^(Zabs_nat p -1))%R.
Proof.
intros n p r H2len H1lep Hinf.
apply (artan_serie_bound_1 n p r); try assumption.
replace (1 / 30 * 1 / 16 ^ (Zabs_nat p - 1))%R with
  (1 / (30 * 16 ^ (Zabs_nat p - 1)))%R.
apply (arctan_bound_limit (Zabs_nat p)).
change 1%nat with (Zabs_nat 1).
apply Zabs_nat_le; omega.
newfield.
apply pow_nonzero.
discrR.
Qed.

Lemma arctan_serie_bound :
  forall n r,
    2 <= n ->
    infinit_sum (fun k : nat => 
      terme_arctan_1_n_R n (Z_of_nat k)) r -> 
    (r <= 1)%R.
Proof.
intros n r H2len Hinf.
generalize (infinit_sum_step _ _ Hinf).
intros Hinfstep.
assert (H1le1Z : 1<=1).
omega.
generalize (arctan_serie_shift_bound n 1 (r - terme_arctan_1_n_R n (Z_of_nat 0))
  H2len H1le1Z Hinfstep).
simpl.
replace (1 / 30 * 1 / 1)%R with (/30)%R; [idtac|newfield].
intros H2.
assert (H0le0: 0<=0); [omega|idtac].
generalize (terme_atan_1_n_bound n 0 H2len H0le0).
replace (IZR (2 ^ (4 * 0 + 1))) with 2%R.
intros H3.
assert (H4: (terme_arctan_1_n_R n 0 <= 1/2)%R).
fourier.
generalize (Rplus_le_compat _ _ _ _ H2 H4).
replace (r - terme_arctan_1_n_R n 0 + terme_arctan_1_n_R n 0)%R with r.
intros.
apply Rle_trans with (/ 30 + 1 / 2)%R.
assumption.
apply Rmult_le_reg_l with 30%R.
fourier.
replace (30 * (/ 30 + 1 / 2))%R with 16%R; [idtac|newfield].
fourier.
newfield.
simpl.
reflexivity.
Qed.

Lemma  arctan_correct_ineq_aux_1:
  forall (n :Z), 2 <= n ->
    0<= 3*n^2-1 <= 3*n^3.
Proof.
intros n H.
assert (H0:0<n).
omega.
assert (H5: 0<=2).
omega.
generalize (Zpow_pos 2 n H0 H5).
intros H1.
clear H5; assert (H5: 0<=3).
omega.
generalize (Zpow_pos 3 n H0 H5).
intros H2.
clear H5.
split.
omega.
assert (H3: 0 <= n ^ 2).
omega.
generalize (Zmult_le_compat_l _ _ _ H H3 ).
replace (n^2*n)%Z with (n^3).
replace (n^2*1)%Z with (n^2).
intros H4.
omega.
omega.
replace 3 with (2+1)%Z; [idtac|trivial].
assert (H4: 1>=0); [omega|idtac].
assert (H5: 2>=0); [omega|idtac].
generalize (Zpower_exp n _ _ H5 H4).
intros H6; rewrite H6.
replace (n^1) with n.
reflexivity.
unfold Zpower.
unfold Zpower_pos.
simpl.
auto with zarith.
Qed.

Lemma  arctan_correct_ineq_aux:
  forall (n i:Z), 2 <= n -> 0 <= i ->
    0<= n^2*(4*i+3)-4*i-1 <= n^(4*i+3)*(4*i+1)*(4*i+3).
Proof.
intros n i H.
split.
assert (1 <= n^2).
change 1 with (1^2).
apply Zpower_incr.
omega.
omega.
apply Zle_trans with ((4 * i + 3) - 4 * i - 1)%Z.
simpl; omega.
assert ( 1* (4 * i + 3) <= n ^ 2 * (4 * i + 3)).
apply Zmult_le_compat_r.
assumption.
omega.
omega.
apply Zle_trans with (n^2*(4*i+3))%Z.
omega.
apply Zmult_le_compat_r; [idtac| omega].
replace (4 * i + 3)%Z with (2 + (4*i+1))%Z; [idtac| auto with zarith]. 
replace (n^2)%Z with (n^2*(1*1))%Z; [idtac| auto with zarith].
rewrite Zpower_exp; try omega.
rewrite <- Zmult_assoc.
apply Zmult_le_compat_l.
apply Zmult_le_compat; try omega.
pattern 1 at 1.
replace 1%Z with (1^(4 * i + 1))%Z.
apply Zpower_incr; omega.
apply eq_IZR.
replace (4 * i + 1)%Z with (Z_of_nat (Zabs_nat (4 * i + 1)))%Z.
rewrite power_IZR.
change (IZR 1)%Z with 1%R.
apply pow1.
apply Z_of_nat_Zabs_nat.
omega.
change 0 with (0^2).
apply Zpower_incr.
omega.
omega.
Qed.

Ltac apply_arctan_1_n_correct1' arctan_1_n' :=
 match goal with 
   arctan_1_n_correct1 : _, id : represents ?v ?vr, id1 : infinit_sum _ ?r
   |- represents (arctan_1_n' ?v (?y, _, _, _)) ?w =>
      replace w with (vr + IZR y * r)%R;
        [ apply arctan_1_n_correct1;clear arctan_1_n_correct1;
	  auto with real1 | clear arctan_1_n_correct1]
 end.

Lemma technical_bound_lemma : forall y r n i log_y,
  log_y <= 4 * i - 3 ->
  y = 2 ^ log_y ->
  1 <= y -> 
  1 <= i -> 
  2 <= n -> 
  infinit_sum (fun k : nat =>
    terme_arctan_1_n_R n (Z_of_nat (k + Zabs_nat i))) r -> 
  (0<2*(IZR y * r) <= 1/4)%R.
intros y r n i log_y Hlelog Hlog H1ley H1lei H2len Hinf.
split.

apply Rmult_lt_0_compat.
fourier.
apply Rmult_lt_0_compat.
apply (IZR_lt 0).
omega.
apply (arctan_serie_shift_pos _ _ _ H2len H1lei Hinf).

assert (H0lty : (0 <IZR y)%R).
apply (IZR_lt 0).
omega.
replace (1/4)%R with (2*(IZR y) *(1/8*1/IZR y))%R; [idtac|newfield].
rewrite <- Rmult_assoc.
apply Rmult_le_compat_l.
rewrite <- (mult_IZR 2).
apply (IZR_le 0).
omega.
apply Rle_trans with (1/30 *1/16^(Zabs_nat i -1))%R.
apply (arctan_serie_shift_bound _ _ _ H2len H1lei Hinf).
unfold Rdiv.
repeat rewrite (Rmult_comm 1).
repeat rewrite Rmult_1_r; try discrR.
repeat rewrite <- Rinv_mult_distr; try discrR.
apply Rlt_le.
apply Rinv_lt_contravar.
repeat apply Rmult_lt_0_compat ; try fourier.
apply pow_lt.
fourier.
subst y.
assert (0<=log_y).
apply (Zpower_aux _ H1ley).
rewrite <- (Z_of_nat_Zabs_nat log_y); [idtac| assumption].
rewrite power_IZR.
change (IZR 2) with 2%R.
apply Rle_lt_trans with (2 ^ (4* Zabs_nat i))%R.
replace 8%R with (2^3)%R.
rewrite <- pow_add.
apply Rle_pow.
fourier.
rewrite <- (Zabs_nat_plus 3); try omega.
rewrite <- (Zabs_nat_mult 4).
apply Zabs_nat_le.
split; omega.
unfold pow.
newfield.
rewrite pow_mult.
replace (2^4)%R with 16%R; [idtac | unfold pow; newfield].
apply Rle_lt_trans with (16 ^1 * 16 ^ (Zabs_nat i - 1))%R.
rewrite <- pow_add.
replace (1 + (Zabs_nat i - 1))%nat with (Zabs_nat i).
fourier.
symmetry.
apply le_plus_minus_r.
apply (Zabs_nat_le 1).
omega.
apply Rmult_lt_compat_r.
apply pow_lt; fourier.
unfold pow.
fourier.
generalize (lt_IZR 0 y H0lty).
intros; omega.
assert (0 < 16 ^ (Zabs_nat i - 1))%R; [idtac | auto with real].
apply pow_lt; fourier.
auto with real.
Qed.

Theorem arctan_1_n_series_body_correct :
  forall (arctan_1_n_series : stream idigit -> Z*Z*Z*Z -> stream idigit),
    (  forall v vr r y i n log_y, 
      y = 2^log_y ->
      2<= n ->
      1 <= y ->
      1 <= i -> 
      log_y-1 <= 4*i -3 ->
      represents v vr ->
      infinit_sum (fun k => terme_arctan_1_n_R n (Z_of_nat (k + Zabs_nat i))) r ->
      (vr + (IZR y)*r <= 1)%R ->
      represents (arctan_1_n_series v (y, i, n, log_y)) (vr+(IZR y)*r)%R) ->
    forall v vr r y i n log_y,
      y = 2^log_y ->      
      1 <= y ->
      1 <= i -> 
      2 <= n ->
      log_y <= 4*i -3 ->
      represents v vr ->
      infinit_sum (fun k => terme_arctan_1_n_R n (Z_of_nat (k + Zabs_nat i))) r ->
      (vr + (IZR y)*r <= 1)%R ->
      represents (series_body _ arctan_1_n_series v ((2*y)%Z, i, n, (log_y +1)%Z))
        (vr+(IZR y)*r)%R.
Proof.
intros arctan_1_n_aux' arctan_1_n_aux_correct1;
intros v vr r y i n log_y;
intros Hlog_y H1ley H1lei H2len Hlogyle Hrep Hinf Hsumle1.
generalize (technical_bound_lemma _ _ _ _ _ Hlogyle Hlog_y H1ley H1lei H2len Hinf).
intros [Hnpros Hboundr].

unfold series_body.

destruct v as [[ | | ] v1]; inversion Hrep.

destruct v1 as [[ | | ] v2];
inversion_to_variable v2;
add_fact v2.

(*L::L*)
constructor' arctan_1_n_aux_correct1;[idtac | clear  arctan_1_n_aux_correct1].
apply_arctan_1_n_correct1' arctan_1_n_aux'.
subst y.
rewrite Zpower_exp. 
rewrite Zmult_comm.
change (2^1)%Z with 2%Z. 
reflexivity.
apply Zle_ge.
apply Zpower_aux; assumption; omega.
omega.
omega.
omega.
dec_tac.
rewrite mult_IZR; change (IZR 2) with 2%R; rewrite Rmult_assoc.
fourier.
rewrite Rmult_plus_distr_l.
replace (2 * (r0 / 2))%R with r0; [idtac| newfield].
rewrite mult_IZR; change (IZR 2) with 2%R; rewrite Rmult_assoc.
reflexivity.
rewrite Rmult_plus_distr_l.
replace (2 * (r0 / 2))%R with r0; [idtac| newfield].
dec_tac.
split; fourier.

(*L::R*)
constructor' arctan_1_n_aux_correct1;[idtac | clear  arctan_1_n_aux_correct1].
apply_arctan_1_n_correct1' arctan_1_n_aux'.
subst y.
rewrite Zpower_exp. 
rewrite Zmult_comm.
change (2^1)%Z with 2%Z. 
reflexivity.
apply Zle_ge.
apply Zpower_aux; assumption; omega.
omega.
omega.
omega.
rewrite mult_IZR; change (IZR 2) with 2%R; rewrite Rmult_assoc.
dec_tac.
fourier.
rewrite Rmult_plus_distr_l.
rewrite mult_IZR; change (IZR 2) with 2%R; rewrite Rmult_assoc.
newfield.
replace (2 * ((r1 + 1) / 2 / 2 + IZR y * r) - 1 / 2)%R with 
  (r1/2 + 2*IZR (y) *r)%R; [idtac|newfield].
rewrite Rmult_assoc.
dec_tac.
split; fourier.

(*L::C*)
constructor' arctan_1_n_aux_correct1;[idtac | clear  arctan_1_n_aux_correct1].
apply_arctan_1_n_correct1' arctan_1_n_aux'.
subst y.
rewrite Zpower_exp. 
rewrite Zmult_comm.
change (2^1)%Z with 2%Z. 
reflexivity.
apply Zle_ge.
apply Zpower_aux; assumption; omega.
omega.
omega.
omega.
dec_tac.
rewrite mult_IZR; change (IZR 2) with 2%R; rewrite Rmult_assoc.
fourier.
rewrite mult_IZR; change (IZR 2) with 2%R; rewrite Rmult_assoc.
newfield.
(replace (2 * (r0 / 2 + IZR y * r))%R with 
  (r0 + 2*IZR (y) *r)%R; [idtac|newfield]).
rewrite Rmult_assoc.
dec_tac.
split; fourier.

(*R*)
inversion_to_variable v1;
add_fact v1.
constructor' arctan_1_n_aux_correct1;[idtac | clear  arctan_1_n_aux_correct1].
apply_arctan_1_n_correct1' arctan_1_n_aux'.
subst y.
rewrite Zpower_exp. 
rewrite Zmult_comm.
change (2^1)%Z with 2%Z. 
reflexivity.
apply Zle_ge.
apply Zpower_aux; assumption; omega.
omega.
omega.
omega.
rewrite mult_IZR; change (IZR 2) with 2%R; rewrite Rmult_assoc.
dec_tac.
replace r0 with (2*vr -1)%R; [idtac| subst vr; newfield].
fourier.
rewrite mult_IZR; change (IZR 2) with 2%R; rewrite Rmult_assoc.
newfield.
split.
replace (2 * ((r0 + 1) / 2 + IZR y * r) - 1)%R with 
  (r0 + 2 * IZR y * r)%R; [idtac|newfield].
rewrite Rmult_assoc.
dec_tac; fourier.
rewrite H2.
rewrite Rmult_plus_distr_l.
fourier.

destruct v1 as [[ | | ] v2];
inversion_to_variable v2;
add_fact v2.

(*C::L*)
constructor' arctan_1_n_aux_correct1;[idtac | clear  arctan_1_n_aux_correct1].
apply_arctan_1_n_correct1' arctan_1_n_aux'.
subst y.
rewrite Zpower_exp. 
rewrite Zmult_comm.
change (2^1)%Z with 2%Z. 
reflexivity.
apply Zle_ge.
apply Zpower_aux; assumption; omega.
omega.
omega.
omega.
dec_tac.
rewrite mult_IZR; change (IZR 2) with 2%R; rewrite Rmult_assoc.
fourier.
rewrite mult_IZR; change (IZR 2) with 2%R; rewrite Rmult_assoc.
newfield.
replace (2 * ((2 * r0 + 1) / 4 + IZR y * r) - 1 / 2)%R with 
  (r0 + 2*IZR y *r)%R; [idtac|newfield].
rewrite Rmult_assoc.
dec_tac; split; fourier.

(*C::R*)
constructor' arctan_1_n_aux_correct1;[idtac | clear  arctan_1_n_aux_correct1].
apply_arctan_1_n_correct1' arctan_1_n_aux'.
subst y.
rewrite Zpower_exp. 
rewrite Zmult_comm.
change (2^1)%Z with 2%Z. 
reflexivity.
apply Zle_ge.
apply Zpower_aux; assumption; omega.
omega.
omega.
omega.
rewrite mult_IZR; change (IZR 2) with 2%R; rewrite Rmult_assoc.
dec_tac; fourier.
rewrite mult_IZR; change (IZR 2) with 2%R; rewrite Rmult_assoc.
newfield.
replace (2 * ((2 * ((r1 + 1) / 2) + 1) / 4 + IZR y * r) - 1)%R with 
  (r1/2 + 2*IZR (y) *r)%R; [idtac|newfield].
dec_tac.
rewrite Rmult_assoc.
split; fourier.

(*C::C*)
constructor' arctan_1_n_aux_correct1;[idtac | clear  arctan_1_n_aux_correct1].
apply_arctan_1_n_correct1' arctan_1_n_aux'.
subst y.
rewrite Zpower_exp. 
rewrite Zmult_comm.
change (2^1)%Z with 2%Z. 
reflexivity.
apply Zle_ge.
apply Zpower_aux; assumption; omega.
omega.
omega.
omega.
dec_tac.
rewrite mult_IZR; change (IZR 2) with 2%R; rewrite Rmult_assoc.
fourier.
rewrite mult_IZR; change (IZR 2) with 2%R; rewrite Rmult_assoc.
newfield.
replace (2 * ((2 * r0 + 1) / 4 + IZR y * r) - 1 / 2)%R with
  (r0 +  2*IZR (y) *r)%R; [idtac|newfield].
rewrite Rmult_assoc.
dec_tac.
split; fourier.
Defined.

Lemma  arctan_step_aux : 
  forall r i n,
    1<=i -> 
    infinit_sum 
    (fun k : nat => terme_arctan_1_n_R n (Z_of_nat (k + Zabs_nat i))) r -> 
    infinit_sum
    (fun k : nat => terme_arctan_1_n_R n (Z_of_nat (k + Zabs_nat (i + 1))))
    (r - terme_arctan_1_n_R n i).
Proof.
intros.
generalize (infinit_sum_step _ _ H0).
simpl.
rewrite Z_of_nat_Zabs_nat.
intros H1.
apply (infinit_sum_ext 
  (fun i0 : nat => terme_arctan_1_n_R n (Z_of_nat (i0 + 1 + Zabs_nat i)))
  (fun k : nat => terme_arctan_1_n_R n (Z_of_nat (k + Zabs_nat (i + 1))))
  (r - terme_arctan_1_n_R n i)).
intros p.
rewrite (Zabs_nat_plus i 1).
replace (Zabs_nat 1) with 1%nat.
rewrite (plus_comm (Zabs_nat i)).
rewrite plus_assoc.
reflexivity.
auto.
omega.
omega.
assumption.
omega.
Qed.

Theorem technical_lemma_fourier : forall a r, (0 < r - a -> a <= r)%R.
intros; fourier.
Qed.

Theorem lemma_inf_step : 
  forall i n r,
    1<=i ->
    infinit_sum
    (fun k : nat => terme_arctan_1_n_R n (Z_of_nat (k + Zabs_nat i))) r -> 
    infinit_sum
    (fun k : nat => terme_arctan_1_n_R n (Z_of_nat (k + Zabs_nat (i + 1))))
    (r - terme_arctan_1_n_R n i).
intros i n r H1lei Hinf.
generalize (infinit_sum_step _ _ Hinf).
simpl.
rewrite Z_of_nat_Zabs_nat.
intros.
apply (infinit_sum_ext 
  (fun i0 : nat => terme_arctan_1_n_R n (Z_of_nat (i0 + 1 + Zabs_nat i)))
  (fun k : nat => terme_arctan_1_n_R n (Z_of_nat (k + Zabs_nat (i + 1))))
  (r - terme_arctan_1_n_R n i)).
intros p.
rewrite (Zabs_nat_plus i 1).
replace (Zabs_nat 1) with 1%nat.
rewrite (plus_comm (Zabs_nat i)).
rewrite plus_assoc.
reflexivity.
auto.
omega.
omega.
assumption.
omega.
Qed.

Theorem technical_lemma_1 : 
  forall x y , 
    (x - 1 <= 4*y-3 -> 
      x > 4*y-3 -> 
      x = 4*y-2)%Z.
intros.
apply Zle_antisym; omega.
Qed.

Theorem technical_lemma_2 :
  forall n i,
    2 <= n -> 
    1 <= i ->
    1 <= 2 ^ (4 * i - 2) ->
    2 ^ (4 * i - 2) * (n ^ 2 * (4 * i + 3) - 4 * i - 1) <=
    n ^ (4 * i + 3) * (4 * i + 1) * (4 * i + 3).
Proof.
intros n i H2len H1lei H1lepow.
apply Zle_trans with (2 ^ (4 * i - 2) * (n ^ 2 * (4 * i + 3)))%Z.
apply Zmult_le_compat_l; omega.
rewrite Zmult_assoc.
apply Zmult_le_compat_r; [idtac| omega].
apply Zle_trans with (n ^ (4 * i - 2) * n ^ 2)%Z.
apply Zmult_le_compat_r.
apply Zpower_incr; omega.
change 0%Z with (0^2)%Z.
apply Zpower_incr; omega.
rewrite <- Zpower_exp; try omega.
replace (4*i-2+2)%Z with (4*i)%Z; [idtac| auto with zarith].
replace (n^(4*i))%Z with (n^(4*i)*1*1)%Z; [idtac| auto with zarith].
rewrite Zpower_exp; try omega.
repeat rewrite <- Zmult_assoc .
apply Zmult_le_compat_l.
apply Zmult_le_compat; try omega.
change 1%Z with (1^3).
apply Zpower_incr; omega.
apply Zpower_le_0_compat; omega.
Qed.

Theorem arctan_1_n_aux_correct :
  forall v vr r y i n log_y, 
    y = 2^log_y ->
    2 <=n ->
    1 <= y ->
    1 <= i -> 
    log_y - 1 <= 4*i -3 ->
    represents v vr ->
    infinit_sum (fun k => terme_arctan_1_n_R n (Z_of_nat(k+ (Zabs_nat i)))) r ->
    (vr + (IZR y)*r <= 1)%R ->
    represents (arctan_1_n_aux v (y, i, n, log_y)) (vr+(IZR y)*r)%R.
Proof.
cofix.
intros v vr r y i n log_y Hlogy H2len H1ley H1lei Hlogle Hrep Hinf.
str_unfold (arctan_1_n_aux v (y, i, n, log_y)).
unfold str_decompose.
lazy zeta beta iota delta [arctan_1_n_aux].
replace (cofix arctan_1_n_aux (v0 : stream idigit) (s : Z * Z * Z * Z) :
              stream idigit :=
              let (p, a) := s in
              let (p0, n0) := p in
              let (y0, i0) := p0 in
              if Z_le_gt_dec a (4 * i0 - 3)
              then
               series_body (Z * Z * Z * Z) arctan_1_n_aux v0
                 (2 * y0, i0, n0, a + 1)
              else
               series_body (Z * Z * Z * Z) arctan_1_n_aux
                 (v0 +
                  rat_to_stream (y0 * (n0 ^ 2 * (4 * i0 + 3) - 4 * i0 - 1))
                    (n0 ^ (4 * i0 + 3) * (4 * i0 + 1) * (4 * i0 + 3)))%S
                 (2 * y0, i0 + 1, n0, a + 1))%Z with
  arctan_1_n_aux; [idtac| apply refl_equal].
elim (Z_le_gt_dec (log_y) (4*i-3)).
intros Hle.
fold (str_decompose (series_body (Z*Z*Z*Z) arctan_1_n_aux v (2*y,i,n, log_y+1)%Z)).
rewrite <- str_dec_thm.
exact (arctan_1_n_series_body_correct arctan_1_n_aux arctan_1_n_aux_correct 
  v vr r y i n log_y Hlogy H1ley H1lei H2len Hle Hrep Hinf).

intros Hgt.
fold (str_decompose (series_body (Z*Z*Z*Z) arctan_1_n_aux 
  (v + rat_to_stream (y * (n ^ 2 * (4 * i + 3) - 4 * i - 1))
    (n ^ (4 * i + 3) * (4 * i + 1) * (4 * i + 3)))%S
  (2 * y, i + 1, n, log_y + 1)))%Z.
rewrite <- str_dec_thm.

assert (H7 : (vr + IZR y * r = (vr + IZR y * (terme_arctan_1_n_R n i)) + 
  (IZR y) * (r - (terme_arctan_1_n_R n i)))%R).
clear arctan_1_n_aux_correct; field.
rewrite H7.

generalize (lemma_inf_step _ _ _ H1lei Hinf).
intros Hinf' H1.
rewrite <- H7 in H1.

match goal with 
  |- represents (series_body _ arctan_1_n_aux ?v _ ) (?vr + (IZR y)*?r) =>
    apply (arctan_1_n_series_body_correct arctan_1_n_aux arctan_1_n_aux_correct
      v vr r y (i+1) n log_y) end; 
clear arctan_1_n_aux_correct; try assumption; try omega.

apply add_correct.
assumption.
unfold terme_arctan_1_n_R.
match goal with |- represents (rat_to_stream ?a ?b) ?c => 
  replace c with (IZR a / IZR b)%R end.
apply rat_to_stream_correct.

split.
apply Zmult_le_0_compat.
omega.
assert (H0lei: 0<=i).
omega.
generalize (arctan_correct_ineq_aux n i H2len H0lei).
intros [H8 H9]; exact H8.

subst y.
generalize (technical_lemma_1 _ _ Hlogle Hgt).
intros H10.
subst log_y.

apply technical_lemma_2; assumption.

repeat apply Zmult_lt_O_compat; try omega.
apply Zle_neq_lt.
assert (n ^ (4 * i + 3) <> 0).
apply Zpower_not_0; omega.
auto.
apply Zpower_le_0_compat; omega.

unfold Rdiv.
repeat rewrite mult_IZR.
apply Rmult_assoc.
apply Rle_trans with (vr + IZR y * r)%R; [idtac| assumption].
apply Rplus_le_compat_l.
apply Rmult_le_compat_l.
apply (IZR_le 0).
omega.
assert (0 < r - (terme_arctan_1_n_R n i))%R.
assert (H8: (1<=i+1)).
omega.
generalize (arctan_serie_shift_pos n (i+1) _ H2len H8 Hinf').
tauto.
apply technical_lemma_fourier.
assumption.
rewrite <- H7; assumption.
Qed.

Theorem arctan_correct :
  forall r n,
    2 <= n ->
  infinit_sum (fun k => terme_arctan_1_n_R n (Z_of_nat k)) r ->
  represents (arctan_1_n n) r.
Proof.
intros r n H2len H.
unfold arctan_1_n.
(**)
assert (H1: (r = (terme_arctan_1_n_R n 0 + 
  (IZR 1)*(r- terme_arctan_1_n_R n 0)))%R).
field.
rewrite H1.
apply  arctan_1_n_aux_correct; try omega.
simpl; trivial.
unfold terme_arctan_1_n_R.
change (4*0+3)%Z with 3.
change (4*0+1)%Z with 1.
replace (n^3*1*3)%Z with (3*n^3)%Z; [idtac|ring].
replace (n ^ 2 * 3 - 4 * 0 - 1)%Z with (3*n^2-1)%Z; [idtac|ring].
apply rat_to_stream_correct.
apply (arctan_correct_ineq_aux_1 _ H2len).
assert (H': 0<n^3).
apply Zpow_pos; omega.
omega.
apply (infinit_sum_step _ _ H).
rewrite <- H1; clear H1.
apply arctan_serie_bound with n; assumption.
Qed.

