Require Export realbasis.
Require Export proof_utils.
Require Export real_facts.

Lemma sub_aux_stable_aux : 
  forall u v, (u <= 1 -> 0 <= v -> (u+1-v)/2 <= 1)%R. 
Proof.
intros u v H1 H2.
replace ((u+1-v)/2)%R with (u/2+(1-v)/2)%R.
apply Rle_1_split.
assumption.
auto with real1.
field.
discrR.
Qed.

Theorem Rsub_1_stable : forall x, (1-x <=1 -> 0<=x)%R.
intros.
auto with real1.
Qed.

Lemma sub_aux_stable : 
  forall u v, (0<=u <= 1 -> 0 <= v <=1-> 0 <= (u+1-v)/2 <= 1)%R. 
Proof.
intros u v; inversion 1; inversion 1.
split.
apply Rsub_1_stable.
replace (1 - (u + 1 - v) / 2)%R with ((v+1-u)/2)%R.
apply sub_aux_stable_aux; assumption.
field; discrR.
apply sub_aux_stable_aux; assumption.
Qed.

(*en fonction du bit d remplace b par ce qu'il faut en fonction de u
  pour que le constructeur de represents trouve ce qu'il faut
*)
Ltac represents_replace :=
  match goal with id : represents ?s ?u
      |- represents (?d::?s) ?b =>
	replace b with 
	  match d with 
	    | L => (u/2)%R
	    | C => ((2*u+1)/4 )%R
	    | R => ((u+1)/2)%R
	  end; 
	  [constructor; [assumption| split; try auto with real1] | field; discrR]
  end.
(*on veut montrer qu'un stream represente un reel et on connait ses 2 premiers digits
  on a besoin d'une hypothese ou on connnait la representation du stream 
  sans le premier digit
*)
Ltac rep_prepare_hyp := 
  match goal with 
    | id1 : represents ?s ?u, id2 : (0 <= ?u <=1)%R
      |- represents (?d1::?d2::?s) ?b =>
	match d2 with 
	  | L => generalize (reprL s u id1 id2)
	  | C => generalize (reprC s u id1 id2)
	  | R => generalize (reprR s u id1 id2)
 	end; intros
  end.

Theorem impl_CL_LR :
  forall s u , represents (C::L::s) u -> represents (L::R::s) u.
Proof.
  intros s u H;  inversion H.
  inversion H1.
  inversion H6.
  rep_prepare_hyp; represents_replace.
  replace ((r0 + 1) / 2)%R with (r0/2+1/2)%R.
auto with real1.
field; discrR.
Qed.

Theorem impl_CR_RL : 
  forall s u , represents (C::R::s) u -> represents (R::L::s) u.
Proof.
  intros s u H;  inversion H.
  inversion H1.
  inversion H6.
  rep_prepare_hyp; represents_replace.
Qed.

Theorem impl_LR_CL : 
  forall s u , represents (L::R::s) u -> represents (C::L::s) u.
Proof.
  intros s u H; inversion H.
  inversion H1.
  inversion H6.
  rep_prepare_hyp; represents_replace.
Qed.

Theorem impl_RL_CR : 
  forall s u , represents (R::L::s) u -> represents (C::R::s) u.
Proof.
  intros s u H; inversion H.
  inversion H1.
  inversion H6.
  rep_prepare_hyp; represents_replace.
  replace ((r0 + 1) / 2)%R with (r0/2+1/2)%R.
auto with real1.
field; discrR.
Qed.

Theorem decalage_C_L : 
  forall d s u, represents (d::C::s) u -> represents (d::L::s) (u-1/8).
Proof.
  intros d; case d; intros s u H; inversion H;
  inversion H1;  inversion H6;
  rep_prepare_hyp;  represents_replace.
Qed.

Theorem decalage_C_R : 
  forall d s u, represents (d::C::s) u -> represents (d::R::s) (u+1/8).
Proof.
  intros d; case d; intros s u H;
  inversion H; inversion H1;  inversion H6;
  rep_prepare_hyp; represents_replace;
  (replace ((r0 + 1) / 2)%R with (r0/2+1/2)%R;
    [auto with real1 |field; discrR]).
Qed.

Theorem decalage_R_C : 
  forall d s u, represents (d::R::s) u -> represents (d::C::s) (u-1/8).
Proof.
  intros d; case d; intros s u H; inversion H;
  inversion H1;  inversion H6;
  rep_prepare_hyp; represents_replace;
  (replace ((2*r0 + 1) / 4)%R with (r0/2+1/4)%R;
    [auto with real1 |field; discrR]).
Qed.

Theorem decalage_L_C : 
  forall d s u, represents (d::L::s) u -> represents (d::C::s) (u+1/8).
Proof.
  intros d; case d; intros s u H; inversion H;
  inversion H1; inversion H6;
  rep_prepare_hyp; represents_replace;
  (replace ((2*r0 + 1) / 4)%R with (r0/2+1/4)%R;
    [auto with real1 |field; discrR]).
Qed.

(*on veut fabriquer une hypothese du type represents (d::x) u
  a partir des autres hypotheses et des equivalences entre prefixe
*)
Ltac find_represents_eq x :=
  match x with 
    |L::?x' => 
      match goal with 
	|id : represents (L::R::x') _ |- _ => 
	  generalize  (impl_LR_CL _ _ id); clear id; inversion 1
	|id : represents (C::R::x') _ |- _ => 
	  generalize  (impl_CR_RL _ _ id); clear id; inversion 1
	| id : represents (_::C::x') _ |- _ => 
	  generalize  (decalage_C_L _ _ _ id); clear id; inversion 1
	| _ => idtac "find_represents_eq L"
      end
    |C::?x' => 
      match goal with 
	|id : represents (_::L::x') _ |- _ => 
	  generalize  (decalage_L_C _ _ _ id); clear id; inversion 1
	|id : represents (_::R::x') _ |- _ => 
	  generalize  (decalage_R_C _ _ _ id); clear id; inversion 1
	| _ =>idtac "find_represents_eq C"
      end     
    |R::?x' => 
      match goal with 
	|id : represents (R::L::x') _ |- _ => 
	  generalize  (impl_RL_CR _ _ id); clear id; inversion 1
	|id : represents (C::L::x') _ |- _ => 
	  generalize  (impl_CL_LR _ _ id); clear id; inversion 1
	| id : represents (_::C::x') _ |- _ => 
	  generalize  (decalage_C_R _ _ _ id); clear id; inversion 1
	| _ => idtac "find_represents_eq R"
      end
  end.

Ltac find_represents x :=
 match goal with 
   |id : represents x _ |- _ => idtac
   |id : represents (_::x) _ |- _ => inversion id; clear id
   |_ => find_represents_eq x 
 end.


(*essaye de fabriquer les hypotheses qui donnent les valeurs representees par 
  les termes sur lesquels on veut appliquer sub_aux 
  si on connait le premier digit d'un argument, on va essayer si necessaire 
  d'utiliser l'equivalence de certains prefixes
*)
Ltac prepare_hyp := 
  match goal with 
    | |- represents (_::sub_aux (?d1::?x) (?d2::?y)) _ =>
      find_represents_eq d1 x;
      find_represents_eq d2 y
    | |- represents (_::sub_aux (?d1::?x) ?y) _ =>
      find_represents_eq d1 x;
      find_represents y
    | |- represents (_::sub_aux ?x (?d2::?y)) _ =>
      find_represents x;
      find_represents_eq d2 y
    | |- represents (_::sub_aux ?x ?y) _ => 
      find_represents x;
      find_represents y
    | _ => idtac "prepare_hyp"
  end.
    
(*on va essayer de transformer b dans le but de type represents (d::sub_aux x y) b
  en connaissant les valeurs representes par x et y, de telle sorte que l'on puisse 
  utiliser le bon constructeur de represents pour simplifier et qu'on puisse ensuite
  appliquer l'hypothese de coinduction H fournie.
  on fait la preuve du replace et de l'inegalite engendree par constructor
 *)
Ltac simpl_represent_sub_aux H :=
match goal with 
  id1 : represents ?x ?u, id2 : represents ?y ?v
  |- represents (?d::sub_aux ?x ?y) ?b => 
    replace b with 
      match d with 
	| L => (((u+1-v)/2)/2)%R 
	| C => ((2*((u+1-v)/2)+1)/4)%R
	| R => ((((u+1-v)/2)+1)/2)%R
      end;
      [constructor; [apply H; assumption| apply sub_aux_stable;assumption ]| field; discrR]
end.

Lemma mult2_sub1_correct : 
  forall x u, (1/2 <= u)%R -> represents x u -> 
    represents (mult2_sub_one x) (2*u-1)%R.
Proof.
cofix .
intros x u H1 H2.
str_unfold (mult2_sub_one x).
generalize H1. clear H1. case H2;clear H2 x u;
intros x r H0 H1 H2.

simpl.
replace r with 1%R.
replace (2*(1/2)-1)%R with (0/2)%R.
constructor.
apply represents_zero.
split; auto with real.
field; discrR.
case H1.
assert (H3: (0<=2)%R).
auto with real.
generalize (Rmult_le_compat_l 2 (1/2) (r/2) H3 H2)%R.
replace (2*(1/2))%R with 1%R; [idtac|field; discrR].
replace (2*(r/2))%R with r%R; [idtac|field; discrR].
intros.
apply Rle_antisym; assumption.

clear mult2_sub1_correct;simpl.
replace (2 * ((r + 1) / 2) - 1)%R with r; [idtac|field; discrR].
generalize H0; case x; tauto.

simpl in *.
replace  (2 * ((2 * r + 1) / 4) - 1)%R with ((2*r-1)/2)%R; [idtac|field; discrR].
assert (r1_2: (1/2 <=r)%R).
clear mult2_sub1_correct x H0.
assert (0 <= 4)%R .
auto with real1.
generalize (Rmult_le_compat_l 4 (1/2) ((2*r+1)/4)  H H2)%R.
replace (4*(1/2))%R with 2%R; [idtac|field; discrR].
replace (4*((2*r+1)/4))%R with (2*r+1)%R; [idtac|field; discrR].
intros H6.
fourier.
constructor.
apply mult2_sub1_correct;clear mult2_sub1_correct; assumption.
clear mult2_sub1_correct.
split.
replace 0%R with (1+(-1))%R;[idtac|field; discrR].
apply (Rplus_le_compat_r (-1) 1 (2*r))%R.
fourier.
destruct H1 as [H3 H4].
assert (H5: (0<=2)%R); auto with real. 
generalize (Rmult_le_compat_l 2 _ _ H5 H4).
intros; fourier.
Qed.


(*on a une hypothese qui nous donne un terme equivalent a "a +/- 1/8",
  on va faire le replace qu'il faut pour pouvoir faire un subst a ensuite
*)
Ltac subst_aux a :=
  match goal with 
    | id : ?b = (a + 1/8)%R |- _ => 
      assert (b-1/8=a)%R; [rewrite id; field; discrR | subst a]
    | id : ?b = (a - 1/8)%R |- _ => 
      assert (b+1/8=a)%R; [rewrite id; field; discrR | subst a]
  end.	

Theorem subs_aux_correct :
  forall x y u v, represents x u -> represents y v -> 
    represents (sub_aux x y) ((u + 1 - v)/2).
Proof.
cofix.
intros x y u v.
str_unfold (sub_aux x y).
(*on enleve les cas simples ou un bit de chaque cote suffit*)
destruct x as [ [ | | ] x'] ; destruct y as [ [ | | ] y']; 
try (simpl; prepare_hyp; simpl_represent_sub_aux subs_aux_correct);
(*on enleve ensuite ceux ou le 2e bit de y est necesaire*)
destruct y' as [ [ | | ] y''];
intros;
try(simpl; prepare_hyp; simpl_represent_sub_aux subs_aux_correct);
(*on enleve enfin ceux ou le 2e bit de x est necesaire*)
destruct x' as [ [ | | ] x''];
try(simpl; prepare_hyp; simpl_represent_sub_aux subs_aux_correct);
(*pour les 8 cas pathetiques ou on a fait des decalages
  il reste a demontrer la validite d'un replace que field n'a pas su demontrer
*)
subst_aux v;
subst_aux u;
field; discrR.
Qed.

Theorem sub_correct : 
  forall x y u v, represents x u -> represents y v -> (0 <= u - v)%R -> 
    represents (x - y)%S (u - v)%R.
Proof.
intros x y u v H1 H2 H3.
unfold sub.
replace (u-v)%R with  (2*((u+1-v)/2)-1)%R;[idtac| field; discrR].
apply mult2_sub1_correct.
replace (1/2)%R with (1/2+0)%R; [idtac| field; discrR]. 
replace ((u+1-v)/2)%R with (1/2+(u-v)/2)%R; [idtac| field; discrR].
apply Rplus_le_compat_l.
auto with real1.
apply subs_aux_correct; assumption.
Qed.
