Require Export realbasis.
Require Export proof_utils.
Require Export real_facts.
Require Export add_correct.
Require Export sub_correct.

Theorem techbound:
  forall b k r, 
    0 < b ->
    -2 * b+1 <= k <= 2 * b-1 ->
    (-1 <= r <= 1 ->
      -1 <= IZR k * r / IZR (2 * b) <= 1)%R.
Proof.  
  intros b k r Hbpos Hkbound Hrbound.
  elim (total_order_T r 0).
  intros H; elim H; clear H.
  intros Hrneg.

  split.
  apply Rle_trans with r%R.
  dec_tac; assumption.
  replace (IZR k * r)%R with ((-r) * IZR (-k))%R; 
    [idtac| rewrite Ropp_Ropp_IZR; ring].
  unfold Rdiv; rewrite Rmult_assoc.
  pattern r at 1; replace r with (-r*-1)%R; [idtac| ring].
  apply Rmult_le_compat_l.
  fourier.
  apply Rmult_le_reg_l with (IZR (2*b)); [apply (IZR_lt 0); omega| idtac].
  rewrite <- Rmult_assoc.
  rewrite Rinv_r_simpl_m; [idtac|discrR; omega].
  change (-1)%R with (IZR (-1)).
  rewrite <- mult_IZR.
  apply IZR_le; omega.
  
  apply Rle_trans with (-r*1)%R.  
  replace (IZR k * r)%R with ((-r) * IZR (-k))%R; 
    [idtac| rewrite Ropp_Ropp_IZR; ring].
  unfold Rdiv; rewrite Rmult_assoc.
  apply Rmult_le_compat_l.
  fourier.
  apply Rmult_le_reg_l with (IZR (2*b)); [apply (IZR_lt 0); omega| idtac].
  rewrite <- Rmult_assoc.
  rewrite Rinv_r_simpl_m; [idtac|discrR; omega].
  rewrite Rmult_1_r.
  apply IZR_le; omega.
  apply Ropp_le_cancel.
  replace (- (- r * 1))%R with r; [idtac|ring].
  dec_tac;assumption.

  intros; subst.
  unfold Rdiv; rewrite Rmult_0_r; rewrite Rmult_0_l; split; fourier.

  intros Hrpos; split.
  apply Rle_trans with (r*-1)%R.
  replace (r * -1)%R with (-r)%R; [idtac|ring].  
  apply Ropp_le_contravar; dec_tac; assumption.
  rewrite (Rmult_comm (IZR k)).
  unfold Rdiv; rewrite Rmult_assoc.
  apply Rmult_le_compat_l.
  fourier.
  apply Rmult_le_reg_l with (IZR (2*b)); [apply (IZR_lt 0); omega| idtac].
  rewrite <- Rmult_assoc.
  rewrite Rinv_r_simpl_m; [idtac|discrR; omega].
  change (-1)%R with (IZR (-1)).
  rewrite <- mult_IZR.
  apply IZR_le; omega.
  
  apply Rle_trans with (r*1)%R.  
  rewrite (Rmult_comm (IZR k)).
  unfold Rdiv; rewrite Rmult_assoc.
  apply Rmult_le_compat_l.
  fourier.
  apply Rmult_le_reg_l with (IZR (2*b)); [apply (IZR_lt 0); omega| idtac].
  rewrite <- Rmult_assoc.
  rewrite Rinv_r_simpl_m; [idtac|discrR; omega].
  rewrite Rmult_1_r.
  apply IZR_le; omega.
  rewrite Rmult_1_r.
  dec_tac;assumption.
Qed.

Theorem Zopp_lt_cancel : 
  forall m n, -m < -n -> n < m.
Proof.
intros.
apply lt_IZR.
apply Ropp_lt_cancel.
repeat rewrite <- Ropp_Ropp_IZR.
apply IZR_lt; assumption.
Qed.

Theorem Zopp_le_cancel : 
  forall m n, -m <= -n -> n <= m.
Proof.
intros.
apply le_IZR.
apply Ropp_le_cancel.
repeat rewrite <- Ropp_Ropp_IZR.
apply IZR_le; assumption.
Qed.

Theorem Zmult_lt_compat : 
  forall n m p q : Z, n <= p -> m < q -> 0 < n -> 0 <= m -> n * m < p * q.
Proof.
intros.
elim (Z_le_gt_dec m 0).
intros; assert (m=0);[omega| subst].
rewrite Zmult_0_r.
apply Zmult_lt_O_compat; omega.
intros.
elim (Z_le_gt_dec p n).
intros; assert (p=n);[omega| subst].
apply Zmult_lt_compat_l; assumption.
intros.
apply lt_IZR.
repeat rewrite mult_IZR.
apply Rmult_gt_0_lt_compat.
assert (0 < IZR m)%R; [apply (IZR_lt 0); omega| fourier].
assert (0 < IZR p)%R; [apply (IZR_lt 0); omega| fourier].
apply IZR_lt; omega.
apply IZR_lt; omega.
Qed.

Theorem tech_bound2 : 
  forall b k l q,
    0<b ->
    (k*l = 2*b*q -> 
    -2*b+1 <= k <= 2*b-1 -> 
    -2*b+1 <= l <= 2*b-1 -> 
    -2*b+1 <= q <= 2*b-1)%Z.
Proof.
  intros b k l q Hbpos Heq Hkbound Hlbound.
  split;
  (apply Zmult_le_reg_r with (2*b); [omega|idtac]);
  (replace ( q * (2 * b)) with (k*l); [idtac| rewrite Heq; ring]);
  elim (Z_le_gt_dec k 0);
  elim (Z_le_gt_dec l 0); intros U1 U2.
  apply Zle_trans with 0.
  apply Zopp_le_cancel.
  replace (- ((-2 * b+1) * (2 * b))) with ((2*b-1)*2*b); [idtac|ring].
  change (-(0)) with 0.
  repeat apply Zmult_le_0_compat; omega.
  replace (k*l) with (-k*-l); [idtac|ring].
  apply Zmult_le_0_compat; omega.

  apply Zopp_le_cancel.
  replace (- ((-2 * b+1) * (2 * b))) with ((2*b-1)*(2*b)); [idtac|ring].
  replace (-(k*l)) with (l*-k); [idtac|ring].
  apply Zmult_le_compat; omega.

  apply Zopp_le_cancel.
  replace (- ((-2 * b+1) * (2 * b))) with ((2*b-1)*(2*b)); [idtac|ring].
  replace (-(k*l)) with (k*-l); [idtac|ring].
  apply Zmult_le_compat; omega.
  
  apply Zle_trans with 0.
  apply Zopp_le_cancel.
  replace (- ((-2 * b+1) * (2 * b))) with ((2*b-1)*(2*b)); [idtac|ring].
  change (-(0)) with 0.
  repeat apply Zmult_le_0_compat; omega.
  apply Zmult_le_0_compat; omega. 

  replace (k*l) with (-k*-l); [idtac|ring].
  apply Zmult_le_compat; omega.

  apply Zle_trans with 0.
  apply Zopp_le_cancel.
  change (-(0)) with 0.
  replace (-(k*l)) with (l*-k); [idtac|ring].
  apply Zmult_le_0_compat; omega. 
  apply Zmult_le_0_compat; omega. 

  apply Zle_trans with 0.
  apply Zopp_le_cancel.
  change (-(0)) with 0.
  replace (-(k*l)) with (-l*k); [idtac|ring].
  apply Zmult_le_0_compat; omega. 
  apply Zmult_le_0_compat; omega. 

  apply Zmult_le_compat; omega.
Qed.

Fixpoint multp (b:Z) (p:positive) (x:stream (idigit b)) {struct p}:
  stream (idigit b):=
  match p with 
    | xI p' => add b (mult2 b (multp b p' x)) x  
    | xO p' => mult2 b (multp b p' x)
    | xH => x
  end.

Definition multk (b k: Z) (x y:stream (idigit b)):  stream (idigit b):=
  match k with 
    | Z0 => x
    | Zpos p => add b x (multp b p (I b 0::y))
    | Zneg p => sub b x (multp b p (I b 0::y))
  end.
 
Theorem multp_correct : 
  forall b x r, 0<b ->
    represents b x r ->
    forall p,
      (-1 <= (IZR (Zpos p) * r) <= 1)%R ->
      represents b (multp b p x) (IZR (Zpos p) * r)%R.
Proof.
intros b x r Hbpos Hrep.
induction p.
rewrite Zpos_xI.
rewrite plus_IZR.
change (IZR 1) with 1%R.
rewrite Rmult_plus_distr_r.
rewrite Rmult_1_l.
rewrite mult_IZR.
simpl (multp).
change (IZR 2) with 2%R.
intros [H1 H2].
apply add_correct; try assumption.
rewrite Rmult_assoc.
assert (G: (-1 / 2 <= IZR (Zpos p) * r <= 1 / 2)%R).
split; (apply Rmult_le_reg_l with 2%R; [fourier| idtac]).
replace  (2 * (-1 / 2))%R with (-1)%R; [idtac| newfield].
elim (total_order_T r 0); intros H3.
elim H3; clear H3; intros H3.
apply Rle_trans with (1:= H1).
rewrite Rmult_assoc.
generalize (2 * (IZR (Zpos p) * r))%R.
intros; fourier.
subst.
repeat rewrite Rmult_0_r.
fourier.
apply Rle_trans with 0%R.
fourier.
apply Rmult_le_pos.
fourier.
apply Rmult_le_pos.
apply (IZR_le 0).
apply Zle_0_pos.
fourier.
replace  (2 * (1 / 2))%R with 1%R; [idtac| newfield].
elim (total_order_T r 0); intros H3.
elim H3; clear H3; intros H3.
apply Rle_trans with 0%R.
apply Ropp_le_cancel.
replace (-0)%R with 0%R; [idtac|ring].
rewrite <- Rmult_assoc.
rewrite <- Ropp_mult_distr_r_reverse.
apply Rmult_le_pos.
apply Rmult_le_pos.
fourier.
apply (IZR_le 0).
apply Zle_0_pos.
fourier.
fourier.
subst; rewrite Rmult_0_r; fourier.
apply Rle_trans with (2:=H2).
rewrite <- Rmult_assoc.
generalize (2 * IZR (Zpos p) * r)%R; intros; fourier.
apply mult2_correct; try assumption.
apply IHp.
dec_tac; split; fourier.
split; assumption.
rewrite Zpos_xO.
simpl multp.
rewrite mult_IZR.
change (IZR 2) with 2%R.
intros [H1 H2].
rewrite Rmult_assoc.
assert (G: (-1 / 2 <= IZR (Zpos p) * r <= 1 / 2)%R).
split; (apply Rmult_le_reg_l with 2%R; [fourier| idtac]).
replace  (2 * (-1 / 2))%R with (-1)%R; [idtac| newfield].
rewrite <- Rmult_assoc; assumption.
replace  (2 * (1 / 2))%R with 1%R; [idtac| newfield].
rewrite <- Rmult_assoc; assumption.
apply mult2_correct; try assumption.
apply IHp.
dec_tac; split; fourier.
intros _ ; simpl.
rewrite Rmult_1_l; assumption.
Qed.

Theorem multk_correct : forall b k x y r s,
  0 < b -> 
  represents b x r -> 
  represents b y s -> 
  -2*b+1 <= k <= 2*b-1 ->
  (-1 <= r + IZR k * s /IZR (2*b) <= 1)%R ->
  represents b (multk b k x y) (r + IZR k * s /IZR (2*b))%R. 
Proof.
intros b k x y r s Hbpos Hrepx Hrepy.
case k; simpl multk.
change (IZR 0) with 0%R.
unfold Rdiv; repeat rewrite Rmult_0_l.
intros; rewrite Rplus_0_r; assumption.
intros p Hkbound Hbound; apply add_correct; try assumption.
unfold Rdiv.
rewrite Rmult_assoc.
generalize (represents_interval b y s Hbpos Hrepy).
intros Hs.
apply multp_correct; try assumption.
replace (s * / IZR (2 * b))%R with ((IZR 0 +s)/IZR (2 * b))%R;[idtac|
  change (IZR 0) with 0%R; newfield; discrR; omega].
constructor; try assumption.
omega.
split;
(apply Rmult_le_reg_l with (IZR (2*b));[apply (IZR_lt 0); omega| idtac]);
(replace ( IZR (2 * b) * (IZR (Zpos p) * (s * / IZR (2 * b))))%R with 
  (IZR (Zpos p) * s)%R;[idtac| newfield; discrR; omega]).
replace (IZR (2 * b) * -1)%R with (IZR (-2*b)); [idtac|
  repeat rewrite mult_IZR; simpl; newfield].
apply Rle_trans with ( IZR (Zpos p)*IZR (-1))%R.
rewrite <- mult_IZR; apply IZR_le ; omega.
apply Rmult_le_compat_l.
apply (IZR_le 0); apply Zle_0_pos.
simpl; dec_tac; assumption.
apply Rle_trans with ( IZR (Zpos p)*1)%R.
apply Rmult_le_compat_l.
apply (IZR_le 0); apply Zle_0_pos.
dec_tac; assumption.
repeat rewrite Rmult_1_r; apply IZR_le ; omega.

intros p.
replace (Zneg p) with (- Zpos p); [idtac| rewrite <- Zopp_neg; ring].
replace (r + IZR (- Zpos p) * s / IZR (2 * b))%R with 
  (r - IZR (Zpos p) * s / IZR (2 * b))%R; [idtac|
    rewrite Ropp_Ropp_IZR; newfield; discrR; omega].
intros Hkbound Hbound.
apply sub_correct; try assumption.
unfold Rdiv.
rewrite Rmult_assoc.
generalize (represents_interval b y s Hbpos Hrepy).
intros Hs.
apply multp_correct; try assumption.
replace (s * / IZR (2 * b))%R with ((IZR 0 +s)/IZR (2 * b))%R;[idtac|
  change (IZR 0) with 0%R; newfield; discrR; omega].
constructor; try assumption.
omega.
split;
(apply Rmult_le_reg_l with (IZR (2*b));[apply (IZR_lt 0); omega| idtac]);
(replace ( IZR (2 * b) * (IZR (Zpos p) * (s * / IZR (2 * b))))%R with 
  (IZR (Zpos p) * s)%R;[idtac| newfield; discrR; omega]).
replace (IZR (2 * b) * -1)%R with (IZR (-2*b)); [idtac|
  repeat rewrite mult_IZR; simpl; newfield].
apply Rle_trans with ( IZR (Zpos p)*IZR (-1))%R.
rewrite <- mult_IZR; apply IZR_le; omega.
apply Rmult_le_compat_l.
apply (IZR_le 0); apply Zle_0_pos.
simpl; dec_tac; assumption.
apply Rle_trans with ( IZR (Zpos p)*1)%R.
apply Rmult_le_compat_l.
apply (IZR_le 0); apply Zle_0_pos.
dec_tac; assumption.
repeat rewrite Rmult_1_r; apply IZR_le ; omega.
Qed.

(*Theorem mult_correct_body :
  forall b, 3<=b ->
  forall mult_a' : stream idigit -> stream idigit*stream idigit -> stream idigit,
    (forall v1 v2 k x u u', 
      (-1/4 <= v3 <= 1/4)%R ->
      (-1 <= v1 +v2*k <= 1)%R ->
      represents x v1 -> 
      represents u v2 -> 
      represents u' v3 ->
      represents (mult_a' x (u, u')) (v1+v2*v3)%R) ->
    forall v1 v2 v3 x u u', 
      (-1/4 <= v3 <= 1/4)%R ->
      (-1 <= v1 +v2*v3 <= 1)%R ->
      represents x v1 -> 
      represents u (2*v2) -> 
      represents u' v3 ->
      represents (series_body _ mult_a' x (u, u')) (v1+v2*v3)%R.
Proof.

Theorem multkdiv2b_correct : 
  forall b k x u, 
    0 < b -> 
    -2*b+1 <= k <= 2*b-1 ->
    represents b x u -> 
    represents b (multkdiv2b b k x) (IZR k * u / IZR (2*b)). 
Proof. 

cofix.
intros b k x u Hbpos Hkbound Hrep.
str_unfold (multkdiv2b b k x).
destruct x as [[l] x'].
Opaque Zmult.
simpl.
Transparent Zmult.
caseEq (Zdiv_eucl (k * l) (2 * b)).
intros q r HZdiv.
assert (G1 : 2*b > 0).
auto with zarith.
assert (G:=Z_div_mod (k*l) (2*b) G1).
rewrite HZdiv in G.
elim G;clear G1 G HZdiv;intros G1 G2.

elim (Z_eq_dec r 0).
intros; subst.
rewrite Zplus_0_r in G1.
clear G2.
inversion Hrep.
replace (IZR k * ((IZR l + r) / IZR (2 * b)))%R with
  (IZR q + IZR k * r / IZR (2 * b))%R.
constructor; try (apply multkdiv2b_correct; assumption); clear multkdiv2b_correct.
apply techbound; assumption.
apply tech_bound2 with (k:=k) (l:=l); assumption.
unfold Rdiv.
rewrite <- Rmult_assoc.
rewrite Rmult_plus_distr_l.
rewrite <- mult_IZR.
rewrite G1.
rewrite (mult_IZR (2*b)).
newfield; discrR; omega.
Admitted.
*)


(* compute x + uw *)
CoFixpoint mult_a (b:Z) (x:stream (idigit b)) 
  (p:stream (idigit b) *stream (idigit b)) : stream (idigit b) :=
  let (u,v) := p in
    match u with
      (I k)::u' => 
      series_body' b _ (mult_a b) (add _ x (multkdiv2b _ k v)) (u',v)
    end.

Lemma mult_correct_aux_lemma :
  forall x y z,
    (0<z)%R ->
    (-z <= y <= z)%R -> 
    (-1 <= x <= 1)%R ->
    (-z<= x * y <= z)%R.
Proof.
intros x y z Hzpos [Hxge Hxle] [Hyge Hyle].
elim (Rlt_le_dec 0 x); elim (Rlt_le_dec 0 y).

intros Hgt1 Hgt2.
split.
apply Rle_trans with 0%R.
fourier.
apply Rlt_le.
apply Rmult_lt_0_compat; fourier.
replace (z)%R  with (1*z)%R; [idtac| newfield].
apply Rmult_le_compat; fourier.

intros Hle Hgt.
split.
apply Ropp_le_cancel.
replace (- (x * y))%R with (x * (-y))%R; [idtac| newfield].
replace (- - z)%R with (1*z)%R; [idtac| newfield].
apply Rmult_le_compat; fourier.
apply Rle_trans with 0%R.
replace 0%R with (x * 0)%R; [idtac| newfield].
apply Rmult_le_compat_l; fourier.
fourier.

intros Hgt Hle.
split.
apply Ropp_le_cancel.
replace (- (x * y))%R with ((-x) * y)%R; [idtac| newfield].
replace (- -z)%R with (1*z)%R; [idtac| newfield].
apply Rmult_le_compat; fourier.
apply Rle_trans with 0%R.
replace 0%R with (0*y)%R; [idtac| newfield].
apply Rmult_le_compat_r; fourier.
fourier.

intros Hle1 Hle2.
replace (x * y)%R with ((-x) * (-y))%R; [idtac| newfield].
split.
apply Rle_trans with 0%R.
fourier.
replace 0%R with (0*0)%R; [idtac| newfield].
apply Rmult_le_compat; fourier.
replace z  with (1*z)%R; [idtac| newfield].
apply Rmult_le_compat; fourier.
Qed.

Theorem tech_bound_1 : 
  forall r b v1 v2 v3,
    0<b ->
    ((IZR (2*b-1) + r) / IZR (2 * b))%R = v1 ->
    (v1 + Rabs v3 <= 1 -> 
      -1 <= v1 - Rabs v3 ->
      -1 <= v2 <= 1 ->
      r + IZR (2 * b) * v2 * v3 <= 1)%R.
Proof.
  intros r b v1 v2 v3 Hbpos Hv1eq Hsum Hmin Hv2.
  subst.
  apply Rplus_le_reg_l with (IZR (2*b)- 1)%R.
  replace (IZR (2 * b) - 1 + 1)%R with (1*IZR (2 * b))%R; [idtac| ring]. 
  replace (IZR (2 * b) - 1 + (r + IZR (2 * b) * v2 * v3))%R with 
    (((IZR (2 * b - 1) + r) / IZR (2 * b) + v2 * v3)*(IZR (2*b)))%R; [idtac|
      rewrite <- Z_R_minus; change (IZR 1) with 1%R; newfield; discrR; omega].
  apply Rmult_le_compat_r.
  apply (IZR_le 0); omega.
  apply Rle_trans with (2:=Hsum).
  apply Rplus_le_compat_l.
  elim (Rle_or_lt 0 v3).
  intros Hv3pos.
  replace (Rabs v3) with (1*v3)%R; [idtac| rewrite Rabs_right;[ring | intuition]].
  dec_tac; apply Rmult_le_compat_r; assumption.
  intros Hv3neg.
  replace (Rabs v3) with (1*-v3)%R;
    [idtac| rewrite Rabs_left;[idtac|assumption]; ring].
  rewrite <- Rmult_opp_opp.
  dec_tac; apply Rmult_le_compat_r; fourier.
Qed.

Theorem mult_body_aux_lemma : 
  forall b x v2,
    0 < b ->
    represents b x (2*IZR b*v2)%R ->
    (-1 <= v2 <= 1)%R.
Proof.  
intros b x v2 Hbpos Hrep.
generalize (represents_interval _ _ _ Hbpos Hrep).
intros [U1 U2].
assert (0<2*IZR b)%R.
apply Rmult_lt_0_compat.  
fourier.
apply (IZR_lt 0); omega.
split; (apply Rmult_le_reg_l with (2*IZR b)%R;[assumption|idtac]).
apply Rle_trans with (2:= U1).
replace (2 * IZR b * -1)%R with (IZR (-2*b));[idtac|rewrite mult_IZR; simpl; ring].
change (-1)%R with (IZR (-1)).
apply IZR_le; omega.
rewrite Rmult_1_r.
apply Rle_trans with (1:=U2).
rewrite <- (mult_IZR 2).
apply (IZR_le 1); omega.
Qed.

Theorem mult_correct_body :
  forall b, 1 < b ->
    forall mult_a' : stream (idigit b) -> stream (idigit b) * stream (idigit b)->
      stream (idigit b),
      (forall v1 v2 v3 x u u', 
	(-1/(4*IZR b) <= v3 <= 1/(4*IZR b))%R ->
	(v1 + Rabs v3<= 1)%R ->
	(-1 <=v1 - Rabs v3)%R ->
	represents _ x v1 ->
	represents _ u v2 ->
	represents _ u' v3 ->
	represents _ (mult_a' x (u, u')) (v1+v2*v3)%R) ->
      forall v1 v2 v3 x u u',
	(-1/(4*IZR b) <= v3 <= 1/(4*IZR b))%R ->
	(v1 + Rabs v3<= 1)%R ->
	(-1 <=v1 - Rabs v3)%R ->
	represents _ x v1 ->
	represents _ u (2*IZR b*v2) ->
	represents _ u' v3 ->
	represents _ (series_body' _ _ mult_a' x (u, u')) (v1+v2*v3)%R.
Proof.
  intros b H1ltb mult_a' mult_correct_body v1 v2 v3 x u u' 
    Hv3bound Hsum Hmin Hv1 Hv2 Hv3.
  assert (Hbpos : 0<b);[omega|idtac].
  assert (H4bpos: (0 < 4 * IZR b)%R).
  clear mult_correct_body.
  replace (4*IZR b)%R with (IZR (4*b)); [idtac| rewrite mult_IZR; simpl; ring].
  apply (IZR_lt 0); omega.
  assert (Hbound: (-1/(4*IZR b)<=2*IZR b*v2*v3 <= 1/(4*IZR b))%R).
  clear mult_correct_body.
  generalize (represents_interval _ _ _ Hbpos Hv2).
  intros.
  replace (-1 / (4 * IZR b))%R with 
    (-(1 / (4 * IZR b)))%R; [idtac|unfold Rdiv; ring].
  apply mult_correct_aux_lemma.
  unfold Rdiv; rewrite Rmult_1_l.
  apply Rinv_0_lt_compat; assumption.
  replace (- (1 / (4 * IZR b)))%R with (-1 / (4 * IZR b))%R;
    [idtac| unfold Rdiv; ring].
  assumption.
  assumption.

  unfold series_body'.

  destruct x as [ [k1] [[k2] x2]].
  elim (Z_le_gt_dec b k2).
  intros Hblek2.
  inversion Hv1.
  assert (Hsumpos: (0 <= r + IZR (2 * b) * v2 * v3)%R).
  clear mult_correct_body.
  rewrite <- Rplus_opp_r with (IZR (2 * b) * v2 * v3)%R.
  rewrite Rplus_comm.
  apply Rplus_le_compat_r.
  generalize (I_interval _ _ _ _ Hbpos H1).
  intros [Hr _].
  apply Rle_trans with (2:=Hr).
  apply Ropp_le_cancel.
  rewrite Ropp_involutive.
  rewrite mult_IZR; simpl.
  generalize Hbound; intros [Hbound1 _].
  apply Rle_trans with (2:= Hbound1).
  replace (-1 / (4 * IZR b))%R with 
    (- (1 / (4 * IZR b)))%R; [idtac|newfield; discrR;omega].
  apply Ropp_le_contravar.
  apply Rmult_le_reg_l with (4 * IZR b)%R. 
  assumption.
  replace (4 * IZR b * (1 / (4 * IZR b)))%R with (IZR 0 + 1)%R;
    [ idtac| simpl; newfield;discrR; auto with zarith].
  replace ( 4 * IZR b * ((IZR k2 - 1) / (2 * IZR b)))%R with (IZR (2*k2-3) +1)%R;
    [idtac|rewrite <- Z_R_minus; rewrite mult_IZR; simpl ; newfield; discrR; omega].
  apply Rplus_le_compat_r.
  apply IZR_le; omega.

  elim (Z_eq_dec k1 (2 * b - 1)).
  intros Hk1eq.
  replace ((IZR k1 + r) / IZR (2 * b) + v2 * v3)%R with 
    ((IZR (k1) + (r+IZR (2*b)*v2*v3))/(IZR (2*b)))%R; 
    [idtac| newfield; discrR; omega].
  assert (-1 <= r + IZR (2 * b) * v2 * v3 <= 1)%R.
  clear mult_correct_body.
  split.
  apply Rle_trans with (2:= Hsumpos); fourier.
  apply tech_bound_1 with v1; try assumption.
  subst; reflexivity.
  apply (mult_body_aux_lemma _ _ _ Hbpos Hv2).
  cut (r + Rabs v3 <= 1)%R.
  cut (-1 <= r - Rabs v3)%R.
  intros.
  constructor; try assumption.
  apply mult_correct_body; clear mult_correct_body; try assumption.
  rewrite mult_IZR; simpl; assumption.
  clear mult_correct_body.
Admitted.
(*  

  dec_tac; assumption.
  constructor; try assumption.
  apply mult_correct_body; clear mult_correct_body; try assumption.
  rewrite mult_IZR; simpl; assumption.
Rabs_left: forall r : R, (r < 0)%R -> Rabs r = (- r)%R
Rabs_right: forall r : R, (r >= 0)%R -> Rabs r = r





  assert (r + Rabs v3 <= 1)%R.
  assert (-1 <= r - Rabs v3)%R
  assert (Hsumbound: (-1 <= r + IZR (2 * b) * v2 * v3 <= 1)%R).
  clear mult_correct_body.
  split.
  apply Rle_trans with (2:= Hsumpos); fourier.
  subst.
  apply Rplus_le_reg_l with (IZR (2*b)- 1)%R.
  replace (IZR (2 * b) - 1 + 1)%R with (1*IZR (2 * b))%R; [idtac| ring]. 
  replace (IZR (2 * b) - 1 + (r + IZR (2 * b) * v2 * v3))%R with 
    (((IZR (2 * b - 1) + r) / IZR (2 * b) + v2 * v3)*(IZR (2*b)))%R; [idtac|
      rewrite <- Z_R_minus; change (IZR 1) with 1%R; newfield; discrR; omega].
  apply Rmult_le_compat_r.
  apply (IZR_le 0); omega.
  dec_tac; assumption.
  constructor; try assumption.
  apply mult_correct_body; clear mult_correct_body; try assumption.
  rewrite mult_IZR; simpl; assumption.

  intros Hk1neq.
  replace ((IZR k1 + r) / IZR (2 * b) + v2 * v3)%R with 
    ((IZR (k1+1) + ((r-1)+IZR (2*b)*v2*v3))/(IZR (2*b)))%R; 
    [idtac| rewrite plus_IZR; change (IZR 1) with 1%R; newfield; discrR; omega].
  assert (Hsumbound: (-1 <= r - 1 + IZR (2 * b) * v2 * v3 <= 1)%R).
  clear mult_correct_body.
  split.
  replace ( r - 1 + IZR (2 * b) * v2 * v3)%R with 
    (r+ IZR (2 * b) * v2 * v3 +-1)%R;[idtac|ring].
  pattern (-1)%R at 1; replace (-1)%R with (0+-1)%R; [idtac|ring].
  apply Rplus_le_compat_r; assumption.
  pattern 1%R at -1; replace 1%R with (0+1)%R; [idtac|ring].
  apply Rplus_le_compat.
  dec_tac; fourier.
  rewrite mult_IZR; simpl.
  generalize Hbound; intros [_ Hbound2].
  apply Rle_trans with (1:= Hbound2).
  apply Rmult_le_reg_l with (4 * IZR b)%R.
  assumption.
  replace (4 * IZR b * (1 / (4 * IZR b)))%R with (IZR 1)%R;
    [ idtac| simpl; newfield;discrR; auto with zarith].
  replace (4 * IZR b * 1)%R with (IZR (4*b)); [idtac | 
    rewrite mult_IZR; simpl; ring].
  apply IZR_le; omega.
  constructor; try assumption.
  apply mult_correct_body; clear mult_correct_body; try assumption.
  inversion_to_variable x2.
  replace ((IZR k2 + r0) / IZR (2 * b) - 1)%R with 
    ((IZR (k2 -2*b) +r0)/IZR (2 * b))%R; 
    [idtac| rewrite <- Z_R_minus; newfield; discrR; auto with zarith].
  constructor; try assumption; omega.
  rewrite mult_IZR; simpl; assumption.
  omega.

  elim (Z_le_gt_dec k2 (- b)).
  intros Hk2leb _.
  inversion Hv1.
  assert (Hsumneg: (r + IZR (2 * b) * v2 * v3 <= 0)%R).
  clear mult_correct_body.
  rewrite <- Rplus_opp_r with (IZR (2 * b) * v2 * v3)%R.
  rewrite Rplus_comm.
  apply Rplus_le_compat_l.
  generalize (I_interval _ _ _ _ Hbpos H1).
  intros [_ Hr].
  apply Rle_trans with (1:=Hr).
  apply Ropp_le_cancel.
  rewrite Ropp_involutive.
  rewrite mult_IZR; simpl.
  generalize Hbound; intros [_ Hbound2].
  apply Rle_trans with (1:= Hbound2).
  apply Rmult_le_reg_l with (4 * IZR b)%R. 
  assumption.
  replace (4 * IZR b * (1 / (4 * IZR b)))%R with (IZR 0 + 1)%R;
    [ idtac| simpl; newfield;discrR; auto with zarith].
  replace ( 4 * IZR b * - ((IZR k2 + 1) / (2 * IZR b)))%R with (IZR (-2*k2-3) +1)%R;
    [idtac|rewrite <- Z_R_minus; rewrite mult_IZR; simpl ; newfield; discrR; omega].
  apply Rplus_le_compat_r.
  apply IZR_le; omega.

  elim (Z_eq_dec k1 (-2 * b + 1)).
  intros Hk1eq.
  replace ((IZR k1 + r) / IZR (2 * b) + v2 * v3)%R with 
    ((IZR k1 + (r+IZR (2*b)*v2*v3))/(IZR (2*b)))%R; 
    [idtac| newfield; discrR; omega].
  assert (Hsumbound: (-1 <= r + IZR (2 * b) * v2 * v3 <= 1)%R).
  clear mult_correct_body.
  split.
  subst.
  apply Rplus_le_reg_l with (-IZR (2*b)+ 1)%R.
  replace (-IZR (2 * b) + 1 + -1)%R with (-1*IZR (2 * b))%R; [idtac| ring]. 
  replace (-IZR (2 * b) + 1 + (r + IZR (2 * b) * v2 * v3))%R with 
    (((IZR (-2 * b + 1) + r) / IZR (2 * b) + v2 * v3)*IZR (2 * b))%R; [idtac|
      rewrite plus_IZR; repeat rewrite mult_IZR; simpl; newfield; discrR; omega].
  apply Rmult_le_compat_r.
  apply (IZR_le 0); omega.
  dec_tac; assumption.
  apply Rle_trans with (1:= Hsumneg); fourier.
  constructor; try assumption.
  apply mult_correct_body; clear mult_correct_body; try assumption.
  rewrite mult_IZR; simpl; assumption.

  intros Hk1neq.
  replace ((IZR k1 + r) / IZR (2 * b) + v2 * v3)%R with 
    ((IZR (k1-1) + ((r+1)+IZR (2*b)*v2*v3))/(IZR (2*b)))%R; [idtac| 
      rewrite <- Z_R_minus; change (IZR 1) with 1%R; newfield; discrR; omega].
  assert (Hsumbound:   (-1 <= r + 1 + IZR (2 * b) * v2 * v3 <= 1)%R).
  clear mult_correct_body.
  split.
  replace (-1)%R with (0+-1)%R; [idtac|ring].
  apply Rplus_le_compat.
  dec_tac; fourier.
  rewrite mult_IZR; simpl.
  generalize Hbound; intros [Hbound1 _].
  apply Rle_trans with (2:= Hbound1).
  apply Rmult_le_reg_l with (4 * IZR b)%R.
  assumption.
  replace (4 * IZR b * (-1 / (4 * IZR b)))%R with (IZR (-1))%R;
    [ idtac| simpl; newfield;discrR; auto with zarith].
  replace (4 * IZR b * -1)%R with (IZR (-4*b)); [idtac | 
    rewrite mult_IZR; simpl; ring].
  apply IZR_le; omega.  
  replace ( r + 1 + IZR (2 * b) * v2 * v3)%R with 
    (r+ IZR (2 * b) * v2 * v3 +1)%R;[idtac|ring].
  pattern 1%R at -1; replace 1%R with (0+1)%R; [idtac|ring].
  apply Rplus_le_compat_r; assumption.
  constructor; try assumption.
  apply mult_correct_body; clear mult_correct_body; try assumption.
  inversion_to_variable x2.
  replace ((IZR k2 + r0) / IZR (2 * b) + 1)%R with 
    ((IZR (k2 +2*b) +r0)/IZR (2 * b))%R; 
    [idtac| rewrite plus_IZR; newfield; discrR; auto with zarith].
  constructor; try assumption; omega.
  rewrite mult_IZR; simpl; assumption.
  omega.

  intros Hk2gt Hk2lt.
  inversion Hv1.
  replace ((IZR k1 + r) / IZR (2 * b) + v2 * v3)%R with 
    ((IZR k1 + (r+IZR (2*b)*v2*v3))/(IZR (2*b)))%R; 
    [idtac| newfield; discrR; omega].
  assert (Hsumbound: (-1 <= r + IZR (2 * b) * v2 * v3 <= 1)%R).
  clear mult_correct_body.
  generalize (I_interval _ _ _ _ Hbpos H1).
  intros.
  rewrite mult_IZR; simpl.
  split.
  apply Rle_trans with ((IZR k2 - 1) / IZR (2 * b) + -1 / (4 * IZR b))%R.
  apply Rmult_le_reg_l with (4 * IZR b)%R. 
  assumption.
  replace (4 * IZR b * -1)%R with (IZR (-4*b)); [idtac | 
    rewrite mult_IZR; simpl; ring].
  replace (4 * IZR b * ((IZR k2 - 1) / IZR (2 * b) + -1 / (4 * IZR b)))%R with
    (IZR (2*k2-3))%R; [ idtac|  rewrite <- Z_R_minus; repeat rewrite mult_IZR;
      simpl; newfield;discrR; auto with zarith].
  apply IZR_le; omega.
  dec_tac; apply Rplus_le_compat; assumption.
  apply Rle_trans with ((IZR k2 + 1) / IZR (2 * b) + 1 / (4 * IZR b))%R.
  dec_tac; apply Rplus_le_compat; assumption.
  apply Rmult_le_reg_l with (4 * IZR b)%R. 
  assumption.
  replace (4 * IZR b * 1)%R with (IZR (4*b)); [idtac | 
    rewrite mult_IZR; simpl; ring].
  replace (4 * IZR b * ((IZR k2 + 1) / IZR (2 * b) + 1 / (4 * IZR b)))%R with
    (IZR (2*k2+3))%R; [ idtac|  rewrite plus_IZR; repeat rewrite mult_IZR;
      simpl; newfield;discrR; auto with zarith].
  apply IZR_le; omega.
  constructor; try assumption.
  apply mult_correct_body; clear mult_correct_body; try assumption.
  rewrite mult_IZR; simpl; assumption.
Defined.  

Ltac force_represents :=
 match goal with
  id: represents _ ?x ?y |- represents _ ?x ?z =>
  replace z with y;[assumption | newfield; discrR; auto with zarith]
 end.

Theorem mult_a_correct :
  forall b, 1 < b ->
    forall v1 v2 v3 x u u', 
      (-1/(4*IZR b) <= v3 <= 1/(4*IZR b))%R ->
      (-1 <=v1 +v2*v3 <= 1)%R ->
      represents _ x v1 ->
      represents _ u v2 ->
      represents _ u' v3 ->
      represents _ (mult_a b x (u, u')) (v1+v2*v3)%R.
Proof.
cofix.
intros b H1ltb v1 v2 v3 x u u' Hv3bound Hsum Hv1 Hv2 Hv3.
rewrite (str_dec_thm (mult_a b x (u, u'))).
lazy beta iota zeta delta [str_decompose mult_a]; fold mult_a.

destruct u as [ [k] u1].
fold (str_decompose (series_body' b (stream (idigit b) * stream (idigit b)) 
         (mult_a b) (add b x (multkdiv2b b k u')) (u1, u'))).
rewrite <- str_dec_thm.
inversion Hv2.
replace ( (v1 + (IZR k + r) / IZR (2 * b) * v3))%R with 
  ((v1 + (IZR k * v3) /IZR (2 * b) + r / IZR (2 * b) * v3))%R;
  [idtac|newfield; discrR; omega].
apply mult_correct_body; try assumption. 
apply mult_a_correct; clear mult_a_correct; assumption.
clear mult_a_correct.
replace (v1+IZR k * v3 / IZR (2 * b) + r / IZR (2 * b) * v3)%R with 
  (v1+v2*v3)%R; [assumption | subst; newfield; discrR; omega].
clear mult_a_correct.
apply add_correct; try assumption.
omega.
apply multkdiv2b_correct; try assumption; omega.
2:rewrite mult_IZR; simpl; force_represents.


subst.
generalize Hsum; clear Hsum; intros [Hsum1 Hsum2].
split.
apply Rle_trans with (1:=Hsum1).
apply Rplus_le_compat_l.








fold (str_decompose (series_body (stream idigit * stream idigit) mult_a (x + (L:: u')) (u1, u'))).
rewrite <- str_dec_thm.
inversion Hv2.
match goal with |- represents _ ?a => 
  replace a with ((v1+v3/2)+r/2*v3)%R; 
    [idtac | clear mult_a_correct; field; auto with real1]
end.
apply mult_correct_body; (exact mult_a_correct || clear mult_a_correct); auto.
replace (v1 + v3/2 + r/2 *v3)%R with (v1 + v2*v3)%R;
  [assumption | subst v2; field; auto with real1].
apply add_correct.
assumption.
constructor.
assumption.
interval_from_represents Hv3; auto.
apply Rle_trans with (2:= Hsum).
apply Rplus_le_compat_l.
replace (v3/2)%R with (1/2*v3)%R.
apply Rmult_le_compat_r.
interval_from_represents Hv3; intros; dec_tac; auto.
subst v2; autorewrite with real1.
dec_tac; fourier.
field; auto with real1.
force_represents.

fold (str_decompose (series_body (stream idigit * stream idigit) mult_a (x + (L::L:: u')) (u1, u'))).
rewrite <- str_dec_thm.
inversion Hv2.

match goal with |- represents _ ?a => 
  replace a with ((v1+(v3/2/2))+r/2*v3)%R; 
    [idtac | clear mult_a_correct; field; auto with real1]
end.
apply mult_correct_body; (exact mult_a_correct || clear mult_a_correct);auto.
replace (v1 + v3/2/2 + r/2 *v3)%R with (v1 + v2*v3)%R;
  [assumption | subst v2; field; auto with real1].
apply add_correct.
assumption.
constructor.
constructor.
assumption.
interval_from_represents Hv3; auto.
interval_from_represents Hv3; intros; dec_tac; split; fourier.
apply Rle_trans with (2:= Hsum).
apply Rplus_le_compat_l.
replace (v3/2/2)%R with (1/4*v3)%R.
apply Rmult_le_compat_r.
interval_from_represents Hv3; intros; dec_tac; fourier.
subst v2; autorewrite with real1.
dec_tac; fourier.
field; auto with real1.
force_represents.
Qed.

Theorem zero_correct : represents zero 0%R.
cofix.
rewrite (str_dec_thm zero); simpl.
replace 0%R with (0/2)%R.
constructor.
apply zero_correct.
clear zero_correct.
split; fourier.
clear zero_correct.
unfold Rdiv; ring.
Qed.

Theorem mult_correct :
  forall x y vx vy, represents x vx -> represents y vy ->
    represents (x * y) (vx * vy).
intros x y vx vy Hx Hy.
replace (vx * vy)%R with (2*(2*(vx * (vy/2 /2))))%R.
unfold mult.
interval_from_represents Hx; interval_from_represents Hy; intros; dec_tac.
assert (vx * (vy/2/2) <= 1/4)%R.
replace (vy/2/2)%R with (1/4*vy)%R;[idtac|field;auto with real1].
replace (1/4)%R with (1* (1/4*1))%R.
apply Rmult_le_compat.
assumption.
repeat apply Rmult_le_pos.
fourier.
fourier.
fourier.
assumption.
assumption.
apply Rmult_le_compat.
repeat apply Rmult_le_pos; fourier.
assumption.
rewrite Rmult_1_l; rewrite Rmult_1_r.
auto with real.
assumption.
ring.
apply mult2_correct.
split.
repeat apply Rmult_le_pos; try fourier.
auto with real1.
auto with real1.
apply mult2_correct.
split.
repeat apply Rmult_le_pos; try fourier.
auto with real1.
auto with real1.
replace (vx * (vy / 2/2))%R with (0 + vx * (vy/2/2))%R;[idtac|ring].
apply mult_a_correct.
replace (vy/2/2)%R with (1/4*vy)%R;[idtac | field; auto with real1].
fourier.
fourier.
apply zero_correct.
assumption.
constructor.
constructor.
assumption.
auto.
auto with real1.
field; auto with real1.
Qed.








Theorem mult_correct_body :
  forall b, 1 < b ->
    forall mult_a' : stream (idigit b) -> stream (idigit b) * stream (idigit b)->
      stream (idigit b),
      (forall v1 v2 v3 x u u', 
	(-1/(4*IZR b) <= v3 <= 1/(4*IZR b))%R ->
	(-1 <=v1 +v2*v3 <= 1)%R ->
	represents _ x v1 ->
	represents _ u v2 ->
	represents _ u' v3 ->
	represents _ (mult_a' x (u, u')) (v1+v2*v3)%R) ->
      forall v1 v2 v3 x u u',
	(-1/(4*IZR b) <= v3 <= 1/(4*IZR b))%R ->
	(-1 <=v1 +v2*v3 <= 1)%R ->
	represents _ x v1 -> 
	represents _ u (2*IZR b*v2) ->
	represents _ u' v3 ->
	represents _ (series_body' _ _ mult_a' x (u, u')) (v1+v2*v3)%R.
Proof.
  intros b H1ltb mult_a' mult_correct_body v1 v2 v3 x u u' Hv3bound Hsum Hv1 Hv2 Hv3.

  assert (Hbpos : 0<b);[omega|idtac].
  assert (H4bpos: (0 < 4 * IZR b)%R).
  clear mult_correct_body.
  replace (4*IZR b)%R with (IZR (4*b)); [idtac| rewrite mult_IZR; simpl; ring].
  apply (IZR_lt 0); omega.
  assert (Hbound: (-1/(4*IZR b)<=2*IZR b*v2*v3 <= 1/(4*IZR b))%R).
  clear mult_correct_body.
  generalize (represents_interval _ _ _ Hbpos Hv2).
  intros.
  replace (-1 / (4 * IZR b))%R with 
    (-(1 / (4 * IZR b)))%R; [idtac|unfold Rdiv; ring].
  apply mult_correct_aux_lemma.
  unfold Rdiv; rewrite Rmult_1_l.
  apply Rinv_0_lt_compat; assumption.
  replace (- (1 / (4 * IZR b)))%R with (-1 / (4 * IZR b))%R;
    [idtac| unfold Rdiv; ring].
  assumption.
  assumption.

  unfold series_body'.

  destruct x as [ [k1] [[k2] x2]].
  elim (Z_le_gt_dec b k2).
  intros Hblek2.
  inversion Hv1.
  assert (Hsumpos: (0 <= r + IZR (2 * b) * v2 * v3)%R).
  clear mult_correct_body.
  rewrite <- Rplus_opp_r with (IZR (2 * b) * v2 * v3)%R.
  rewrite Rplus_comm.
  apply Rplus_le_compat_r.
  generalize (I_interval _ _ _ _ Hbpos H1).
  intros [Hr _].
  apply Rle_trans with (2:=Hr).
  apply Ropp_le_cancel.
  rewrite Ropp_involutive.
  rewrite mult_IZR; simpl.
  generalize Hbound; intros [Hbound1 _].
  apply Rle_trans with (2:= Hbound1).
  replace (-1 / (4 * IZR b))%R with 
    (- (1 / (4 * IZR b)))%R; [idtac|newfield; discrR;omega].
  apply Ropp_le_contravar.
  apply Rmult_le_reg_l with (4 * IZR b)%R. 
  assumption.
  replace (4 * IZR b * (1 / (4 * IZR b)))%R with (IZR 0 + 1)%R;
    [ idtac| simpl; newfield;discrR; auto with zarith].
  replace ( 4 * IZR b * ((IZR k2 - 1) / (2 * IZR b)))%R with (IZR (2*k2-3) +1)%R;
    [idtac|rewrite <- Z_R_minus; rewrite mult_IZR; simpl ; newfield; discrR; omega].
  apply Rplus_le_compat_r.
  apply IZR_le; omega.

  elim (Z_eq_dec k1 (2 * b - 1)).
  intros Hk1eq.
  replace ((IZR k1 + r) / IZR (2 * b) + v2 * v3)%R with 
    ((IZR (k1) + (r+IZR (2*b)*v2*v3))/(IZR (2*b)))%R; 
    [idtac| newfield; discrR; omega].
  assert (Hsumbound: (-1 <= r + IZR (2 * b) * v2 * v3 <= 1)%R).
  clear mult_correct_body.
  split.
  apply Rle_trans with (2:= Hsumpos); fourier.
  subst.
  apply Rplus_le_reg_l with (IZR (2*b)- 1)%R.
  replace (IZR (2 * b) - 1 + 1)%R with (1*IZR (2 * b))%R; [idtac| ring]. 
  replace (IZR (2 * b) - 1 + (r + IZR (2 * b) * v2 * v3))%R with 
    (((IZR (2 * b - 1) + r) / IZR (2 * b) + v2 * v3)*(IZR (2*b)))%R; [idtac|
      rewrite <- Z_R_minus; change (IZR 1) with 1%R; newfield; discrR; omega].
  apply Rmult_le_compat_r.
  apply (IZR_le 0); omega.
  dec_tac; assumption.
  constructor; try assumption.
  apply mult_correct_body; clear mult_correct_body; try assumption.
  rewrite mult_IZR; simpl; assumption.

  intros Hk1neq.
  replace ((IZR k1 + r) / IZR (2 * b) + v2 * v3)%R with 
    ((IZR (k1+1) + ((r-1)+IZR (2*b)*v2*v3))/(IZR (2*b)))%R; 
    [idtac| rewrite plus_IZR; change (IZR 1) with 1%R; newfield; discrR; omega].
  assert (Hsumbound: (-1 <= r - 1 + IZR (2 * b) * v2 * v3 <= 1)%R).
  clear mult_correct_body.
  split.
  replace ( r - 1 + IZR (2 * b) * v2 * v3)%R with 
    (r+ IZR (2 * b) * v2 * v3 +-1)%R;[idtac|ring].
  pattern (-1)%R at 1; replace (-1)%R with (0+-1)%R; [idtac|ring].
  apply Rplus_le_compat_r; assumption.
  pattern 1%R at -1; replace 1%R with (0+1)%R; [idtac|ring].
  apply Rplus_le_compat.
  dec_tac; fourier.
  rewrite mult_IZR; simpl.
  generalize Hbound; intros [_ Hbound2].
  apply Rle_trans with (1:= Hbound2).
  apply Rmult_le_reg_l with (4 * IZR b)%R.
  assumption.
  replace (4 * IZR b * (1 / (4 * IZR b)))%R with (IZR 1)%R;
    [ idtac| simpl; newfield;discrR; auto with zarith].
  replace (4 * IZR b * 1)%R with (IZR (4*b)); [idtac | 
    rewrite mult_IZR; simpl; ring].
  apply IZR_le; omega.
  constructor; try assumption.
  apply mult_correct_body; clear mult_correct_body; try assumption.
  inversion_to_variable x2.
  replace ((IZR k2 + r0) / IZR (2 * b) - 1)%R with 
    ((IZR (k2 -2*b) +r0)/IZR (2 * b))%R; 
    [idtac| rewrite <- Z_R_minus; newfield; discrR; auto with zarith].
  constructor; try assumption; omega.
  rewrite mult_IZR; simpl; assumption.
  omega.

  elim (Z_le_gt_dec k2 (- b)).
  intros Hk2leb _.
  inversion Hv1.
  assert (Hsumneg: (r + IZR (2 * b) * v2 * v3 <= 0)%R).
  clear mult_correct_body.
  rewrite <- Rplus_opp_r with (IZR (2 * b) * v2 * v3)%R.
  rewrite Rplus_comm.
  apply Rplus_le_compat_l.
  generalize (I_interval _ _ _ _ Hbpos H1).
  intros [_ Hr].
  apply Rle_trans with (1:=Hr).
  apply Ropp_le_cancel.
  rewrite Ropp_involutive.
  rewrite mult_IZR; simpl.
  generalize Hbound; intros [_ Hbound2].
  apply Rle_trans with (1:= Hbound2).
  apply Rmult_le_reg_l with (4 * IZR b)%R. 
  assumption.
  replace (4 * IZR b * (1 / (4 * IZR b)))%R with (IZR 0 + 1)%R;
    [ idtac| simpl; newfield;discrR; auto with zarith].
  replace ( 4 * IZR b * - ((IZR k2 + 1) / (2 * IZR b)))%R with (IZR (-2*k2-3) +1)%R;
    [idtac|rewrite <- Z_R_minus; rewrite mult_IZR; simpl ; newfield; discrR; omega].
  apply Rplus_le_compat_r.
  apply IZR_le; omega.

  elim (Z_eq_dec k1 (-2 * b + 1)).
  intros Hk1eq.
  replace ((IZR k1 + r) / IZR (2 * b) + v2 * v3)%R with 
    ((IZR k1 + (r+IZR (2*b)*v2*v3))/(IZR (2*b)))%R; 
    [idtac| newfield; discrR; omega].
  assert (Hsumbound: (-1 <= r + IZR (2 * b) * v2 * v3 <= 1)%R).
  clear mult_correct_body.
  split.
  subst.
  apply Rplus_le_reg_l with (-IZR (2*b)+ 1)%R.
  replace (-IZR (2 * b) + 1 + -1)%R with (-1*IZR (2 * b))%R; [idtac| ring]. 
  replace (-IZR (2 * b) + 1 + (r + IZR (2 * b) * v2 * v3))%R with 
    (((IZR (-2 * b + 1) + r) / IZR (2 * b) + v2 * v3)*IZR (2 * b))%R; [idtac|
      rewrite plus_IZR; repeat rewrite mult_IZR; simpl; newfield; discrR; omega].
  apply Rmult_le_compat_r.
  apply (IZR_le 0); omega.
  dec_tac; assumption.
  apply Rle_trans with (1:= Hsumneg); fourier.
  constructor; try assumption.
  apply mult_correct_body; clear mult_correct_body; try assumption.
  rewrite mult_IZR; simpl; assumption.

  intros Hk1neq.
  replace ((IZR k1 + r) / IZR (2 * b) + v2 * v3)%R with 
    ((IZR (k1-1) + ((r+1)+IZR (2*b)*v2*v3))/(IZR (2*b)))%R; [idtac| 
      rewrite <- Z_R_minus; change (IZR 1) with 1%R; newfield; discrR; omega].
  assert (Hsumbound:   (-1 <= r + 1 + IZR (2 * b) * v2 * v3 <= 1)%R).
  clear mult_correct_body.
  split.
  replace (-1)%R with (0+-1)%R; [idtac|ring].
  apply Rplus_le_compat.
  dec_tac; fourier.
  rewrite mult_IZR; simpl.
  generalize Hbound; intros [Hbound1 _].
  apply Rle_trans with (2:= Hbound1).
  apply Rmult_le_reg_l with (4 * IZR b)%R.
  assumption.
  replace (4 * IZR b * (-1 / (4 * IZR b)))%R with (IZR (-1))%R;
    [ idtac| simpl; newfield;discrR; auto with zarith].
  replace (4 * IZR b * -1)%R with (IZR (-4*b)); [idtac | 
    rewrite mult_IZR; simpl; ring].
  apply IZR_le; omega.  
  replace ( r + 1 + IZR (2 * b) * v2 * v3)%R with 
    (r+ IZR (2 * b) * v2 * v3 +1)%R;[idtac|ring].
  pattern 1%R at -1; replace 1%R with (0+1)%R; [idtac|ring].
  apply Rplus_le_compat_r; assumption.
  constructor; try assumption.
  apply mult_correct_body; clear mult_correct_body; try assumption.
  inversion_to_variable x2.
  replace ((IZR k2 + r0) / IZR (2 * b) + 1)%R with 
    ((IZR (k2 +2*b) +r0)/IZR (2 * b))%R; 
    [idtac| rewrite plus_IZR; newfield; discrR; auto with zarith].
  constructor; try assumption; omega.
  rewrite mult_IZR; simpl; assumption.
  omega.

  intros Hk2gt Hk2lt.
  inversion Hv1.
  replace ((IZR k1 + r) / IZR (2 * b) + v2 * v3)%R with 
    ((IZR k1 + (r+IZR (2*b)*v2*v3))/(IZR (2*b)))%R; 
    [idtac| newfield; discrR; omega].
  assert (Hsumbound: (-1 <= r + IZR (2 * b) * v2 * v3 <= 1)%R).
  clear mult_correct_body.
  generalize (I_interval _ _ _ _ Hbpos H1).
  intros.
  rewrite mult_IZR; simpl.
  split.
  apply Rle_trans with ((IZR k2 - 1) / IZR (2 * b) + -1 / (4 * IZR b))%R.
  apply Rmult_le_reg_l with (4 * IZR b)%R. 
  assumption.
  replace (4 * IZR b * -1)%R with (IZR (-4*b)); [idtac | 
    rewrite mult_IZR; simpl; ring].
  replace (4 * IZR b * ((IZR k2 - 1) / IZR (2 * b) + -1 / (4 * IZR b)))%R with
    (IZR (2*k2-3))%R; [ idtac|  rewrite <- Z_R_minus; repeat rewrite mult_IZR;
      simpl; newfield;discrR; auto with zarith].
  apply IZR_le; omega.
  dec_tac; apply Rplus_le_compat; assumption.
  apply Rle_trans with ((IZR k2 + 1) / IZR (2 * b) + 1 / (4 * IZR b))%R.
  dec_tac; apply Rplus_le_compat; assumption.
  apply Rmult_le_reg_l with (4 * IZR b)%R. 
  assumption.
  replace (4 * IZR b * 1)%R with (IZR (4*b)); [idtac | 
    rewrite mult_IZR; simpl; ring].
  replace (4 * IZR b * ((IZR k2 + 1) / IZR (2 * b) + 1 / (4 * IZR b)))%R with
    (IZR (2*k2+3))%R; [ idtac|  rewrite plus_IZR; repeat rewrite mult_IZR;
      simpl; newfield;discrR; auto with zarith].
  apply IZR_le; omega.
  constructor; try assumption.
  apply mult_correct_body; clear mult_correct_body; try assumption.
  rewrite mult_IZR; simpl; assumption.
Defined.  

Ltac force_represents :=
 match goal with
  id: represents _ ?x ?y |- represents _ ?x ?z =>
  replace z with y;[assumption | newfield; discrR; auto with zarith]
 end.

Theorem mult_a_correct :
  forall b, 1 < b ->
    forall v1 v2 v3 x u u', 
      (-1/(4*IZR b) <= v3 <= 1/(4*IZR b))%R ->
      (-1 <=v1 +v2*v3 <= 1)%R ->
      represents _ x v1 ->
      represents _ u v2 ->
      represents _ u' v3 ->
      represents _ (mult_a b x (u, u')) (v1+v2*v3)%R.
Proof.
cofix.
intros b H1ltb v1 v2 v3 x u u' Hv3bound Hsum Hv1 Hv2 Hv3.
rewrite (str_dec_thm (mult_a b x (u, u'))).
lazy beta iota zeta delta [str_decompose mult_a]; fold mult_a.

destruct u as [ [k] u1].
fold (str_decompose (series_body' b (stream (idigit b) * stream (idigit b)) 
         (mult_a b) (add b x (multkdiv2b b k u')) (u1, u'))).
rewrite <- str_dec_thm.
inversion Hv2.
replace ( (v1 + (IZR k + r) / IZR (2 * b) * v3))%R with 
  ((v1 + (IZR k * v3) /IZR (2 * b) + r / IZR (2 * b) * v3))%R;
  [idtac|newfield; discrR; omega].
apply mult_correct_body; try assumption. 
apply mult_a_correct; clear mult_a_correct; assumption.
clear mult_a_correct.
replace (v1+IZR k * v3 / IZR (2 * b) + r / IZR (2 * b) * v3)%R with 
  (v1+v2*v3)%R; [assumption | subst; newfield; discrR; omega].
clear mult_a_correct.
apply add_correct; try assumption.
omega.
apply multkdiv2b_correct; try assumption; omega.
2:rewrite mult_IZR; simpl; force_represents.


subst.
generalize Hsum; clear Hsum; intros [Hsum1 Hsum2].
split.
apply Rle_trans with (1:=Hsum1).
apply Rplus_le_compat_l.








fold (str_decompose (series_body (stream idigit * stream idigit) mult_a (x + (L:: u')) (u1, u'))).
rewrite <- str_dec_thm.
inversion Hv2.
match goal with |- represents _ ?a => 
  replace a with ((v1+v3/2)+r/2*v3)%R; 
    [idtac | clear mult_a_correct; field; auto with real1]
end.
apply mult_correct_body; (exact mult_a_correct || clear mult_a_correct); auto.
replace (v1 + v3/2 + r/2 *v3)%R with (v1 + v2*v3)%R;
  [assumption | subst v2; field; auto with real1].
apply add_correct.
assumption.
constructor.
assumption.
interval_from_represents Hv3; auto.
apply Rle_trans with (2:= Hsum).
apply Rplus_le_compat_l.
replace (v3/2)%R with (1/2*v3)%R.
apply Rmult_le_compat_r.
interval_from_represents Hv3; intros; dec_tac; auto.
subst v2; autorewrite with real1.
dec_tac; fourier.
field; auto with real1.
force_represents.

fold (str_decompose (series_body (stream idigit * stream idigit) mult_a (x + (L::L:: u')) (u1, u'))).
rewrite <- str_dec_thm.
inversion Hv2.

match goal with |- represents _ ?a => 
  replace a with ((v1+(v3/2/2))+r/2*v3)%R; 
    [idtac | clear mult_a_correct; field; auto with real1]
end.
apply mult_correct_body; (exact mult_a_correct || clear mult_a_correct);auto.
replace (v1 + v3/2/2 + r/2 *v3)%R with (v1 + v2*v3)%R;
  [assumption | subst v2; field; auto with real1].
apply add_correct.
assumption.
constructor.
constructor.
assumption.
interval_from_represents Hv3; auto.
interval_from_represents Hv3; intros; dec_tac; split; fourier.
apply Rle_trans with (2:= Hsum).
apply Rplus_le_compat_l.
replace (v3/2/2)%R with (1/4*v3)%R.
apply Rmult_le_compat_r.
interval_from_represents Hv3; intros; dec_tac; fourier.
subst v2; autorewrite with real1.
dec_tac; fourier.
field; auto with real1.
force_represents.
Qed.

Theorem zero_correct : represents zero 0%R.
cofix.
rewrite (str_dec_thm zero); simpl.
replace 0%R with (0/2)%R.
constructor.
apply zero_correct.
clear zero_correct.
split; fourier.
clear zero_correct.
unfold Rdiv; ring.
Qed.

Theorem mult_correct :
  forall x y vx vy, represents x vx -> represents y vy ->
    represents (x * y) (vx * vy).
intros x y vx vy Hx Hy.
replace (vx * vy)%R with (2*(2*(vx * (vy/2 /2))))%R.
unfold mult.
interval_from_represents Hx; interval_from_represents Hy; intros; dec_tac.
assert (vx * (vy/2/2) <= 1/4)%R.
replace (vy/2/2)%R with (1/4*vy)%R;[idtac|field;auto with real1].
replace (1/4)%R with (1* (1/4*1))%R.
apply Rmult_le_compat.
assumption.
repeat apply Rmult_le_pos.
fourier.
fourier.
fourier.
assumption.
assumption.
apply Rmult_le_compat.
repeat apply Rmult_le_pos; fourier.
assumption.
rewrite Rmult_1_l; rewrite Rmult_1_r.
auto with real.
assumption.
ring.
apply mult2_correct.
split.
repeat apply Rmult_le_pos; try fourier.
auto with real1.
auto with real1.
apply mult2_correct.
split.
repeat apply Rmult_le_pos; try fourier.
auto with real1.
auto with real1.
replace (vx * (vy / 2/2))%R with (0 + vx * (vy/2/2))%R;[idtac|ring].
apply mult_a_correct.
replace (vy/2/2)%R with (1/4*vy)%R;[idtac | field; auto with real1].
fourier.
fourier.
apply zero_correct.
assumption.
constructor.
constructor.
assumption.
auto.
auto with real1.
field; auto with real1.
Qed.


CoFixpoint abs_idigit (b: Z) (x : stream (idigit b)) : (stream (idigit b)) := 
  match x with 
    | (I k)::x' => (I b (Zabs k))::abs_idigit b x'
  end.

CoInductive abs_bound (b:Z) :  stream (idigit b) -> Prop :=
  C: forall x k, 
    (forall u u', represents b ((I b::x) u -> 
      represents b (abs_idigit b ((I b k)::x)) u' -> (-u'<=u<=u')%R) ->
    abs_bound b x -> abs_bound b ((I b k)::x).

Theorem abs_bound_ex:
  forall b x,
  0 < b ->
  abs_bound b x.
Proof.
cofix.
intros b x Hbpos.
case x; clear x. 
intros i x'.
case i; clear i.
intros k.
constructor.
generalize (abs_bound_ex b x' Hbpos).
intros U.
split.

intros u u'.
inversion 1.
inversion 1.
str_unfold (abs_idigit b (I b k ::I b k0:: x)).
intros Hrep.
simpl in Hrep.
inversion Hrep; clear Hrep.
generalize (H0 r r0 H6 H12).
intros [G1 G2].
unfold Rdiv.
replace (- ((IZR (Zabs k) + r0) */ IZR (2 * b)))%R with 
  ((-IZR (Zabs k) + -r0) */ IZR (2 * b))%R; [idtac|ring].
split; (apply Rmult_le_compat_r;
  [apply Rlt_le; apply Rinv_0_lt_compat; apply (IZR_lt 0); omega|idtac]);
(apply Rplus_le_compat; try assumption);
try rewrite <- Ropp_Ropp_IZR;
apply IZR_le;
(elim (Z_le_gt_dec k 0);[intros Hkneg; rewrite Zabs_non_eq; omega|
  intros Hkpos; rewrite Zabs_eq; omega]).
apply abs_bound_ex; assumption.
CoInductive abs_bound (b:Z) :  stream (idigit b) -> Prop :=
  C: forall x k, 
    (forall u u', represents b ((I b k)::x) u -> 
      represents b (abs_idigit b ((I b k)::x)) u' -> (-u'<=u<=u')%R) ->
    abs_bound b x -> abs_bound b ((I b k)::x).

Theorem abs_bound_ex:
  forall b x,
  0 < b ->
  abs_bound b x.
Proof.
cofix.
intros b x Hbpos.
case x; clear x. 
intros i x'.
case i; clear i.
intros k.
constructor.
generalize (abs_bound_ex b x' Hbpos).
intros U.
split.

intros u u'.
inversion 1.
inversion 1.
str_unfold (abs_idigit b (I b k ::I b k0:: x)).
intros Hrep.
simpl in Hrep.
inversion Hrep; clear Hrep.
generalize (H0 r r0 H6 H12).
intros [G1 G2].
unfold Rdiv.
replace (- ((IZR (Zabs k) + r0) */ IZR (2 * b)))%R with 
  ((-IZR (Zabs k) + -r0) */ IZR (2 * b))%R; [idtac|ring].
split; (apply Rmult_le_compat_r;
  [apply Rlt_le; apply Rinv_0_lt_compat; apply (IZR_lt 0); omega|idtac]);
(apply Rplus_le_compat; try assumption);
try rewrite <- Ropp_Ropp_IZR;
apply IZR_le;
(elim (Z_le_gt_dec k 0);[intros Hkneg; rewrite Zabs_non_eq; omega|
  intros Hkpos; rewrite Zabs_eq; omega]).
apply abs_bound_ex; assumption.

*)
(*
Definition tl (b : Z) (x: stream (idigit b)) : stream (idigit b) :=
  match x with 
    | _::x' => x'
  end.

CoInductive abs_bound' (b:Z) :  stream (idigit b) -> Prop :=
  C': forall x, 
    (forall u u', represents b x u -> 
      represents b (abs_idigit b x) u' -> (-u'<=u<=u')%R) -> 
    abs_bound' b (tl b x) -> abs_bound' b x.

Theorem abs_bound_ex':
  forall b x,
  0 < b ->
  abs_bound' b x.
Proof.
cofix.
intros b x Hbpos.
constructor.
destruct x as [[k] x'].



case x; clear x. 
intros i x'.
case i; clear i.
intros k.
constructor.
intros u u'.
inversion 1.
inversion 1.
str_unfold (abs_idigit b (I b k ::I b k0:: x)).
intros Hrep.
simpl in Hrep.
inversion Hrep; clear Hrep.
generalize (H0 r r0 H6 H12).
intros [G1 G2].
unfold Rdiv.
replace (- ((IZR (Zabs k) + r0) */ IZR (2 * b)))%R with 
  ((-IZR (Zabs k) + -r0) */ IZR (2 * b))%R; [idtac|ring].
split; (apply Rmult_le_compat_r;
  [apply Rlt_le; apply Rinv_0_lt_compat; apply (IZR_lt 0); omega|idtac]);
(apply Rplus_le_compat; try assumption);
try rewrite <- Ropp_Ropp_IZR;
apply IZR_le;
(elim (Z_le_gt_dec k 0);[intros Hkneg; rewrite Zabs_non_eq; omega|
  intros Hkpos; rewrite Zabs_eq; omega]).
apply abs_bound_ex; assumption.
*)

