Require Export realbasis.
Require Export proof_utils.
Require Export Field_tac.

Definition str_decompose (A:Set)(x:stream A) :=
  match x with Cons a tl => Cons a tl end.

Implicit Arguments str_decompose.

Theorem str_dec_thm : forall A:Set, forall x : stream A,
  x = str_decompose x.
intros; case x; reflexivity.
Qed.

Implicit Arguments str_dec_thm.

Ltac str_unfold term :=
  rewrite (str_dec_thm term).

Ltac inversion_to_variable x :=
  match goal with
  | id : represents _ x _ |- _ => idtac
  | id : represents _ (Cons ?d x) _ |- context [Cons ?d x] => idtac
  | id : represents _ (Cons ?d1 x) _ |- context [Cons ?d2 x] => inversion id
  | id : represents _ (Cons ?d1 (Cons ?d2 x)) _ |- context [ Cons ?d3 x] =>
    inversion id; inversion_to_variable x
  end.

(*
Theorem redondance_ex1 : 
  forall b x k r,
    0<b ->
    -2*b+1 < k ->
    represents b ((I b k)::(I b (-2*b+1))::x) r ->
    represents b ((I b (k-1))::(I b 1)::x) r.
Proof.
intros b x k r Hbpos Hkgt Hrep.
Opaque Zmult.
inversion_to_variable x.
Transparent Zmult.
replace ((IZR k + (IZR (-2 * b + 1) + r1) / IZR (2 * b)) / IZR (2 * b))%R with 
  ((IZR (k-1) + ((IZR 1+r1)/IZR (2*b)))/ IZR (2*b))%R.
constructor.
constructor; try assumption; omega.
split.
apply Rle_trans with 0%R.
fourier.
unfold Rdiv.
apply Rle_mult_inv_pos.
simpl; dec_tac; fourier.
apply (IZR_lt 0); omega.




replace (IZR 1 + r1)%R with (IZR (2*b) + (IZR (-2 * b + 1) + r1))%R;[idtac|
  rewrite plus_IZR; repeat rewrite mult_IZR; simpl; ring].
unfold Rdiv in *.
rewrite Rmult_plus_distr_r.
rewrite H9.
rewrite <- Rinv_r_sym.
*)



Theorem IZR_div_ineq1 :
  forall a b c,
    0 <c ->
    a*c<= b ->
    (IZR a <= IZR b/IZR c)%R. 
Proof.
intros a b c Hcpos H.
apply Rmult_le_reg_l with (IZR c).
apply (IZR_lt 0 c Hcpos).
unfold Rdiv.
rewrite <- Rmult_assoc.
rewrite Rinv_r_simpl_m.
rewrite <- mult_IZR.
apply IZR_le.
rewrite Zmult_comm.
exact H.
apply (IZR_neq c 0).
omega.
Qed.

Theorem IZR_div_ineq2 :
  forall a b c,
    0 <c ->
    a<= b*c ->
    (IZR a/IZR c <= IZR b)%R. 
Proof.
intros a b c Hcpos H.
apply Rmult_le_reg_l with (IZR c).
apply (IZR_lt 0 c Hcpos).
unfold Rdiv.
rewrite <- Rmult_assoc.
rewrite Rinv_r_simpl_m.
rewrite <- mult_IZR.
apply IZR_le.
rewrite Zmult_comm.
exact H.
apply (IZR_neq c 0).
omega.
Qed.

Theorem IZR_div_ineq3 :
  forall a b c d,
    0 <b ->
    0 <d ->
    a*b<= c*d ->
    (IZR a/IZR d <= IZR c/IZR b)%R. 
Proof.
intros a b c d Hbpos Hdpos H.
apply Rmult_le_reg_l with (IZR b*IZR d)%R.
rewrite <- mult_IZR.
apply (IZR_lt 0).
apply Zmult_lt_O_compat; assumption.
replace (IZR b * IZR d * (IZR a / IZR d))%R with 
  (IZR b *IZR a)%R;[idtac|newfield; discrR;omega].
replace (IZR b * IZR d * (IZR c / IZR b))%R with 
  (IZR d *IZR c)%R;[idtac|newfield; discrR;omega].
repeat rewrite <- mult_IZR.
apply IZR_le.
rewrite (Zmult_comm b).
rewrite (Zmult_comm d).
exact H.
Qed.

Theorem Zdiv_bound : 
  forall a b, 0<b -> a-b<=a/b*b<=a.
Proof.
intros a b H1.
assert (H: b>0).
omega.
split.
pattern a at 1.
rewrite (Z_div_mod_eq a b H).
replace (a/b*b) with (a/b*b +0)%Z; [idtac|ring].
rewrite Zmult_comm.
replace ( a / b * b + a mod b - b)%Z with 
  (a / b * b + (a mod b - b))%Z; [idtac|ring].
apply Zplus_le_compat_l.
generalize (Z_mod_lt a b H).
intros [_ H2]; omega.
rewrite Zmult_comm.
apply Z_mult_div_ge.
assumption.
Qed.

Theorem Zdiv_IZR_bound : 
  forall a b,
    0<b -> 
    (IZR a/IZR b -1 <=IZR (a/b) <=IZR a/IZR b)%R.
Proof.
intros a b Hbpos.
generalize (Zdiv_bound a b Hbpos).
split; 
(apply Rmult_le_reg_l with (IZR b); [apply (IZR_lt 0 b Hbpos) |idtac]);
(replace (IZR b * IZR (a / b))%R with (IZR (a/b*b)); [idtac |
rewrite Zmult_comm; apply mult_IZR]);
(replace (IZR b * (IZR a / IZR b - 1))%R with 
  (IZR b * (IZR a / IZR b) - IZR b)%R; [idtac|ring]);
(replace (IZR b * (IZR a / IZR b))%R with (IZR a);
  [idtac|newfield; apply (IZR_neq b 0); omega]);
try rewrite Z_R_minus;
apply IZR_le; omega.
Qed.

Theorem Zdiv_bound' : forall a b,
  0<b -> a<=(a+b)/b*b<=a+b.
intros a b H.
pattern a at 1.
replace a with (a+b-b)%Z.
apply Zdiv_bound.
assumption.
ring.
Qed.

Theorem Zdiv_IZR_bound' : 
  forall a b,
    0<b -> 
    (IZR a/IZR b <=IZR ((a+b)/b) <=IZR a/IZR b +1)%R.
Proof.
intros a b Hbpos.
generalize (Zdiv_IZR_bound (a+b) b Hbpos).
replace (IZR (a + b) / IZR b - 1)%R with (IZR a / IZR b)%R.
replace (IZR (a + b) / IZR b)%R with (IZR a/IZR b +1)%R.
tauto.
rewrite plus_IZR; newfield; apply (IZR_neq b 0); omega.
rewrite plus_IZR; newfield; apply (IZR_neq b 0); omega.
Qed.

Lemma IZR_div_le : forall a b, 
  0 < b ->
  (-1 <= IZR a / IZR b <= 1)%R ->
  -b <= a <= b.
Proof.
intros a b Hbpos [H1 H2].
split.

apply le_IZR.
apply Rmult_le_reg_l with (/IZR b).
rewrite <- Rmult_1_l.
apply Rlt_mult_inv_pos.
fourier.
apply (IZR_lt 0 b Hbpos).
rewrite Ropp_Ropp_IZR.
replace (/ IZR b * - IZR b)%R with 
  (-(/ IZR b * IZR b))%R; [idtac|ring].
rewrite Rinv_l.
unfold Rdiv in H1.
rewrite Rmult_comm.
assumption.
discrR; omega.

apply le_IZR.
apply Rmult_le_reg_l with (/IZR b).
rewrite <- Rmult_1_l.
apply Rlt_mult_inv_pos.
fourier.
apply (IZR_lt 0 b Hbpos).
rewrite Rinv_l.
unfold Rdiv in H2.
rewrite Rmult_comm.
assumption.
discrR; omega.
Qed.

Lemma IZR_le_div : forall a b, 
  0 < b ->
  -b <= a <= b ->
  (-1 <= IZR a / IZR b <= 1)%R.
Proof.
intros a b Hbpos [H1 H2].
split;
(apply Rmult_le_reg_l with (IZR b); [apply (IZR_lt 0 b Hbpos)|idtac]);
(replace (IZR b * (IZR a / IZR b))%R 
  with (IZR a); [idtac|newfield; discrR; omega]).
replace (IZR b * -1)%R with (-IZR b)%R; [idtac| ring].
rewrite <- Ropp_Ropp_IZR.
apply IZR_le; assumption.

replace (IZR b * 1)%R with (IZR b)%R; [idtac| ring].
apply IZR_le; assumption.
Qed.

Theorem rts_find_k_correct : 
  forall b a c k, 
    0 < b ->
    0 < c ->
    k = rts_find_k b a c ->
    (-1 <= IZR a / IZR c <= 1)%R ->
    (IZR k <= IZR (a*b)/IZR c <= IZR (k+1))%R
      /\ -b<=k<=b-1.
Proof.  
intros b a c k Hbpos Hcpos Hk H.
subst k; unfold rts_find_k.

assert (Hcpos' : c>0);[omega| idtac].
generalize(Z_mod_lt (a*b) c Hcpos').
intros [Hmodpos Hmodltc].

assert (Hmod: (IZR (a*b) / IZR c = IZR (a*b / c) + IZR ((a*b) mod c) / IZR c)%R).
replace (IZR (a * b / c) + IZR ((a * b) mod c) / IZR c)%R with
  ((IZR c * IZR (a * b / c) + IZR ((a * b) mod c)) / IZR c)%R;
  [idtac|newfield; discrR;omega].
rewrite <- mult_IZR.
rewrite <- plus_IZR.
rewrite <- (Z_div_mod_eq (a*b) c); [reflexivity|omega].

rewrite Hmod.
assert (IZR (a * b / c) <= IZR (a * b / c) + IZR ((a * b) mod c) / IZR c)%R.
pattern (IZR (a * b / c)) at 1.
replace (IZR (a * b / c)) with (IZR (a * b / c) +0)%R; [idtac|ring].
apply Rplus_le_compat_l.
apply Rmult_le_reg_l with (IZR c).
apply (IZR_lt 0 c Hcpos).
replace (IZR c * 0)%R with 0%R; [idtac|ring].
unfold Rdiv; rewrite <- Rmult_assoc.
rewrite Rinv_r_simpl_m; [idtac|discrR; omega].
apply (IZR_le 0).
omega.

elim (Z_le_gt_dec (-b) (a * b / c)).
elim (Z_le_gt_dec (a * b / c) (b-1)).
intros Hle1 Hle2.
split.
split.
assumption.
rewrite plus_IZR.
apply Rplus_le_compat_l.
apply IZR_div_ineq2.
exact Hcpos.
omega.
split; assumption.

intros Hgt _.
split; [idtac|split;omega].
split.
apply Rle_trans with (IZR (a * b / c)).
apply IZR_le.
omega.
assumption.
rewrite <- Hmod.
ring (b-1+1)%Z.
replace (IZR b) with (IZR b *1)%R;[idtac|ring].
rewrite mult_IZR.
rewrite Rmult_comm.
unfold Rdiv in *.
rewrite Rmult_assoc.
apply Rmult_le_compat_l.
apply (IZR_le 0); omega.
dec_tac; assumption.

intros Hgt.
split; [idtac|split;omega].
split.
rewrite <- Hmod.
rewrite mult_IZR.
rewrite Rmult_comm.
unfold Rdiv in *.
rewrite Rmult_assoc.
replace (IZR (-b)) with (IZR b*-1)%R;[idtac|rewrite Ropp_Ropp_IZR; ring].
apply Rmult_le_compat_l.
apply (IZR_le 0); omega.
dec_tac; assumption.
rewrite plus_IZR.
apply Rplus_le_compat.
apply IZR_le; omega.
apply IZR_div_ineq2; omega.
Qed.

Theorem rts_lemma1 : 
  forall  b a c, 
    0<b -> 
    0<c ->
    (-1<=(IZR a/ IZR c)<=1)%R ->
    (-1 <= IZR (2 * a * b - (2 * rts_find_k b a c + 1) * c) / IZR c <= 1)%R.
Proof.
intros b a c Hbpos Hcpos H.
generalize (rts_find_k_correct b a c (rts_find_k b a c) Hbpos Hcpos 
(refl_equal (rts_find_k b a c)) H).
intros [[H1 H2] [H3 H4]].
rewrite <- (Z_R_minus).
match goal with |- (_ <= (?a - IZR (?b *?c)) / IZR ?c <= _)%R => 
  replace ((a - IZR (b *c)) / IZR c)%R with ((a/IZR c) - IZR b)%R
end; [idtac| repeat rewrite mult_IZR; newfield; discrR; omega].
rewrite plus_IZR.
rewrite <- Zmult_assoc.
repeat rewrite (mult_IZR 2).
simpl.
split.
replace (-1)%R with (2*0+-1)%R; [idtac|ring].
replace ( 2 * IZR (a * b) / IZR c - (2 * IZR (rts_find_k b a c) + 1))%R with 
  (2*(IZR (a * b) / IZR c- IZR (rts_find_k b a c)) + -1)%R;
  [idtac|newfield; discrR; omega].
apply Rplus_le_compat_r.
apply Rmult_le_compat_l; fourier.
replace ( 2 * IZR (a * b) / IZR c - (2 * IZR (rts_find_k b a c) + 1))%R with 
  (2*(IZR (a * b) / IZR c- (IZR (rts_find_k b a c) +1)) + 1)%R;
  [idtac|newfield; discrR; omega].
pattern 1%R at 5.
replace 1%R with (2*0+1)%R; [idtac|ring].
apply Rplus_le_compat_r.
pattern 1%R at 3.
replace 1%R with (IZR 1);[idtac|reflexivity].
rewrite <- plus_IZR. 
apply Rmult_le_compat_l; fourier.
Qed.

Theorem rts_lemma2 : 
  forall  b a c, 
    0<b -> 
    c<0 ->
    (-1<=(IZR a/ IZR c)<=1)%R ->
    (-1 <= IZR (2*a*b - (2*rts_find_k b (-a) (-c) + 1) * c) / IZR c <= 1)%R.
Proof.
intros b a c Hbpos Hcneg H.
match goal with |- (_ <= IZR ?a /IZR ?b <= _)%R =>
  replace (IZR a /IZR b)%R with (IZR (-a) /IZR (-b))%R
end.
match goal with |- (_ <= IZR (- (2*?a*?e -?c*?b)) / _ <= _)%R =>
  replace (- (2*a*e -c*b)) with (2*(-a)*e-c*(-b))%Z
end; [idtac|ring].
apply rts_lemma1.
assumption.
omega.
repeat rewrite Ropp_Ropp_IZR.
unfold Rdiv in *.
rewrite <- Ropp_inv_permute.
rewrite Rmult_opp_opp.
assumption.
discrR; omega.

repeat rewrite Ropp_Ropp_IZR.
unfold Rdiv.
rewrite <- Ropp_inv_permute.
apply Rmult_opp_opp.
discrR; omega.
Qed.

Ltac rat_to_str_tac :=
  match goal with 
    |- represents ?e (Cons (I ?e ?k) (rat_to_stream ?e ?a ?b)) ?r =>
      replace r with ((IZR  k + (IZR a/ IZR b))/ IZR (2*e))%R
end.

Theorem rat_to_stream_correct :
  forall b a c, 
    0<b -> 
    0<>c ->
    (-1<=(IZR a/ IZR c)<=1)%R ->
    represents b (rat_to_stream b a c) (IZR a/ IZR c)%R.
cofix.
intros b a c Hbpos Hcneq0 Hbound.
str_unfold (rat_to_stream b a c).
unfold str_decompose.
unfold rat_to_stream.
replace (cofix rat_to_stream (b0 a0 c0 : Z) : stream (idigit b0) :=
            I b0
              (2 *
               (if Z_le_gt_dec 0 c0
                then rts_find_k b0 a0 c0
                else rts_find_k b0 (- a0) (- c0)) + 1)
            :: rat_to_stream b0
                 (2 * a0 * b0 -
                  (2 *
                   (if Z_le_gt_dec 0 c0
                    then rts_find_k b0 a0 c0
                    else rts_find_k b0 (- a0) (- c0)) + 1) * c0) c0)%Z
with rat_to_stream ; [idtac| apply refl_equal]. 
elim (Z_le_gt_dec 0 c).

intros Hle.
rat_to_str_tac.
constructor.
apply rat_to_stream_correct; clear rat_to_stream_correct.
assumption.
assumption.
apply rts_lemma1; try assumption; try omega.
apply rts_lemma1; try assumption; try omega.
assert (Hcpos : 0<c);[omega|idtac].
generalize (rts_find_k_correct b a c (rts_find_k b a c) Hbpos 
  Hcpos (refl_equal (rts_find_k b a c)) Hbound).
omega.
generalize (2*rts_find_k b a c + 1)%Z; intros x.
rewrite <- Z_R_minus.
replace ((IZR (2*a*b) - IZR (x *c)) / IZR c)%R with 
  (IZR (2*a*b)/ IZR c - IZR (x *c) / IZR c)%R;
  [idtac|newfield; discrR; omega].
repeat rewrite mult_IZR.
newfield.
repeat split; discrR; omega.

intros Hgt.
rat_to_str_tac.
constructor.
apply rat_to_stream_correct; clear rat_to_stream_correct.
assumption.
assumption.
apply rts_lemma2; try assumption; try omega.
apply rts_lemma2; try assumption; try omega.
assert (Hcpos : 0< -c);[omega|idtac].
assert (Hbound': (-1 <= IZR (-a) / IZR (-c) <= 1)%R).
repeat rewrite Ropp_Ropp_IZR.
unfold Rdiv.
rewrite <- Ropp_inv_permute.
rewrite Rmult_opp_opp; assumption.
discrR; omega.
generalize (rts_find_k_correct b (-a) (-c) (rts_find_k b (-a) (-c)) Hbpos 
  Hcpos (refl_equal (rts_find_k b (-a) (-c))) Hbound').
omega.
rewrite <- Z_R_minus.
generalize (2*rts_find_k b (-a) (-c) + 1)%Z; intros x.
replace ((IZR (2*a*b) - IZR (x *c)) / IZR c)%R with 
  (IZR (2*a*b)/ IZR c - IZR (x *c) / IZR c)%R;
  [idtac|newfield; discrR; omega].
repeat rewrite mult_IZR.
newfield.
repeat split; discrR; omega.
Qed.

Theorem represents_one :
  forall b,
  0 < b ->
  represents _ (one b) 1%R.
Proof.
cofix.
intros b H.
str_unfold (one b).
Opaque Zmult.
simpl.
Transparent Zmult.
replace (1)%R with ((IZR (2 * b - 1) + 1) / IZR (2 * b))%R.
constructor.
apply represents_one.
assumption.
clear represents_one.
split; fourier.
omega.
replace (IZR (2 * b - 1) + 1)%R with (IZR (2*b)).
field.
apply (IZR_neq (2*b) 0).
omega.
rewrite <- Z_R_minus.
ring.
Qed.

Theorem represents_sub_one :
  forall b,
    0 < b ->
    represents _ (sub_one b) (-1)%R.
Proof.
cofix.
intros b H.
str_unfold (sub_one b).
Opaque Zmult.
simpl.
Transparent Zmult.
replace (-1)%R with ((IZR (-2 * b + 1) + (-1)) / IZR (2 * b))%R.
constructor.
apply represents_sub_one.
assumption.
clear represents_sub_one.
split; fourier.
omega.
replace (IZR (-2 * b + 1) + -1)%R with (- IZR (2*b))%R.
field.
apply (IZR_neq (2*b) 0).
omega.
rewrite plus_IZR.
repeat rewrite mult_IZR.
ring.
rewrite <- Ropp_mult_distr_r_reverse.
apply Rmult_eq_compat_l.
ring.
Qed.

Theorem represents_zero :
  forall b,
    0 < b->
    represents _ (zero b) 0%R.
Proof.
cofix.
intros b H.
str_unfold (zero b).
simpl.
replace 0%R with ((IZR 0 + 0) / IZR (2 * b))%R.
constructor.
apply represents_zero.
assumption.
clear represents_zero.
split; fourier.
omega.
field.
apply (IZR_neq (2*b) 0).
omega.
Qed.

Theorem I_interval :
  forall b x y k,
    0 < b -> 
    represents _ (Cons (I b k) x) y ->
    ((IZR k - 1)/IZR (2*b) <= y <= (IZR k + 1)/IZR (2*b))%R.
Proof.
intros b x y k H0ltb Hrep.
assert (0< IZR (2*b))%R.
apply (IZR_lt 0).
omega.
inversion Hrep.
unfold Rdiv.
unfold Rminus.
dec_tac.
split; apply Rmult_le_compat_r; auto with real.
Qed.

Theorem I_interval' :
  forall b x y k,
    0 < b -> 
    represents _ (Cons (I b k) x) y ->
    ((IZR k - 1) <= y*IZR (2*b) <= (IZR k + 1))%R.
Proof.
intros b x y k H0ltb Hrep.
assert (0< IZR (2*b))%R.
apply (IZR_lt 0).
omega.
inversion Hrep.
unfold Rdiv.
rewrite Rmult_comm.
rewrite <- Rmult_assoc.
rewrite Rinv_r_simpl_m; auto with real.
unfold Rminus.
dec_tac.
split; fourier.
Qed.

Theorem represents_interval : 
  forall b x r, 
    0 < b -> 
    represents b x r -> 
    (-1 <= r <= 1)%R.
Proof.
intros b x r H0ltb Hrep.
assert (0< IZR (2*b))%R.
apply (IZR_lt 0).
omega.
inversion Hrep.
subst.
generalize (I_interval _ _ _ _ H0ltb Hrep).
intros [H3 H4].
split; 
[apply Rle_trans with (2:= H3) |
apply Rle_trans with (1:= H4)];
apply Rmult_le_reg_l with (IZR (2 * b))%R; auto;
unfold Rdiv;
rewrite <- Rmult_assoc;
rewrite Rinv_r_simpl_m; auto with real.
unfold Rminus.
change (-1)%R with (IZR (-1)%Z)%R.
rewrite <- plus_IZR.
rewrite <- mult_IZR.
apply IZR_le.
omega.
change 1%R with (IZR 1)%R.
rewrite <- plus_IZR.
rewrite <- mult_IZR.
apply IZR_le.
omega.
Qed.

Theorem bounds_correct : 
  forall n e s a b c r,
    0 < e ->
    (a, b, c) = bounds e n s ->
    represents e s r -> 
    (IZR a <= IZR ((2*e)^c) * r <= IZR b)%R.
Proof.
induction n.
intros e s a b c r Hepos Hbounds Hrep.
inversion Hbounds.
inversion Hrep.
change (IZR (-1)) with (-1)%R.
change (IZR 1) with 1%R.
replace (IZR ((2 * e) ^ 0)) with 1%R.
rewrite H6.
ring (1*r)%R.
apply (represents_interval _ _ _ Hepos Hrep).
reflexivity.
intros e s a b c r Hepos Hbounds Hrep.
destruct s as [[k] s'].
inversion Hrep.
Admitted.

Ltac rewrite_all H :=
 match type of H with
 | ?t1 = ?t2 =>
   let rec aux H :=
     match goal with
     | id : context [t1] |- _ =>
       match type of id with
       | t1 = t2 => fail 1
       | _ => generalize id;clear id; try aux H; intro id
       end
     | _ => try rewrite H
     end in
   aux H
 end. 

Ltac rep_constr hyp:=
  match goal with
    |- represents ?e ((I ?e ?k)::?s) ?r =>
      replace r with ((IZR k +(r*IZR(2*e)-IZR k))/ IZR (2*e))%R
  end; [constructor| clear hyp;newfield;discrR;omega].

Theorem mult2_correct_aux1:
  forall b r0,
    0 < b ->
    (-1 <= r0 <= 1 ->
      -1 <= (IZR 0 + r0) / IZR (2 * b) <= 1 ->
      -1 / 2 <= (IZR (- b) + (IZR 0 + r0) / IZR (2*b)) / IZR (2*b) <= 1 / 2 ->
      -1 / 2 <= (IZR (- b) + r0) / IZR (2 * b) <= 1 / 2)%R.
Proof.
change (IZR 0) with 0%R.
intros b r0.
ring (0 +r0)%R.
unfold Rdiv.
intros; dec_tac. 
assert (0<=r0<= 2*IZR b)%R.
split.
apply Rmult_le_reg_l with (/IZR(2*b))%R.
apply Rinv_0_lt_compat; apply (IZR_lt 0); omega.
rewrite Rmult_0_r.
rewrite Rmult_comm.
apply Rplus_le_reg_l with (IZR (- b)).
rewrite Rplus_0_r .
apply Rmult_le_reg_l with (/IZR(2*b))%R.
apply Rinv_0_lt_compat; apply (IZR_lt 0); omega.
repeat rewrite (Rmult_comm (/ IZR (2 * b))).
replace (IZR (- b) * / IZR (2 * b))%R with (-1*/2)%R.
assumption.
rewrite mult_IZR.
rewrite Ropp_Ropp_IZR.
change (IZR 2) with 2%R.
newfield; discrR; omega.
apply Rle_trans with (1:=H6).
rewrite <- (mult_IZR 2); apply (IZR_le 1); omega.
dec_tac; split;
(apply Rmult_le_reg_l with (IZR (2*b)); [apply (IZR_lt 0); omega| idtac]);
repeat rewrite <- Rmult_assoc;
(rewrite Rinv_r_simpl_m; [ idtac| discrR; omega]);
rewrite mult_IZR; change (IZR 2) with 2%R;
rewrite Ropp_Ropp_IZR;
[replace (2 * IZR b * -1 * / 2)%R with (- IZR b)%R; [idtac|newfield] |
replace (2 * IZR b * 1 * / 2)%R with ( IZR b)%R; [idtac|newfield]];
fourier.
Qed.

Theorem mult2_correct_aux2:
  forall b r0,
    0 < b ->
    (-1 <= r0 <= 1 ->
      -1 <= (IZR 0 + r0) / IZR (2 * b) <= 1 ->
      -1 / 2 <= (IZR b + (IZR 0 + r0) / IZR (2*b)) / IZR (2*b) <= 1 / 2 ->
      -1 <= 2 * ((IZR b + r0) / IZR (2 * b)) <= 1)%R.
Proof.
change (IZR 0) with 0%R.
intros b r0.
ring (0 +r0)%R.
unfold Rdiv.
intros; dec_tac. 
assert (-2*IZR b<=r0<= 0)%R.
split.

apply Rle_trans with (2:=H1).
rewrite <- (mult_IZR (-2)). 
change (-1)%R with (IZR (-1)); apply IZR_le; omega.
apply Rmult_le_reg_l with (/IZR(2*b))%R.
apply Rinv_0_lt_compat; apply (IZR_lt 0); omega.
rewrite Rmult_0_r.
rewrite Rmult_comm.
apply Rplus_le_reg_l with (IZR b).
rewrite Rplus_0_r .
apply Rmult_le_reg_l with (/IZR(2*b))%R.
apply Rinv_0_lt_compat; apply (IZR_lt 0); omega.
repeat rewrite (Rmult_comm (/ IZR (2 * b))).
replace (IZR b * / IZR (2 * b))%R with (1*/2)%R.
assumption.
rewrite mult_IZR.
change (IZR 2) with 2%R.
newfield; discrR; omega.
dec_tac; split;
(apply Rmult_le_reg_l with (IZR b); [apply (IZR_lt 0); omega| idtac]);
repeat rewrite <- Rmult_assoc;
(replace (IZR b * 2 * (IZR b + r0) * / IZR (2 * b))%R with (IZR b + r0)%R; 
  [idtac| rewrite mult_IZR; change (IZR 2) with 2%R; newfield; discrR; omega]);
[replace (IZR b * -1 )%R with (- IZR b)%R; [idtac|newfield] |
replace (IZR b * 1)%R with ( IZR b)%R; [idtac|newfield]];
fourier.
Qed.

Theorem mult2_correct_aux3:
  forall b r,
    0 < b ->
    ((IZR 0 - 1) / IZR (2 * b) <= r <= (IZR 0 + 1) / IZR (2 * b) ->
      -1/2<=r<=1/2)%R.
Proof.
intros b r Hbpos.
assert  (/ IZR b<=1)%R.
replace 1%R with (/IZR 1)%R ; [idtac| simpl; newfield].
assert (H:1=b \/ 1 <b).
omega.
elim H.
intros; subst b; auto with real.
intros Hlt.
apply Rlt_le.
apply Rinv_lt_contravar.
rewrite <- mult_IZR.
apply (IZR_lt 0); omega.
apply IZR_lt; omega.
change (IZR 0) with 0%R.
repeat rewrite mult_IZR.
change (IZR 2) with 2%R.
replace ((0 - 1) / (2 * IZR b))%R with 
  (1/2 * -/IZR b)%R; [idtac|newfield;discrR; omega].
replace ((0 + 1) / (2 * IZR b))%R with 
  (1/2 * /IZR b)%R; [idtac|newfield;discrR; omega].
intros [U1 U2].
replace (1/2)%R with (1/2*1)%R; [idtac|ring].
replace (-1/2)%R with ((1/2)*-1)%R; [idtac|newfield].
split; [apply Rle_trans with (2:=U1)| apply Rle_trans with (1:=U2)];
(apply Rmult_le_compat_l; [fourier| idtac]); fourier.
Qed.

Theorem mult2_correct_aux4:
  forall b r k,
    0 < b ->
    k>0 -> 
    k<=2*b-1 ->
    (-1<= r <= 1 ->
      -1 / 2 <= (IZR (k - b) + r) / IZR (2 * b) <= 1 / 2)%R.
Proof.
intros b r k Hbpos Hkpos Hkle Hrbound.
assert (U1: (1 <= IZR k)%R).
apply (IZR_le 1); omega.
generalize (IZR_le _ _ Hkle).
clear Hkle; intros Hkle.
rewrite <- Z_R_minus in Hkle.
rewrite mult_IZR in Hkle.
change (IZR 2) with 2%R in Hkle.
change (IZR 1) with 1%R in Hkle.
dec_tac; unfold Rdiv; split;
(apply Rmult_le_reg_l with (IZR (2*b)); [apply (IZR_lt 0); omega| idtac]);
repeat rewrite <- Rmult_assoc;
(rewrite Rinv_r_simpl_m; [ idtac| discrR; omega]);
rewrite mult_IZR; change (IZR 2) with 2%R;
rewrite <- Z_R_minus;
[replace (2 * IZR b * -1 * / 2)%R with (- IZR b)%R; [idtac|newfield] |
replace (2 * IZR b * 1 * / 2)%R with ( IZR b)%R; [idtac|newfield]]; 
fourier.
Qed.

Theorem mult2_correct_aux5:
  forall b r k,
    0 < b ->
    k<0 -> 
    -2*b+1 <= k ->
    (-1<= r <= 1 ->
      -1 / 2 <= (IZR (k + b) + r) / IZR (2 * b) <= 1 / 2)%R.
Proof.
intros b r k Hbpos Hkpos Hkle Hrbound.
assert (U1: (IZR k <= -1)%R).
apply (IZR_le k (-1)); omega.
generalize (IZR_le _ _ Hkle).
clear Hkle; intros Hkle.
rewrite plus_IZR in Hkle.
rewrite mult_IZR in Hkle.
change (IZR (-2)) with (-2)%R in Hkle.
change (IZR 1) with 1%R in Hkle.
dec_tac; unfold Rdiv; split;
(apply Rmult_le_reg_l with (IZR (2*b)); [apply (IZR_lt 0); omega| idtac]);
repeat rewrite <- Rmult_assoc;
(rewrite Rinv_r_simpl_m; [ idtac| discrR; omega]);
rewrite mult_IZR; change (IZR 2) with 2%R;
rewrite plus_IZR;
[replace (2 * IZR b * -1 * / 2)%R with (- IZR b)%R; [idtac|newfield] |
replace (2 * IZR b * 1 * / 2)%R with ( IZR b)%R; [idtac|newfield]]; 
fourier.
Qed.

Theorem mult2_correct :
  forall b x u, 
    0 < b ->
    (-1/2 <= u <= 1/2)%R -> 
    represents b x u ->
   represents b (mult2 b x) (2*u)%R.
Proof.
cofix.
intros b x u Hbpos H Hrep.
str_unfold (mult2 b x).
destruct x as [[k] x'].
Opaque Zmult.
simpl.
Transparent Zmult.
elim (Z_le_gt_dec k (-b-1)).

(* case I k, k <=-b-1  :  x <= -1/2*)
clear mult2_correct.
generalize (str_dec_thm (sub_one b)); unfold str_decompose.
intros Hdec H1; rewrite <- Hdec; clear Hdec.
replace (2*u)%R with (-1)%R.
apply represents_sub_one; assumption.
apply Rle_antisym.
dec_tac; fourier.
inversion Hrep.
unfold Rdiv.
rewrite mult_IZR.
apply Rmult_le_reg_l with (IZR b).
apply (IZR_lt 0); assumption.
change (IZR 2) with 2%R.
replace  (IZR b * (2 * ((IZR k + r) * / (2 * IZR b))))%R with 
  (IZR k + r)%R; [idtac|newfield; discrR; omega].
apply Rplus_le_reg_l with (-r)%R.
ring (- r + (IZR k + r))%R.
apply Rle_trans with (IZR (- b - 1)).
apply IZR_le; assumption.
rewrite <- Z_R_minus.
replace (IZR (- b) - IZR 1)%R with (-1+-IZR b)%R; 
  [idtac |  rewrite Ropp_Ropp_IZR; ring].
dec_tac; fourier.

elim (Z_le_gt_dec (b+1) k).
(* case I k, k => b+1  :  x <= -1/2*)
clear mult2_correct.
intros H1 H2.
generalize (str_dec_thm (one b)); unfold str_decompose.
intros Hdec; rewrite <- Hdec; clear Hdec.
replace (2*u)%R with 1%R.
apply represents_one; assumption.
apply Rle_antisym.
inversion Hrep.
unfold Rdiv.
rewrite mult_IZR.
apply Rmult_le_reg_l with (IZR b).
apply (IZR_lt 0); assumption.
change (IZR 2) with 2%R.
replace  (IZR b * (2 * ((IZR k + r) * / (2 * IZR b))))%R with 
  (IZR k + r)%R; [idtac|newfield; discrR; omega].
apply Rplus_le_reg_l with (-r)%R.
ring (- r + (IZR k + r))%R.
apply Rle_trans with (IZR ( b + 1)).
rewrite plus_IZR.
simpl.
dec_tac; fourier.
apply IZR_le; assumption.
dec_tac; fourier.

destruct x' as [[k'] x''].
intros H3 H4.
assert (H1 : -b <= k);[omega|idtac].
assert (H2 : k <= b);[omega|idtac].
clear H3 H4.

(*case k' =0*)
elim (Z_le_gt_dec k' 0).
elim (Z_le_gt_dec 0 k').
intros H3 H4.
generalize (Zle_antisym _ _ H3 H4).
clear H3 H4; intros; subst k'.

(*case k' =0 and k = -b :  -1/2 - 1/2e <= x <= -1/2 + 1/2e *)
elim (Z_le_gt_dec k (- b)).
intros H3.
generalize (Zle_antisym _ _ H1 H3).
intros; subst k; clear H1 H2 H3.

inversion_to_variable x''.
replace (2 * ((IZR (- b) + (IZR 0 + r0) / IZR (2 * b)) / IZR (2 * b)))%R with 
  ((IZR (-2*b+1)+2*((IZR (-b)+r0)/(IZR (2*b))))/(IZR (2*b)))%R; [idtac |
rewrite plus_IZR; repeat rewrite mult_IZR; rewrite Ropp_Ropp_IZR;
change (IZR 0) with 0%R; change (IZR 1) with 1%R;
change (IZR 2) with 2%R; change (IZR (-2)) with (-2)%R;
newfield; discrR; omega].
assert (-1 / 2 <= (IZR (- b) + r0) / IZR (2 * b) <= 1 / 2)%R.
subst.
apply mult2_correct_aux1; assumption.
constructor.
apply mult2_correct.
assumption.
assumption.
constructor; assumption.
clear mult2_correct.
dec_tac; split; fourier.
omega.

elim (Z_le_gt_dec b k).
intros H3.
generalize (Zle_antisym _ _ H2 H3).
intros U _ ; subst k; clear H1 H2 H3.

(*case k'=0 and 
  k = b :  1/2 - 1/2e <= x <= 1/2 + 1/2e*)
inversion_to_variable x''.
replace (2 * ((IZR b + (IZR 0 + r0) / IZR (2 * b)) / IZR (2 * b)))%R with 
  ((IZR (2*b-1)+2*((IZR b+r0)/(IZR (2*b))))/(IZR (2*b)))%R; [idtac |
rewrite <- Z_R_minus; repeat rewrite mult_IZR;
change (IZR 0) with 0%R; change (IZR 1) with 1%R;
change (IZR 2) with 2%R; change (IZR (-2)) with (-2)%R;
newfield; discrR; omega].
assert (-1 <= 2* ((IZR b + r0) / IZR (2 * b)) <= 1)%R.
subst.
apply mult2_correct_aux2; assumption.
constructor.
apply mult2_correct.
assumption.
clear mult2_correct.
dec_tac; split; fourier.
constructor; assumption.
assumption.
omega.

(*case k' = 0 and -b < k < b*)
inversion_to_variable x''.
replace (2 * ((IZR k + r) / IZR (2 * b)))%R with 
  ((IZR (2*k) + 2*r) / IZR (2 * b))%R; 
  [idtac| rewrite mult_IZR; change (IZR 2) with 2%R; newfield; discrR; omega].
assert (U1 : (-1 / 2 <= r <= 1 / 2)%R).
generalize (I_interval _ _ _ _ Hbpos H4).
intros.
apply (mult2_correct_aux3 b); assumption.
constructor.
apply mult2_correct; assumption.
clear mult2_correct.
dec_tac; split; fourier.
omega.

(*case k'<=1*)
elim (Z_le_gt_dec k (- b)).
intros H3.
generalize (Zle_antisym _ _ H1 H3).
clear H1 H2 H3; intros H2; subst k.

(*case k=-b and k' <= 1  : x <= -1/2*)
clear mult2_correct.
generalize (str_dec_thm (sub_one b)); unfold str_decompose.
intros Hdec Hk'neg _; rewrite <- Hdec; clear Hdec.
replace (2*u)%R with (-1)%R.
apply represents_sub_one; assumption.
apply Rle_antisym.
dec_tac; fourier.
inversion Hrep.
match goal with 
  id : represents _ (_ :: x'') _ |- _ => inversion id
end.
apply Rmult_le_reg_l with (IZR b).
apply (IZR_lt 0); assumption.
unfold Rdiv.
rewrite mult_IZR.
change (IZR 2) with 2%R.
replace (IZR b * (2*((IZR (-b) + (IZR k' + r0) * / (2*IZR b)) * / (2*IZR b))))%R with
  (IZR (- b) + (IZR k' + r0) * / (2 * IZR b))%R; [idtac|newfield; discrR; omega].
replace (IZR b * -1)%R with (IZR (-b) + 0)%R; [apply Rplus_le_compat_l |
  rewrite Ropp_Ropp_IZR; ring].
apply Rmult_le_reg_l with (2*IZR b)%R.
rewrite <- (mult_IZR 2).
apply (IZR_lt 0); omega.
replace (2 * IZR b * ((IZR k' + r0) * / (2 * IZR b)))%R with 
  (IZR k' + r0)%R; [idtac| newfield ; discrR; omega].
replace ( 2 * IZR b * 0)%R with 0%R; [idtac| ring].
assert (U: (IZR k' <= -1)%R).
apply (IZR_le k' (-1)); omega.
dec_tac; fourier.

(*case k>-b and k' <= 1  : x <= -1/2*)
intros Hgt H3 _.
inversion_to_variable x''.
replace (2 * ((IZR k + (IZR k' + r0) / IZR (2 * b)) / IZR (2 * b)))%R with 
  ((IZR (2*k-1)+2*((IZR (k'+b)+r0)/(IZR (2*b))))/(IZR (2*b)))%R ; [idtac |
rewrite plus_IZR; rewrite <- Z_R_minus; repeat rewrite mult_IZR ;
change (IZR 1) with 1%R; change (IZR 2) with 2%R;  newfield; discrR; omega].
assert (-1 / 2 <= (IZR (k' + b) + r0) / IZR (2 * b) <= 1 / 2)%R.
apply mult2_correct_aux5; try assumption; omega.
constructor.
apply mult2_correct; try  assumption.
constructor; try assumption; omega.
clear mult2_correct.
dec_tac; split; fourier.
omega.

(*case k'>=1*)
elim (Z_le_gt_dec b k).
intros H3.
generalize (Zle_antisym _ _ H2 H3).
clear H1 H2 H3; intros H2; subst k.

(*case k=e and k' => 1  : x => 1/2*)
clear mult2_correct.
generalize (str_dec_thm (one b)); unfold str_decompose.
intros Hdec Hk'neg; rewrite <- Hdec; clear Hdec.
replace (2*u)%R with 1%R.
apply represents_one; assumption.
apply Rle_antisym.
inversion Hrep.
match goal with 
  id : represents _ (_ :: x'') _ |- _ => inversion id
end.
apply Rmult_le_reg_l with (IZR b).
apply (IZR_lt 0); assumption.
unfold Rdiv.
rewrite mult_IZR.
change (IZR 2) with 2%R.
replace (IZR b * (2*((IZR b + (IZR k' + r0) * / (2*IZR b)) * / (2*IZR b))))%R with
  (IZR b + (IZR k' + r0) * / (2 * IZR b))%R; [idtac|newfield; discrR; omega].
replace (IZR b * 1)%R with (IZR b + 0)%R; [apply Rplus_le_compat_l | ring].
apply Rmult_le_reg_l with (2*IZR b)%R.
rewrite <- (mult_IZR 2).
apply (IZR_lt 0); omega.
replace (2 * IZR b * ((IZR k' + r0) * / (2 * IZR b)))%R with 
  (IZR k' + r0)%R; [idtac| newfield ; discrR; omega].
replace ( 2 * IZR b * 0)%R with 0%R; [idtac| ring].
assert (U: (1<=IZR k')%R).
apply (IZR_le 1); omega.
dec_tac; fourier.
dec_tac; fourier.

(* case k < b, k' >= 1*)
intros Hgt1 Hgt2.
inversion_to_variable x''.
replace (2 * ((IZR k + (IZR k' + r0) / IZR (2 * b)) / IZR (2 * b)))%R with 
  ((IZR (2*k+1)+2*((IZR (k'-b)+r0)/(IZR (2*b))))/(IZR (2*b)))%R ; [idtac |
rewrite plus_IZR; rewrite <- Z_R_minus; repeat rewrite mult_IZR ;
change (IZR 1) with 1%R; change (IZR 2) with 2%R;  newfield; discrR; omega].
assert (-1 / 2 <= (IZR (k' - b) + r0) / IZR (2 * b) <= 1 / 2)%R.
apply mult2_correct_aux4; try assumption; omega.
constructor.
apply mult2_correct; try  assumption.
constructor; try assumption; omega.
clear mult2_correct.
dec_tac; split; fourier.
omega.
Qed.

(*
Ltac rep_tac := 
  match goal with
    | id : represents ?x ?r |- represents ?x ?r => assumption
    | id : represents ?x ?r1 |- represents ?x ?r2 =>
	replace r1 with r2; [idtac|newfield]
    |id : represents ?x ?r1 |- represents (?d::?x) ?r2 => 
	replace r2 with 
	  match d with 
	    | L => (r1/2 -1/2)%R
	    | C => (r1/2)%R
	    | R => (r1/2 +1/2)%R
	  end; [constructor| newfield]
  end.

Ltac constructor' hyp :=
  match goal with
    |- represents (C::_) ?v =>
    replace v with ((2*v)/2)%R;
         [constructor; [idtac|clear hyp] | clear hyp; newfield]
  | |- represents (L::_) ?v =>
    replace v with ((2*v+1)/2-1/2)%R;
         [constructor; [idtac|clear hyp] | clear hyp; newfield]
  | |- represents (R::_) ?v  =>
    replace v with ((2*v-1)/2+1/2)%R;
         [constructor; [idtac|clear hyp] | clear hyp; newfield]
  end.

Ltac rat_to_str_tac hyp :=
  match goal with 
    |- represents (rat_to_stream ?q ?r) ?w =>
      replace w with (IZR q / IZR r)%R; [apply hyp; clear hyp| clear hyp]
end.

Ltac inversion_to_variable x :=
  match goal with
  | id : represents x _ |- _ => idtac
  | id : represents (Cons ?d x) _ |- context [Cons ?d x] => idtac
  | id : represents (Cons ?d1 x) _ |- context [Cons ?d2 x] => inversion id
  | id : represents (Cons ?d1 (Cons ?d2 x)) _ |- context [ Cons ?d3 x] =>
    inversion id; inversion_to_variable x
  end.


Ltac prepare_represents_constructor :=
  match goal with
    |- represents (Cons R ?a) (?b/2) =>
       replace b with ((b - 1)+1)%R;[idtac | field ; auto with real1]
  | |- represents (Cons L ?a) (?b/2) => idtac
  | |- represents (Cons C ?a) (?b/2) =>
       replace (b/2)%R with ((2*(b - 1/2)+1)/4)%R;
       [idtac | field; auto with real1]
  | |- _ => idtac
  end.
  
Ltac interval_from_represents H :=
 (generalize (R_interval _ _ H) || 
  generalize (L_interval _ _ H) ||
  generalize (C_interval _ _ H) ||
  generalize (represents_interval _ _ H)).

Ltac intervals_from_represents H H0 :=
  interval_from_represents H; interval_from_represents H0; intros; dec_tac.

Ltac inversion_to_variable x :=
  match goal with
  | id : represents x _ |- _ => idtac
  | id : represents (Cons ?d x) _ |- context [Cons ?d x] => idtac
  | id : represents (Cons ?d1 x) _ |- context [Cons ?d2 x] => inversion id
  | id : represents (Cons ?d1 (Cons ?d2 x)) _ |- context [ Cons ?d3 x] =>
    inversion id; inversion_to_variable x
  end.

Ltac constructor' hyp :=
  match goal with
    |- represents (C::_) ?v =>
    replace v with ((2*(2*v - 1/2)+1)/4)%R;
         [constructor | field; clear hyp; auto with real1]
  | |- represents (L::_) ?v =>
    replace v with ((2*v)/2)%R;
         [constructor | field; clear hyp; auto with real1]
  | |- represents (R::_) ?v  =>
    replace v with (((2*v-1)+1)/2)%R;
         [constructor | field; clear hyp; auto with real1]
  end.


*)
