Require Export realbasis.
Require Export proof_utils.
Require Export real_facts.
Require Export add_correct.

Open Scope Z_scope.

Theorem fact_pos :
  forall p n, p = Z_of_nat (fact n) -> 0 < p.
  intros; subst; apply (inj_lt 0); apply lt_O_fact.
Qed.

Theorem fact_step : forall n, (fact n * (n + 1) = fact (n+1))%nat.
  intros; rewrite plus_comm; simpl;
    rewrite mult_comm; simpl; auto.
Qed.

Theorem fact_gt_1 : forall p, 2 <= p -> 1 < (Z_of_nat (fact (Zabs_nat p))).
  intros p; elim p using (well_founded_ind (Zwf_well_founded 2)).
  clear p.
  intros p Hrec Hle; elim (Z_le_gt_dec p 2); intros Hcase.
  replace p with 2; compute; auto.
  omega.
  assert (Zabs_nat p - 1 = Zabs_nat (p - 1))%nat.
  rewrite Zabs_nat_minus.
  auto.
  omega.
  replace (Zabs_nat p) with (S (Zabs_nat (p - 1))).
  rewrite fact_simpl.
  replace (S (Zabs_nat (p - 1))) with (1+Zabs_nat (p - 1))%nat.
  rewrite inj_mult.
  assert (1 < Z_of_nat (fact (Zabs_nat (p - 1)))).
  apply Hrec; unfold Zwf; omega.
  apply Zlt_le_trans with (Z_of_nat (1 + (Zabs_nat (p - 1)))*1).
  rewrite Zmult_1_r.
  rewrite inj_plus.
  rewrite Zabs_nat_minus.
  rewrite inj_minus1.
  repeat rewrite Z_of_nat_Zabs_nat; simpl Z_of_nat; omega.

  rewrite Zabs_nat_1.
  assert (0 < Zabs_nat p)%nat.
  apply (Zabs_nat_lt 0);omega.
  omega.

  omega.
  apply Zmult_le_compat_l.
  omega.
  rewrite inj_plus.
  apply Zplus_le_0_compat.
  omega.
  omega.
  ring_nat.
  rewrite <- H.
  match goal with |- S ?x = _ => change (S x) with (1 + x)%nat end.
  rewrite le_plus_minus_r.
  auto.
  assert (0 < Zabs_nat p)%nat.
  apply (Zabs_nat_lt 0);omega.
  omega.
Qed.

Theorem fact_step_minus : forall n, (1 <= n -> fact (n - 1) * n = fact n)%nat.
  intros n H; induction H; simpl; auto.
  rewrite <- minus_n_O; ring_nat.
Qed.

Theorem lt_0_inv_fact :
  forall k, (0 < 1 /INR(fact k))%R.
  intros k; unfold Rdiv.
  apply Rlt_mult_inv_pos.
  fourier.
  rewrite INR_IZR_INZ.
  apply (IZR_lt 0).
  apply (inj_lt 0).
  apply lt_O_fact.
Qed.

Theorem fact_under_estimate :
  forall k i, 0<= k ->
    Z_of_nat (fact(Zabs_nat k)) * k^Z_of_nat i
    <= Z_of_nat (fact(i+(Zabs_nat k))).
  intros k i Hkpos; induction i.
  simpl.
  rewrite Zmult_1_r; auto with zarith.
  pattern (S i) at 1;replace (S i) with (i + 1)%nat.
  rewrite inj_plus.
  rewrite Zpower_exp.
  rewrite Zmult_assoc.
  replace (k^Z_of_nat 1) with k.
  simpl (S i + Zabs_nat k)%nat.
  rewrite fact_simpl.
  rewrite inj_mult.
  rewrite Zmult_comm.
  apply Zmult_le_compat.
  replace (S (i + Zabs_nat k)) with (1 + (i + Zabs_nat k))%nat.
  repeat rewrite inj_plus.
  simpl Z_of_nat.
  rewrite Z_of_nat_Zabs_nat.
  omega.
  assumption.
  ring_nat.
  exact IHi.
  assumption.
  apply Zmult_le_0_compat.
  apply Zle_0_nat.
  apply Zpower_le_0_compat; assumption.
  simpl; unfold Zpower_pos; simpl; ring.
  omega.
  omega.
  ring_nat.
Qed.

Theorem fact_pm1_fact_p :
  forall p fact_pm1, 2 <= p ->
    fact_pm1 = Z_of_nat (fact (Zabs_nat (p - 1))) ->
    fact_pm1 * p = Z_of_nat (fact (Zabs_nat p)).
  intros p fact_pm1 Hpge2 Hfact.
  rewrite Hfact.
  pattern p at 2; rewrite <- Z_of_nat_Zabs_nat.
  rewrite <- inj_mult.
  rewrite Zabs_nat_minus.
  rewrite Zabs_nat_1.
  rewrite fact_step_minus.
  auto.
  apply (Zabs_nat_le 1).
  omega.
  omega.
  omega.
Qed.

Theorem infinit_sum_inv_fact_bound :
  forall p r, 
    2 <= p ->
    infinit_sum (fun i => 1/INR(fact (i+Zabs_nat p)))%R r ->
    (0 < r <= 1/(INR (fact(Zabs_nat p-1))*(IZR p - 1)))%R.
  intros p r Hpeg2 Hinf.
  generalize (infinit_sum_cv _ _ Hinf).
  intros Hcv.
  split.
  apply Rlt_le_trans with (1/(INR(fact(Zabs_nat p))))%R.
  apply lt_0_inv_fact.
  assert (Hpos : (forall i,  0 <= 1/(INR(fact(i + Zabs_nat p))))%R).
  intros n.
  apply Rlt_le.
  apply lt_0_inv_fact.
  exact (sum_incr _ 0 r Hcv Hpos).
  simpl sum_f_R0.

  assert ((forall n, 
    sum_f_R0 (fun i => (1/INR(fact (i+Zabs_nat p)))%R) n <=
    1/(INR (fact(Zabs_nat p - 1))*(IZR p -1)))%R).
  intros n.
  replace (1/(INR(fact (Zabs_nat p - 1))*(IZR p - 1)))%R with
    (1/(INR(fact (Zabs_nat p))) *(IZR p/(IZR p - 1)))%R.
  assert (H1ltp : 1 < p).
  omega.
  assert (Hfactpos : (0 < INR(fact(Zabs_nat p)))%R).
  apply lt_INR_0.
  apply lt_O_fact.

  apply Rle_trans with
    (2:= tech_exp (INR (fact (Zabs_nat p))) p n H1ltp Hfactpos).
  apply sum_Rle.
  intros n' _; unfold Rdiv; repeat rewrite Rmult_1_l.
  apply Rle_Rinv.
  apply Rmult_lt_0_compat.
  apply (lt_INR 0); apply lt_O_fact.
  rewrite power_IZR; apply pow_lt; apply (IZR_lt 0); omega.
  apply (lt_INR 0); apply lt_O_fact.
  repeat rewrite INR_IZR_INZ.
  rewrite <- mult_IZR.
  apply IZR_le.
  apply fact_under_estimate.
  omega.
  replace (INR(fact (Zabs_nat p))) with
    (INR(fact (Zabs_nat p - 1)*(Zabs_nat p))).
  repeat rewrite INR_IZR_INZ.
  rewrite inj_mult.
  rewrite Z_of_nat_Zabs_nat.
  rewrite mult_IZR.
  field.
  assert ((IZR p <> 0)%R).
  apply not_O_IZR; omega.
  assert ((IZR p + -1 <> 0)%R).
  replace (-1)%R with (IZR (-1)).
  rewrite <- plus_IZR.
  apply not_O_IZR; omega.
  apply (Ropp_Ropp_IZR 1).
  assert ((IZR (Z_of_nat (fact (Zabs_nat p - 1))) <> 0)%R).
  rewrite <- INR_IZR_INZ.
  apply not_O_INR.
  apply fact_neq_0.
  repeat apply prod_neq_R0;auto.
  omega.
  rewrite fact_step_minus.
  auto.
  apply (Zabs_nat_le 1).
  omega.
  apply Rle_cv_lim  with
    (Vn:= (fun i:nat => 1/(INR(fact (Zabs_nat p - 1))*(IZR p -1)))%R) 
    (1:= H) (2:=Hcv).
  intros eps; exists 0%nat; intros n _; rewrite R_dist_eq.
  auto with real.
Qed.

Ltac generalize_in_all x k :=
  match goal with 
    id : context [x] |- _ => 
      generalize id; clear id; generalize_in_all x k; intros id
    | _ => generalize x;intro k
  end.

Theorem IZR_Z_of_nat_0_lt_neq : forall a, 
  (O < a)%nat -> IZR (Z_of_nat a) <> 0%R. 
Proof.
intros.
change 0%R with (IZR (Z_of_nat O)).
apply IZR_neq.
generalize (inj_lt 0 a H).
omega.
Qed.

Theorem e_init_correct : 
  forall b p q, 
    1 < b -> 
    (p, q) = e_minus2_init (Zabs_nat (2*b)) ->
    q = Z_of_nat (fact (Zabs_nat (2*b-1))) /\
    sum_f_R0 (fun i : nat => 
      (1 / INR (fact (i + Zabs_nat 2)))%R) (Zabs_nat (2*b-3))= (IZR p / IZR q)%R /\
    (0<p<q).
Proof.
induction b.
discriminate 1.
2:inversion 1.
intros p0 q H.
assert (2<Zabs_nat (2 * Zpos p))%nat.
change 2%nat with (Zabs_nat 2).
apply Zabs_nat_lt.
omega.
rewrite Zabs_nat_minus.
2:omega.
rewrite Zabs_nat_minus.
2:omega.
clear H.
generalize H0 p0 q; clear H0 p0 q.
generalize (Zabs_nat (2 * Zpos p)).
clear p.
change (Zabs_nat 3) with 3%nat.
change (Zabs_nat 2) with 2%nat.
change (Zabs_nat 1) with 1%nat.
induction 1.
simpl.
inversion 1.
simpl; split.
reflexivity.
split.
reflexivity.
omega.
intros p q.
simpl.
caseEq (e_minus2_init m).
intros p1 q1 Hp1q1 Hpq.
destruct m.
absurd (3<=0); omega.
destruct m.
absurd (3<=1); omega.
destruct m.
absurd (3<=2); omega.
generalize_in_all (S (S (S m))) k.
inversion Hpq; clear Hpq.
symmetry in Hp1q1.
generalize (IHle _ _ Hp1q1); clear IHle.
intros [IH1 [IH2 [IH3 IH4]]].
subst.
rewrite <- inj_mult.
rewrite fact_step_minus; [idtac|omega].
split.
replace (k-0)%nat with k; [reflexivity|omega].
replace (k-2)%nat with (S (k-3))%nat;[idtac|omega].
simpl.
rewrite IH2.
replace (k-3+2)%nat with (k-1)%nat; [idtac|omega].
replace (fact (k - 1) + (k - 1) * fact (k - 1))%nat with 
  (fact (k-1)*(k-1+1))%nat; [idtac| ring_nat].
replace (k-1+1)%nat with k; [idtac|omega].
rewrite INR_IZR_INZ.
rewrite <- (fact_step_minus k); [idtac|omega].
rewrite inj_mult.
rewrite plus_IZR.
repeat rewrite mult_IZR.
simpl.
split.
newfield.
split; apply IZR_Z_of_nat_0_lt_neq.
omega.
apply lt_O_fact.
split.
change 0 with (0+0).
apply Zplus_lt_compat; [idtac|omega].
apply Zmult_lt_O_compat; omega.
apply Zlt_le_trans with ((p1+1)*Z_of_nat k).
rewrite Zmult_plus_distr_l.
apply Zplus_lt_compat_l.
omega.
apply Zmult_le_compat_r;[idtac|omega].
omega.
Qed.

Theorem tech_bound :
  forall b y r u, 
    0<b ->
    (0 < IZR u)%R -> 
    (0 < r <= 1/IZR u)%R -> 
    4*b*b*y <= u -> 
    (2*IZR b * (IZR y * r) <= 1/(2*IZR b))%R.
Proof.
  intros; dec_tac.
  apply Rmult_le_reg_l with (2*IZR b)%R.
  rewrite <- (mult_IZR 2).
  apply (IZR_lt 0); omega.
  replace (2 * IZR b * (1 / (2 * IZR b)))%R with 
    1%R;[idtac | newfield; discrR; omega].
  replace (2 * IZR b * (2 * IZR b * (IZR y * r))) %R with 
    (IZR (4*b*b*y) *r)%R.
  apply Rle_trans with (IZR u * r)%R.
  apply Rmult_le_compat_r.
  fourier.
  apply IZR_le; assumption.
  replace 1%R with (IZR u * (1/IZR u))%R; [idtac| newfield].
  apply Rmult_le_compat_l.
  fourier.
  assumption.
  apply Rlt_neq_0; assumption.
  repeat rewrite mult_IZR; simpl IZR;ring.
Qed.

Ltac apply_e_correct1' e_series' :=
  match goal with 
    e_correct1 : _, id : represents ?v ?vr, id1 : infinit_sum _ ?r
    |- represents (e_series' ?v (?n, ?p, ?f)) ?w =>
      replace w with (vr + IZR n * r)%R;
        [ apply e_correct1;clear e_correct1;auto with real1 | idtac]
  end.

Ltac add_fact x :=
  match goal with 
    | id: represents _ (Cons (I ?b ?d1) x) _,
      Hbpos : 0 < ?b  
      |- context [Cons (I _ ?d1) x] =>
	generalize (I_interval _ _ _ _ Hbpos id); intros
    | _ => idtac
  end.

Theorem tech_bound2 : 
  forall b k' y r r0,
    0<b ->
    k' <= 2 * b - 3 ->
    (2 * IZR b * (IZR y * r) <= 1 / (2 * IZR b) ->
      0 < IZR y * r ->
      -1<= r0 ->
      r0 <= (IZR k' + 1) / IZR (2 * b) ->
      -1 <= r0 + (IZR (2*b*y) * r) <= 1)%R.
Proof.
  intros b k' y r r0 Hbpos Hk'le H2byr Hyrpos Hr0ge Hr0le.
  split.
  apply Rle_trans with (r0+0)%R.
  intuition.
  apply Rplus_le_compat_l.
  rewrite mult_IZR.
  rewrite Rmult_assoc.
  apply Rmult_le_pos.
  apply (IZR_le 0); omega.
  apply Rlt_le; assumption.
  apply Rle_trans with ((IZR k' + 1) / IZR (2 * b)+ 1 / (2 * IZR b))%R.
  apply Rplus_le_compat.
  dec_tac; assumption.
  repeat rewrite mult_IZR; simpl.
  rewrite Rmult_assoc.
  assumption.
  apply Rmult_le_reg_l with (2*IZR b)%R.
  rewrite <- (mult_IZR 2).
  apply (IZR_lt 0); omega.
  replace (2 * IZR b * ((IZR k' + 1) / IZR (2 * b) + 1 / (2 * IZR b)))%R with 
    (IZR k' + 2)%R.
  rewrite Rmult_1_r.
  change 2%R with (IZR 2).
  rewrite <- (plus_IZR k').
  rewrite <- mult_IZR.
  apply IZR_le.
  omega.
  rewrite mult_IZR; simpl.
  newfield; discrR; omega.
Qed.

Theorem tech_bound3 : 
  forall b y r r0 vr,
    0<b ->
    (vr + IZR y * r <= 1 ->
      (IZR (2 * b - 1) + r0) / IZR (2 * b) = vr ->
      0 < IZR y * r ->
      -1<= r0 ->
      -1 <= r0 + IZR (2 * b * y) * r <= 1)%R.
Proof.
  intros b y r r0 vr Hbpos U1 Hvreq Hyrpos Hr0.  
  split.
  apply Rle_trans with (r0+0)%R.
  intuition.
  apply Rplus_le_compat_l.
  rewrite mult_IZR.
  rewrite Rmult_assoc.
  apply Rmult_le_pos.
  apply (IZR_le 0); omega.
  apply Rlt_le; assumption.
  replace 1%R with (IZR (2*b)*(1+-1)+1)%R; [idtac|ring].
  replace r0 with (vr*IZR(2*b)-IZR(2*b)+1)%R; [idtac| subst vr; 
  rewrite <- Z_R_minus; repeat rewrite mult_IZR; simpl; newfield; discrR; omega].
  replace (vr * IZR (2 * b) - IZR (2 * b) + 1 + IZR (2 * b * y) * r)%R with 
    ((IZR (2*b) * (vr + IZR y * r + -1))+1)%R; [idtac| repeat rewrite mult_IZR; 
      simpl; ring].
  apply Rplus_le_compat_r.
  apply Rmult_le_compat_l.
  apply (IZR_le 0); omega.
  apply Rplus_le_compat_r.
  assumption.
Qed.

Theorem tech_bound4 : 
  forall b y r r1 k' k,
    k' > 2 * b - 3 ->
    k' <= 2 * b - 1 ->
    3 <= b ->
    2 * b - 1 > k -> 
    (-1 <= r1 <= 1->  
      2 * IZR b * (IZR y * r) <= 1 / (2 * IZR b) ->
      0 < IZR y * r ->
      -1 <= (IZR (k' - 2 * b) + r1) / IZR (2 * b) + IZR (2 * b * y) * r <= 1)%R.
Proof.
  intros b y r r1 k' k Hk'gt Hk'le H3leb Hkle Hr1 H2byr Hyr.
  split.
  apply Rle_trans with (-2/IZR b+0)%R.
  apply Rmult_le_reg_l with (IZR b).
  apply (IZR_lt 0); omega. 
  replace (IZR b * (-2/IZR b+0))%R with (-2)%R;[idtac|newfield; discrR; omega].
  replace (IZR b*-1)%R with (IZR (-b)); [idtac| rewrite Ropp_Ropp_IZR; ring].
  apply (IZR_le (-b) (-2)); omega.
  apply Rplus_le_compat.
  apply Rmult_le_reg_l with (IZR (2*b)).
  apply (IZR_lt 0); omega.
  replace (IZR (2 * b) * (-2 / IZR b))%R with (-3+-1)%R; [idtac| 
    rewrite mult_IZR; simpl; newfield; discrR; omega].
  replace (IZR (2 * b) * ((IZR (k' - 2 * b) + r1) / IZR (2 * b)))%R with 
    (IZR (k' - 2 * b) + r1)%R; [idtac| newfield; discrR; omega].
  apply Rplus_le_compat.
  replace (-3)%R with (IZR (-3)); [idtac| simpl; ring].
  apply IZR_le; omega.
  intuition.
  rewrite mult_IZR.
  rewrite Rmult_assoc.
  apply Rmult_le_pos.
  apply (IZR_le 0); omega.
  apply Rlt_le; assumption.

  apply Rle_trans with ((IZR (k' - 2 * b) + r1) / IZR (2 * b) + 1 / (2 * IZR b))%R.
  apply Rplus_le_compat_l.
  repeat rewrite mult_IZR; simpl; rewrite Rmult_assoc; assumption.
  apply Rmult_le_reg_l with (IZR (2*b))%R.
  apply (IZR_lt 0); omega.
  replace (IZR (2*b) * ((IZR (k'-2*b) + r1) / IZR (2*b) + 1 / (2*IZR b)))%R with 
    (IZR (k'-2*b) + (r1 +1))%R; [idtac| 
      rewrite mult_IZR; simpl; newfield; discrR; omega].
  apply Rle_trans with (IZR (k' - 2 * b) + 2)%R.
  apply Rplus_le_compat_l.
  intuition.
  rewrite Rmult_1_r.
  change 2%R with (IZR 2).
  rewrite <- plus_IZR.
  apply IZR_le; omega.
Qed.

Theorem tech_bound5 : 
  forall b y r r1 k' k,
    k' < -2 * b + 3 ->
    k' >= -2 * b + 1 ->
    3 <= b ->
    -2 * b + 1 < k -> 
    (-1 <= r1 <= 1->  
      2 * IZR b * (IZR y * r) <= 1 / (2 * IZR b) ->
      0 < IZR y * r ->
      -1 <= (IZR (k' + 2 * b) + r1) / IZR (2 * b) + IZR (2 * b * y) * r <= 1)%R.
Proof.
  intros b y r r1 k' k Hk'gt Hk'le H3leb Hkle Hr1 H2byr Hyr.
  split.
  apply Rle_trans with (0+0)%R.
  intuition.
  apply Rplus_le_compat.
  apply Rmult_le_reg_l with (IZR (2*b)).
  apply (IZR_lt 0); omega.
  replace (IZR (2 * b) * 0)%R with (1+-1)%R; [idtac| 
    rewrite mult_IZR; simpl; newfield; discrR; omega].
  replace (IZR (2 * b) * ((IZR (k' + 2 * b) + r1) / IZR (2 * b)))%R with 
    (IZR (k' + 2 * b) + r1)%R; [idtac| newfield; discrR; omega].
  apply Rplus_le_compat.
  apply (IZR_le 1); omega.
  intuition.
  rewrite mult_IZR.
  rewrite Rmult_assoc.
  apply Rmult_le_pos.
  apply (IZR_le 0); omega.
  apply Rlt_le; assumption.

  apply Rle_trans with ((IZR (k' + 2 * b) + r1) / IZR (2 * b) + 1 / (2 * IZR b))%R.
  apply Rplus_le_compat_l.
  repeat rewrite mult_IZR; simpl; rewrite Rmult_assoc; assumption.
  apply Rmult_le_reg_l with (IZR (2*b))%R.
  apply (IZR_lt 0); omega.
  replace (IZR (2*b) * ((IZR (k'+2*b) + r1) / IZR (2*b) + 1 / (2*IZR b)))%R with 
    (IZR (k'+2*b) + (r1 +1))%R; [idtac| 
      rewrite mult_IZR; simpl; newfield; discrR; omega].
  apply Rle_trans with (IZR (k' + 2 * b) + 2)%R.
  apply Rplus_le_compat_l.
  intuition.
  rewrite Rmult_1_r.
  change 2%R with (IZR 2).
  rewrite <- plus_IZR.
  apply IZR_le; omega.
Qed.

Theorem e_series_body_correct :
  forall b, 3 <= b ->
    forall (e_series : stream (idigit b) -> Z*Z*Z -> stream (idigit b)),
    (forall v vr r y p fact_pm1,
      fact_pm1 = (Z_of_nat (fact (Zabs_nat (p-1)))) ->
      2*b * y <= fact_pm1 * (p - 1) ->
      2*b <= p -> 
      1 <= y -> 
      represents b v vr ->
      infinit_sum (fun i => (1/INR(fact (i+Zabs_nat p)))%R) r ->
      (vr + (IZR y) *r <= 1)%R ->
      represents b (e_series v (y,p,fact_pm1))(vr+(IZR y)*r)%R) ->
    forall v vr r y p fact_pm1,
      fact_pm1 = (Z_of_nat (fact (Zabs_nat (p-1)))) ->
      4*b*b* y <= fact_pm1 * (p - 1) ->
      2*b <= p ->
      1 <= y -> 
      represents b v vr ->
      infinit_sum (fun i => (1/INR(fact (i+Zabs_nat p)))%R) r ->
      (vr + (IZR y) *r <= 1)%R ->
      represents b (series_body b (Z*Z*Z) e_series v (2*b*y, p, fact_pm1))
      (vr+(IZR y)*r)%R.
Proof.
  intros b H4leb e_series' e_correct1.
  intros v vr r y p fact_pm1 Hfact Hle Hpgeb Hnge1 Hv Hsum Hin_res.
  unfold series_body.
  assert (Hpge2 : 2<= p).
  clear e_correct1.
  omega.
  assert (Hfactpos : 0 < fact_pm1).
  clear e_correct1.
  rewrite Hfact.
  apply (inj_lt 0).
  apply lt_O_fact.
  assert (Hprodpos : (0 < IZR(fact_pm1 *(p-1)))%R).
  clear e_correct1.
  apply (IZR_lt 0); apply Zmult_lt_O_compat.
  assumption.
  omega.
  assert (Hboundr : (2*IZR b*(IZR y* r) <= 1/(2*IZR b))%R).
  clear e_correct1.
  apply tech_bound with (fact_pm1 * (p - 1)).
  omega.
  assumption.
  generalize (infinit_sum_inv_fact_bound _ _ Hpge2  Hsum).
  rewrite INR_IZR_INZ; rewrite mult_IZR; rewrite Hfact.
  rewrite <- Z_R_minus; simpl IZR.
  rewrite Zabs_nat_minus.
  intros; assumption.
  omega.
  assumption.
  assert (Hnrpos : (0 < IZR y * r)%R).
  clear e_correct1.
  apply Rmult_lt_0_compat.
  apply (IZR_lt 0); omega.
  generalize (infinit_sum_inv_fact_bound _ _ Hpge2  Hsum).
  intros; dec_tac; assumption.
  assert (Hbpos: 0 < b).
  clear e_correct1.
  omega.
  assert (H1le2by: 1<= 2*b*y).
  clear e_correct1.
  apply Zle_trans with (2*3*1).
  omega.
  repeat rewrite <- Zmult_assoc.
  apply Zmult_le_compat_l.
  apply Zmult_le_compat; omega.
  omega.
  assert (H2b2b: 2 * b * (2 * b * y) <= fact_pm1 * (p - 1)).
  clear e_correct1.
  replace (2*b*(2*b*y)) with (4*b*b*y);[assumption|ring].
  
  destruct v as [ [k] v']; inversion Hv.
  destruct v' as [ [k'] v'']; inversion_to_variable v''.
  elim (Z_le_gt_dec (-2 * b + 3) k').
  elim (Z_le_gt_dec k' (2 * b - 3)).
  intros Hle1 Hle2.

(* case : -2*b+3 <= k' <= 2*b-3*)
  add_fact v''.
  replace ((IZR k + r0) / IZR (2 * b) + IZR y * r)%R with 
    ((IZR k + (r0 + (IZR (2*b*y)*r)))/(IZR (2*b)))%R; [idtac| clear e_correct1; 
      repeat rewrite mult_IZR; newfield; split; discrR; omega].
  assert (Hbound : (-1<=r0 + (IZR (2*b*y) * r) <= 1)%R).
  clear e_correct1.
  dec_tac; apply tech_bound2 with k'; assumption.
  constructor; try (clear e_correct1; assumption).
  apply e_correct1; clear e_correct1; intuition.

(* case 2 * b - 3 < k'*) 
  intros Hgt _.
  elim  (Z_le_gt_dec (2 * b - 1) k).
  intros Hle1.
  assert (k=2 * b - 1).
  clear e_correct1.
  apply Zle_antisym; omega.

(*case 2 * b - 3 < k' and k = 2 * b - 1 *)
  add_fact v''.
  replace ((IZR k + r0) / IZR (2 * b) + IZR y * r)%R with 
    ((IZR k + (r0 + (IZR (2*b*y)*r)))/(IZR (2*b)))%R; [idtac|clear e_correct1; 
      repeat rewrite mult_IZR; newfield; split; discrR; omega].
  assert (-1 <= r0 + IZR (2 * b * y) * r <= 1)%R.
  clear e_correct1.
  subst k; subst k0; dec_tac.
  apply tech_bound3 with vr; assumption.
  constructor; try (clear e_correct1; assumption).
  apply e_correct1; clear e_correct1; intuition.
  
(*case 2 * b - 3 < k' and k < 2 * b - 1 *)
  intros Hgt1.
  inversion_to_variable v''.
  replace ((IZR k + (IZR k' + r1) / IZR (2 * b)) / IZR (2 * b) + IZR y * r)%R with 
    ((IZR (k+1) + ((IZR (k'-2*b) + r1) / IZR (2*b) + IZR (2*b*y) *r))/(IZR (2*b)))%R;
    [idtac| clear e_correct1; rewrite plus_IZR; rewrite <- Z_R_minus; 
      repeat rewrite mult_IZR; simpl ; newfield; discrR; omega].
  assert (-1 <= (IZR (k'-2*b) + r1) / IZR (2*b) + IZR (2*b*y)*r <= 1)%R.
  clear e_correct1.
  dec_tac; apply tech_bound4 with k; intuition.
  constructor; try (clear e_correct1; assumption).
  apply e_correct1; clear e_correct1; intuition.
  constructor; intuition.
  split; omega.

(* case -2 * b + 3 >  k'*) 
  intros Hgt.
  elim  (Z_le_gt_dec k (-2 * b + 1)).
  intros Hle1.
  assert (k=-2 * b + 1).
  apply Zle_antisym; omega.

(*case -2 * b + 3 < k' and k = -2 * b + 1 *)
  add_fact v''.
  replace ((IZR k + r0) / IZR (2 * b) + IZR y * r)%R with 
    ((IZR k + (r0 + (IZR (2*b*y)*r)))/(IZR (2*b)))%R; [idtac|clear e_correct1; 
      repeat rewrite mult_IZR; newfield; split; discrR; omega].
  assert (-1 <= r0 + IZR (2 * b * y) * r <= 1)%R.
  clear e_correct1.
  subst k; subst k0; dec_tac.
  apply tech_bound2 with k'; intuition.
  constructor; try (clear e_correct1; assumption).
  apply e_correct1; clear e_correct1; intuition.

(*case -2 * b + 3 < k' and k > -2 * b + 1 *)
  intros Hgt1.
  inversion_to_variable v''.
  replace ((IZR k + (IZR k' + r1) / IZR (2 * b)) / IZR (2 * b) + IZR y * r)%R with 
    ((IZR (k-1) + ((IZR (k'+2*b) + r1) / IZR (2*b) + IZR (2*b*y) *r))/(IZR (2*b)))%R;
    [idtac| clear e_correct1; rewrite plus_IZR; rewrite <- Z_R_minus; 
      repeat rewrite mult_IZR; simpl ; newfield; discrR; omega].
  assert (-1 <= (IZR (k'+2*b) + r1) / IZR (2*b) + IZR (2*b*y)*r <= 1)%R.
  clear e_correct1.
  apply tech_bound5 with k; intuition.
  constructor; try (clear e_correct1; assumption).
  apply e_correct1; clear e_correct1; intuition.
  constructor; intuition.
  split; omega.
Defined.

Theorem e_correct1_tech_lemma : 
  forall b y fact_pm1 p,
    0 < b -> 
    fact_pm1 = Z_of_nat (fact (Zabs_nat (p - 1))) ->
    2*b <= p ->
    1 <= y ->
    2*b * y <= fact_pm1 * (p - 1) ->
    (-1 <= IZR y / IZR (fact_pm1 * p) <= 1)%R.
Proof. 
intros b y fact_pm1 p Hbpos Hfact Hpge2 Hnge1 Hinv.
unfold Rdiv.
assert (0 < IZR (fact_pm1 * p))%R.
apply (IZR_lt 0).
apply Zlt_le_trans with (fact_pm1 * (p - 1)).
apply Zlt_le_trans with (2:=Hinv).
apply Zmult_lt_O_compat; omega.
apply Zmult_le_compat_l; omega.

split.
apply Rle_trans with 0%R.
fourier.
apply Rle_mult_inv_pos.
apply (IZR_le 0); omega.
assumption.

apply Rmult_le_reg_l with (IZR (fact_pm1 * p))%R.
assumption.
rewrite <- Rmult_assoc.
rewrite Rinv_r_simpl_m.
rewrite Rmult_1_r.
apply IZR_le.
replace y with (1*y); [idtac|ring].
apply Zle_trans with (2*b*y).
apply Zmult_le_compat_r; omega.
apply Zle_trans with (1:=Hinv).
apply Zmult_le_compat_l; omega.
apply Rgt_not_eq.
auto with real.
Qed.

Theorem e_correct1 :
  forall b, 3<= b ->
  forall v vr r y p fact_pm1, 
    fact_pm1 = (Z_of_nat (fact (Zabs_nat (p-1)))) ->
    2*b*y <= fact_pm1 * (p-1) ->
    2*b <= p ->
    1 <= y ->
    represents b v vr ->
    infinit_sum (fun i => (1/INR(fact (i+Zabs_nat p)))%R) r ->
    (vr + (IZR y)*r <= 1)%R ->
    represents b (e_series b v (y, p, fact_pm1)) (vr+(IZR y)*r)%R.
Proof.
  cofix.
  intros b H3leb v vr r y p fact_pm1 Hfact Hinv Hpgeb Hyge1 Hrepr Hinf Hin.
  rewrite (str_dec_thm (e_series b v (y, p, fact_pm1))); unfold str_decompose.
  lazy zeta beta iota delta [e_series]; fold e_series.
  elim (Z_le_gt_dec (4*b*b*y) (fact_pm1 *(p - 1))).
  intros Hle.
  fold (str_decompose (series_body b (Z*Z*Z) (e_series b) v (2*b*y,p,fact_pm1))).
  rewrite <- str_dec_thm.
  exact (e_series_body_correct b H3leb (e_series b) (e_correct1 b H3leb)  v vr r y p fact_pm1
    Hfact Hle Hpgeb Hyge1 Hrepr Hinf Hin).

  intros Hgt.
  replace (vr + IZR y * r)%R with
    ((vr + (IZR y)/(IZR (fact_pm1 * p))) + (IZR y) *
      (r - 1/(IZR (fact_pm1 * p))))%R;
    [idtac | clear e_correct1].
  fold (str_decompose (series_body b (Z*Z*Z) (e_series b)
    (add b v (rat_to_stream b y (fact_pm1 * p)))
    (2 * b*y, p + 1, fact_pm1 * p))); rewrite <- str_dec_thm.
  apply (e_series_body_correct b H3leb (e_series b) (e_correct1 b H3leb) 
    (add b v (rat_to_stream b y (fact_pm1 * p)))
    (vr + (IZR y)/(IZR (fact_pm1 * p)))%R
    (r - 1/(IZR (fact_pm1 * p)))%R
    y (p + 1) (fact_pm1 * p)); clear e_correct1.
  ring (p + 1 - 1).
  pattern p at 1; rewrite <- (Z_of_nat_Zabs_nat p).
  rewrite Hfact.
  rewrite <- inj_mult.
  rewrite Zabs_nat_minus.
  rewrite Zabs_nat_1.
  rewrite (fact_step_minus (Zabs_nat p)).
  auto.
  rewrite <- Zabs_nat_1.
  apply lt_le_weak.
  apply Zabs_nat_lt.
  omega.
  omega.
  omega.
  apply Zle_trans with (2*b* (fact_pm1 * (p - 1))).
  replace ( 4 * b * b * y) with ((2*b)*(2*b*y)); [idtac|ring].
  apply Zmult_le_compat_l; omega.
  rewrite Zmult_permute.
  rewrite <- (Zmult_assoc fact_pm1).
  apply Zmult_le_compat_l.
  apply Zmult_le_compat; omega.
  rewrite Hfact.
  apply Zle_0_nat.
  omega.
  omega.
  apply add_correct.
  omega.
  assumption.
  apply rat_to_stream_correct.
  omega.

  apply Zlt_not_eq.
  apply Zlt_le_trans with (fact_pm1 * (p - 1)).
  apply Zlt_le_trans with (2:=Hinv).
  apply Zmult_lt_O_compat; omega.
  apply Zmult_le_compat_l; omega.

  apply e_correct1_tech_lemma with b; intuition.

  split.
  rewrite mult_IZR.
  assert (Hbpos: 0<b).
  omega.
  generalize (represents_interval _ _ _ Hbpos Hrepr).
  intros [Hvr _].
  assert (0 <= IZR y / (IZR fact_pm1 * IZR p))%R.
  apply Rdiv_le_0_compat.
  apply (IZR_le 0); omega.
  rewrite <- mult_IZR.
  apply (IZR_lt 0). 
  apply Zmult_lt_O_compat.
  apply (fact_pos _ _ Hfact).
  omega.
  fourier.

  apply Rle_trans with (2:= Hin).
  apply Rplus_le_compat_l.
  replace (IZR y/IZR(fact_pm1 * p))%R with (IZR y * (1/IZR(fact_pm1 * p)))%R.
  apply Rmult_le_compat_l.
  apply (IZR_le 0).
  omega.

  2: field;auto with real1.
  5: field;auto with real1.
  generalize (infinit_sum_step _ _ Hinf); intros Hinf_step.
  simpl (0 + Zabs_nat p)%nat in Hinf_step.
  assert (Hext:forall n,
    ((fun i => 1/INR(fact (i+1+Zabs_nat p))) n=
      (fun i => 1/INR(fact (i+(Zabs_nat (p + 1))))) n)%R).
  intros i; simpl; replace (i+1+Zabs_nat p)%nat 
    with (i + (Zabs_nat (p + 1)))%nat.
  auto.
  rewrite Zabs_nat_plus.
  rewrite Zabs_nat_1.
  ring.
  omega.
  omega.
  generalize 
    (infinit_sum_ext 
      (fun i => 1/INR(fact(i + 1 + Zabs_nat p)))%R
      (fun i => 1/INR(fact(i + (Zabs_nat (p + 1)))))%R  _ Hext Hinf_step).
  intros Hinf_step'.
  assert (Hpp1ge2 : 2 <= p + 1).
  omega.
  generalize (infinit_sum_inv_fact_bound (p+1) _ Hpp1ge2 Hinf_step').
  intros Hint'.
  rewrite INR_IZR_INZ in Hint'.
  dec_tac.
  rewrite <- fact_step_minus in H.
  rewrite <- Zabs_nat_1 in H.
  rewrite <- Zabs_nat_minus in H.
  rewrite inj_mult in H.
  rewrite <- Hfact in H.
  rewrite Z_of_nat_Zabs_nat in H.
  fourier.
  omega.
  omega.
  rewrite <- Zabs_nat_1.
  apply Zabs_nat_le.
  omega.

  rewrite mult_IZR.
  apply prod_neq_R0.
  apply not_O_IZR.
  assert (0 < fact_pm1).
  rewrite Hfact.
  apply (inj_lt 0).
  apply lt_O_fact.
  omega.
  apply not_O_IZR.
  omega.

  apply infinit_sum_ext with (fun i => 1/INR(fact ((i+1)+Zabs_nat p)))%R.
  replace (IZR (fact_pm1 * p)) with (INR (fact (0 + (Zabs_nat p)))).
  intros i; rewrite Zabs_nat_plus.
  rewrite Zabs_nat_1.
  rewrite (plus_comm (Zabs_nat p)).
  rewrite plus_assoc.
  auto.
  omega.
  omega.
  simpl.
  rewrite INR_IZR_INZ.
  rewrite Hfact.
  pattern p at 3; rewrite <- Z_of_nat_Zabs_nat.
  rewrite <- inj_mult.
  rewrite Zabs_nat_minus.
  rewrite Zabs_nat_1.
  rewrite fact_step_minus.
  auto.
  rewrite <- Zabs_nat_1; apply Zabs_nat_le.
  omega.
  omega.
  omega.
  rewrite fact_pm1_fact_p.
  rewrite <- INR_IZR_INZ.
  apply infinit_sum_step with (An := (fun i => 1/INR (fact (i + Zabs_nat p)))%R).
  assumption.
  omega.
  assumption.

  match goal with |- (?x <= 1)%R => replace x with (vr + IZR y * r)%R end.
  assumption.
  unfold Rdiv; ring.

  rewrite mult_IZR.
  apply prod_neq_R0.
  apply not_O_IZR.
  assert (0 < fact_pm1).
  rewrite Hfact.
  apply (inj_lt 0).
  apply lt_O_fact.
  omega.
  apply not_O_IZR.
  omega.
Qed.

Theorem infinit_sum_step_k:
  forall (An : nat -> R) (l : R) n,
    infinit_sum An l ->
    infinit_sum (fun i : nat => An (i + S n)%nat) 
      (l - sum_f_R0 An n).
Proof.
induction n.
intros Hinf.
unfold sum_f_R0.
apply infinit_sum_step.
exact Hinf.
intros Hinf.
simpl.
generalize (IHn Hinf).
intros U.
generalize (infinit_sum_step _ _ U).
simpl.
replace (l - (sum_f_R0 An n + An (S n)))%R with 
  ((l-sum_f_R0 An n) - An (S n))%R; [idtac|ring].
apply infinit_sum_ext.
intros n0.
replace (n0 +1 + S n)%nat with (n0 +S (S n))%nat; [reflexivity| ring_nat].
Qed.

Theorem e_correct :
  forall b r,
    3 <= b -> 
    infinit_sum (fun i => (1/INR(fact(i+Zabs_nat 2)))%R) r ->
    represents b (number_e_minus2 b)  r.
Proof.
  intros b r Hbpos H.
  unfold number_e_minus2.
  caseEq (e_minus2_init (Zabs_nat (2 * b))).
  intros p q Hpqeq1.

  assert (Hpqeq : (p,q) = e_minus2_init (Zabs_nat (2 * b))).
  rewrite Hpqeq1; reflexivity.
  clear Hpqeq1.

  assert (H1ltb: 1<b).
  omega.
  generalize (e_init_correct _ _ _ H1ltb Hpqeq).
  intros [U1 [U2 [U3 U4]]].

  assert (U5: (-1 <= IZR p/IZR q<= 1)%R).
  generalize (IZR_lt _ _ U3).
  generalize (IZR_lt _ _ U4).
  simpl.
  intros; split.
  apply Rle_trans with 0%R.
  fourier.
  unfold Rdiv.
  apply Rle_mult_inv_pos; fourier.
  apply Rmult_le_reg_l with (IZR q); [fourier| idtac].
  unfold Rdiv.
  rewrite <- Rmult_assoc.
  rewrite Rinv_r_simpl_m.
  fourier.
  discrR; omega.

  assert (H1ltq: 1<q).
  subst;  apply fact_gt_1; omega.
  
  replace r with (IZR p/IZR q+(IZR 1)*(r-(IZR p/ IZR q)))%R;[idtac | ring].
  apply e_correct1; intuition.
  rewrite Zmult_1_r.
  apply Zmult_le_compat; intuition.
  apply rat_to_stream_correct; intuition.
  generalize (infinit_sum_step_k _ _ (Zabs_nat (2*b-3)) H).
  rewrite U2.
  apply infinit_sum_ext.
  intros.
  replace (n+ S (Zabs_nat (2 * b - 3)) + Zabs_nat 2)%nat with 
    (n + Zabs_nat (2 * b))%nat.
  reflexivity.
  replace (S (Zabs_nat (2 * b - 3))) with (Zabs_nat (2 * b - 3) + 1)%nat.
  2:ring_nat.
  change 1%nat with (Zabs_nat 1).
  rewrite <- Zabs_nat_plus; try omega.
  rewrite <- plus_assoc.
  rewrite <- Zabs_nat_plus; try omega.
  replace (2 * b - 3 + 1 + 2) with (2*b); ring.
  replace (IZR p/IZR q+(IZR 1)*(r-(IZR p/ IZR q)))%R with r%R;[idtac | ring].
  assert (U : 2<=2); [omega|idtac].
  generalize (infinit_sum_inv_fact_bound 2 r U H).
  simpl.
  replace (1/(1*(2-1)))%R with 1%R;[intuition| newfield].
Qed.
